#!/usr/bin/perl -w

# Copyright 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use File::Spec;
use List::Util 'max','min','sum';
use Math::Libm 'hypot';
use Math::BaseCnv 'cnv';
use Math::Trig 'rad2deg';
use Math::PlanePath::R5DragonCurve 117; # v.117 for level_to_n_range()
use Memoize 'memoize';
use Math::PlanePath::Base::Digits 'bit_split_lowtohigh',
  'digit_split_lowtohigh', 'digit_join_lowtohigh';
use Math::PlanePath;
*_divrem_mutate = \&Math::PlanePath::_divrem_mutate;
our @ipc;

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyR5Dragon ':all';
use MyFLAT ':all';

# uncomment this to run the ### lines
# use Smart::Comments;

$|=1;
our (@enc_table,@right_enc_table,@left_enc_table);


{
  # turns subword factor complexity by FLAT
  # 2,4,6,8,10,12,14,16,18,
  # 2*len

  require FLAT;
  my @pending = (FLAT::Regex->new('(0|1|2|3|4)*')->as_dfa); # any
  foreach my $len (1 .. 100) {
    my @new_pending;
    while (defined(my $f = shift @pending)) {
      $f = $f->MyFLAT::digits_increment(add => -1);
      foreach my $t (TurnLpred_flat(),
                     TurnRpred_flat()) {
        my $new = $f->intersect($t);
        if (! $new->is_empty) {
          $new = $new->MyFLAT::minimize(verbose=>0);
          push @new_pending, $new;
        }
      }
    }    
    @pending = @new_pending;
    # print scalar(@pending),",";
    print "len=$len ",scalar(@pending),
      " states ",sum(map{$_->num_states}@pending)/scalar(@pending),"\n";
  }
  print "\n";
  exit 0;
}


{
  # QuintetCurve inside area
  require Math::PlanePath::QuintetCurve;
  require MyPlanar;
  my $path = Math::PlanePath::QuintetCurve->new;
  foreach my $k (1,
                 # 0 .. 6
                ) {
    my $points = [];
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      push @$points, [$x,$y];
    }
    my ($xend,$yend) = $path->n_to_xy($n_hi);
    my $points2 = MyPlanar::points_move($points, -$xend,-$yend);
    $points2 = MyPlanar::points_rotate_plus90($points2, 2);
    push @$points, @$points2;

    foreach my $i (0 .. $#$points) {
      abs($points->[$i]->[0] - $points->[$i-1]->[0])
        + abs($points->[$i]->[1] - $points->[$i-1]->[1])
        == 1 or die;
    }

    ### $points
    my $A = MyPlanar::points_to_area($points);
    print "$A\n";
    foreach my $p (@$points) {
      print '(',$p->[0],',',$p->[1],')--';
    }      
  }    
  exit 0;
}


{
  # EpredR three consecutive

  my $orig = EpredL_flat();
  my $f = $orig;
  for (1 .. 2) {
    $f = $orig->intersect($f->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize(verbose=>1);
  }
  $orig->MyFLAT::reverse->MyFLAT::minimize->MyFLAT::view;
  $f->MyFLAT::reverse->MyFLAT::minimize->MyFLAT::view;
  exit 0;
}
{
  # EpredSix runs

  my $f = Epred_flat();
  for (1 .. 5) {
    $f = Epred_flat()->intersect($f->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize(verbose=>1);
  }
  $f->MyFLAT::reverse->MyFLAT::minimize->MyFLAT::view;
  exit 0;
}

{
  require MyPlanar;
  require Math::Libm;

  my $Hwpf = 0.57698859;
  my $Hepf = 0.17698859;
  my $points = [ [-$Hepf, -$Hwpf], [1+$Hepf, -$Hwpf],
                 [1+$Hepf, $Hwpf], [-$Hepf, $Hwpf] ];
  my $ex = 1;
  my $ey = 0;
  foreach (1 .. 2) {
    ($points, $ex,$ey) = MyR5Dragon::hull_expand($points, $ex,$ey, unsorted=>1);

    $points = [ map {
      my ($x,$y) = @$_;
      ($x,$y) = (   $x + 2*$y,
                    -2*$x +   $y);
      $x /= 5; $y /= 5;
      [$x,$y]
    } @$points ];
    ($ex,$ey) = (   $ex + 2*$ey,
                    -2*$ex +   $ey);
    $ex /= 5; $ey /= 5;
  }
  # say points_str($points);
  print scalar(@$points)," points, end $ex,$ey\n";
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    print("($x,$y)--");
  }
  print "\n";

  #    1,2
  #        |C
  #        v
  #    1,1  --D->
  #        .
  #        .
  #  0-A-->
  #        |B
  #        v
  #          1,-1

  my $A = $points;
  my $B = MyPlanar::points_move(MyPlanar::points_rotate_plus90($points),  1,-1);

  my $C = MyPlanar::points_move(MyPlanar::points_rotate_plus90($points), 1,1);
  my $D = MyPlanar::points_move($points, 1,1);

  say "A,B  ", defined(MyPlanar::poly_intersection($A,$B)) ? "inter" : "not";
  say "C,D  ", defined(MyPlanar::poly_intersection($C,$D)) ? "inter" : "not";
  say '';

  my @AC = MyPlanar::poly_intersection($A,$C);
  ### @AC

  my @AD = MyPlanar::poly_intersection($A,$D);
  ### @AD

  my @BC = MyPlanar::poly_intersection($B,$C);
  ### @BC

  my @BD = MyPlanar::poly_intersection($B,$D);
  ### @BD

  exit 0;
}
{
  # Rsides MyFSM low to high

  require MyFSM;
  my %table;
  foreach my $state (0 .. $#Rpred_state_table) {
    foreach my $digit (0 .. 4) {
      $table{$state}->{$digit} = $Rpred_state_table[$state]->[$digit];
    }
  }
  my $fsm = MyFSM->new(table => \%table,
                       initials => { 0 => 'start' },
                       accepting => { 0 => '1side',
                                      1 => '3side',
                                      2 => '3side',
                                      3 => '3side',
                                      4 => '4side' },
                       title => 'Rsides',
                       flow => 'west',
                      );
  $fsm->simplify (verbose => 1);
  $fsm->show;
  $fsm->view;
  print "\n";

  print "reversal\n";
  $fsm = $fsm->reverse;
  $fsm->renumber_states;
  $fsm->show;
  $fsm->simplify (verbose => 1);
  $fsm->renumber_states;
  $fsm->show;
  $fsm->view;
  exit 0;
}
{
  # Rsides flat low to high
  my $f = Rpred_flat()->clone;
  $f->unset_accepting($f->get_states);
  # $f->set_accepting(1,2);   # X,Y = 3 sides
  $f->set_accepting(0);   # R = 1 side
  MyFLAT::view($f);
  $f = $f->MyFLAT::reverse;
  $f = $f->MyFLAT::minimize;
  MyFLAT::view($f);
  exit 0;
}
{
  # QAdiampart_Graph()
  my $n = 15;
  my $graph = MyR5Dragon::QAdiampart_Graph($n);
  my $root = $graph->get_graph_attribute('root_vertex');
  $graph->set_graph_attribute('vertex_name_type','');
  my $num_vertices = $graph->vertices;
  print "n=$n  num_vertices $num_vertices\n";
  # MyGraphs::Graph_view($graph);

  my @vpar1
    = (0,
       0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 2, 2, 20, 20,
       22, 20, 23, 5, 26, 26, 28, 26, 29, 5, 8, 8, 11, 11, 45, 39, 46,
       46, 46, 40, 58, 55, 40, 40, 36, 44, 35, 45, 45, 37, 49, 50, 50,
       50, 38, 54, 55, 55, 61, 61, 61, 63, 61, 64, 14, 76, 70, 77, 77,
       77, 71, 89, 86, 71, 71, 67, 75, 66, 76, 76, 68, 80, 81, 81, 81,
       69, 85, 86, 86, 92, 92, 92, 94, 92, 95, 14
      );
  my $g2 = Graph->new (undirected=>1);
  foreach my $i (1 .. $#vpar1) {
    my $p = $vpar1[$i] || next;
    $g2->add_edge($i, $p);
  }
  my $isomorphic = MyGraphs::Graph_is_isomorphic($graph, $g2);
  $num_vertices = $g2->vertices;
  print "g2  num_vertices $num_vertices\n";
  ### $isomorphic

  my $s1 = MyGraphs::graph6_str_to_canonical(MyGraphs::Graph_to_sparse6_str($graph));
  my $s2 = MyGraphs::graph6_str_to_canonical(MyGraphs::Graph_to_sparse6_str($g2));
  print "s1\n";
  print "s2\n";

  $graph = MyGraphs::Graph_from_graph6_str($s1);
  $g2    = MyGraphs::Graph_from_graph6_str($s2);
  $graph->set_graph_attribute (name => 'g1');
  $g2   ->set_graph_attribute (name => 'g2');

  MyGraphs::Graph_view($graph);
  MyGraphs::Graph_view($g2);

  print join(' ',@{graph6_str_to_parent_aref($s1)}),"\n";
  print join(' ',@{graph6_str_to_parent_aref($s2)}),"\n";
  exit 0;

  sub graph6_str_to_parent_aref {
    my ($str) = @_;
    MyGraphs::edge_aref_to_parent_aref(graph6_str_to_edge_aref($str));
  }
  sub graph6_str_to_edge_aref {
    my ($str) = @_;
    Graph::Graph6::read_graph(str => $str,
                              edge_aref => \my @edges);
    return \@edges;
  }
}

{
  # left boundary squares num sides
  # R -> U R R
  # U -> U U R
  # R=1 side, U=3 sides
  # 

  foreach my $n (0 .. 40) {
    my $s = 3 - TernaryLowestNon1($n);
    print $s;
  }
  print "\n";

  require Tie::IxHash;
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $n = 0;
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my %sides;
  tie %sides, 'Tie::IxHash';
  foreach (1 .. 100) {
    my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
    my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
    my $xopp = $x + $dx + $lx;
    my $yopp = $y + $dy + $ly;
    my ($x1,$y1, $x2,$y2) = rect_corners_sort ($x,$y, $xopp,$yopp);
    $sides{"$x1,$y1"}++;

    $n++;
    $x += $dx;
    $y += $dy;
    ### at: "n=$n $x,$y dxdy $dx,$dy right $rx,$ry"
    my $turn;
    if (defined $path->xyxy_to_n_either($x,$y, $x+$lx,$y+$ly)) {
      ($dx,$dy) = ($lx,$ly);
    } elsif (defined $path->xyxy_to_n_either($x,$y, $x+$rx,$y+$ry)) {
      ($dx,$dy) = ($rx,$ry);
    } else {
      die;
    }
  }
  print values %sides, "\n";

  my $str = 'U';
  foreach (1 .. 3) {
    $str =~ s{([RU])}{$1 eq 'R' ? 'URR' : 'UUR'}ge;
  }
  print "$str\n";
  $str =~ s{([RU])}{$1 eq 'R' ? '1' : '3'}ge;
  print "$str\n";

  exit 0;
}

{
  # right boundary squares num sides
  # R -> R R U
  # U -> R U U
  # R=1 side, U=3 sides

  require Tie::IxHash;
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $n = 0;
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  my %sides;
  tie %sides, 'Tie::IxHash';
  foreach (1 .. 100) {
    my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
    my $xopp = $x + $dx + $rx;
    my $yopp = $y + $dy + $ry;
    my ($x1,$y1, $x2,$y2) = rect_corners_sort ($x,$y, $xopp,$yopp);
    $sides{"$x1,$y1"}++;

    my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
    $n++;
    $x += $dx;
    $y += $dy;
    ### at: "n=$n $x,$y dxdy $dx,$dy right $rx,$ry"
    my $turn;
    if (defined $path->xyxy_to_n_either($x,$y, $x+$rx,$y+$ry)) {
      ($dx,$dy) = ($rx,$ry);
    } elsif (defined $path->xyxy_to_n_either($x,$y, $x+$lx,$y+$ly)) {
      ($dx,$dy) = ($lx,$ly);
    } else {
      die;
    }
  }
  print values %sides, "\n";

  my $str = 'R';
  foreach (1 .. 4) {
    $str =~ s{([RU])}{$1 eq 'R' ? 'RRU' : 'RUU'}ge;
  }
  $str =~ s{([RU])}{$1 eq 'R' ? '1' : '3'}ge;
  $str = uc($str);
  print "$str\n";

  foreach my $n (0 .. 40) {
    my $s = TernaryLowestNon1($n) + 1;
    print $s;
  }
  print "\n";

  exit 0;

  sub TernaryLowestNon1 {
    my ($n) = @_;
    while ($n % 3 == 1) { $n = int($n/3); }
    return $n % 3;
  }
}

{
  # boundary turn sequence
  # R -> R l R l U
  # U -> R l U l U
  # bottom R = empty, U = r r

  # not in OEIS: 1,1,-1,-1,1,1,1,-1,-1,1,1,-1,-1,1,-1,-1,1,1,1,-1,-1

  my $path = Math::PlanePath::R5DragonCurve->new;
  my $n = 0;
  my ($x,$y) = $path->n_to_xy($n);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  foreach (1 .. 70) {
    my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
    my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
    $n++;
    $x += $dx;
    $y += $dy;
    ### at: "n=$n $x,$y dxdy $dx,$dy right $rx,$ry"
    my $turn;
    if (defined $path->xyxy_to_n_either($x,$y, $x+$rx,$y+$ry)) {
      $turn = 'R';
      ($dx,$dy) = ($rx,$ry);
    } elsif (defined $path->xyxy_to_n_either($x,$y, $x+$lx,$y+$ly)) {
      $turn = 'L';
      ($dx,$dy) = ($lx,$ly);
    } else {
      die;
    }
    # print "n=$n  $turn\n";
    print "$turn";
  }
  print "\n";

  # R -> R l R l U
  # U -> R l U l U
  # bottom R = empty, U = r r
  my $str = 'R';
  foreach (1 .. 4) {
    $str =~ s{([RU])}{$1 eq 'R' ? 'RlRlU' : 'RlUlU'}ge;
  }
  $str =~ s{([RU])}{$1 eq 'R' ? '' : 'rr'}ge;
  $str = uc($str);
  print "$str\n";

  $str =~ s{([RL])}{$1 eq 'R' ? '-1,' : '1,'}ge;
  print "$str\n";
  exit 0;
}
{
  # turn preds view
  # MyR5Dragon::TurnLpred_flat->MyFLAT::view;
  # MyR5Dragon::TurnLLpred_flat->MyFLAT::view;
  # MyR5Dragon::TurnLLLpred_flat->MyFLAT::view;
  MyR5Dragon::TurnRRRpred_flat->MyFLAT::view;
  exit 0;
}
{
  # HQ by pred
  my $f = MyR5Dragon::HQpred_flat();
  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $k (3) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my $n5 = sprintf '%0*s', $k, cnv($n,10,5);
      if ($f->contains($n5)) {
        print "$n $n5\n";
      }
    }
  }
  {
    my $max_k = 10;
    print join(',', MyFLAT::FLAT_count_contains($f,$max_k)),"\n";
  }
  exit 0;
}

{
  # area tree
  #   k=0 single vertex
  #   k=1 5-star
  #   k=2 https://hog.grinvin.org/ViewGraphInfo.action?id=26973
  #   k=3 https://hog.grinvin.org/ViewGraphInfo.action?id=27759
  #
  # turns graph
  #   k=0 empty
  #   k=1 4-cycle
  #   k=2 https://hog.grinvin.org/ViewGraphInfo.action?id=26975
  #   k=3 https://hog.grinvin.org/ViewGraphInfo.action?id=27761

  require MyGraphs;
  my @graphs;
  foreach my $k (1) {
    my $graph = r5quad_turn_Graph($k);
    # my $graph = r5quad_area_tree_Graph($k);
    MyGraphs::Graph_view($graph);
    MyGraphs::Graph_xy_print($graph);
    push @graphs, $graph;
  }
  MyGraphs::hog_searches_html(@graphs);
  exit 0;

}
{
  # R5 quad area tree domsets
  # not in OEIS: 1,17,2025505,50074443104657772268669935200625
  # not in OEIS: 1,2,318,31154264830044

  # vector(4,k,k--; vpar_domsets_count(QA_vpar(k)))

  require MyGraphs;
  my @domsets_count;
  my @minimal_domsets_count;
  foreach my $k (0 .. 4) {
    my $graph = r5quad_area_tree_Graph($k);
    my $domsets_count         = MyGraphs::Graph_tree_domsets_count($graph);
    my $minimal_domsets_count = MyGraphs::Graph_tree_minimal_domsets_count($graph);
    print "k=$k  $domsets_count\n";
    print "     $minimal_domsets_count\n";

    push @domsets_count, $domsets_count;
    push @minimal_domsets_count, $minimal_domsets_count;
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(name => 'domsets_count',
                           array => \@domsets_count, verbose=>1);
  Math::OEIS::Grep->search(name => 'minimal_domsets_count',
                           array => \@minimal_domsets_count, verbose=>1);
  exit 0;
}
{
  # number of independent sets, by tree depth first
  # want 2, 17, 540497

  # w,z,a
  # vw = z^3
  # vz = (w+z)^3 = a^3
  #
  # pw = vz    = a^3
  # pz = vz+vw = a^3 + z^3
  #
  # cw = pz    =   a^3 + z^3
  # cz = pz+pw = 2*a^3 + z^3

  # vw = (2*a^3 + z^3)^3
  # vz = (3*a^3 + z^3)^3
  #
  # pw = vz    = (3*a^3 + z^3)^3
  # pz = vz+vw = (3*a^3 + z^3)^3 + (2*a^3 + z^3)^3

  require MyGraphs;
  my %with;
  my %without;
  my @ww_at_depth;
  my $k = 4;
  my $graph = r5quad_area_tree_Graph($k-1);
  my $root = MyGraphs::Graph_tree_root($graph);
  Graph_traverse_depth_first
    ($graph, subr => sub {
       my ($parent, $v, $depth) = @_;

       my $w = ($with{$v}    //= bigint_class()->new(1));
       my $z = ($without{$v} //= bigint_class()->new(1));

       $ww_at_depth[$depth]->{"$w,$z"}++;

       if (defined $parent) {
         $with{$parent}    //= bigint_class()->new(1);
         $without{$parent} //= bigint_class()->new(1);
         $with{$parent}    *= $z;
         $without{$parent} *= ($z + $w);
       }

       # print "$v at $depth (parent ",$parent//'undef',")  $w, $z\n";
     });

  my $total = $with{$root} + $without{$root};
  print "root $with{$root} + $without{$root} = $total\n";

  {
    foreach my $depth (0 .. $#ww_at_depth) {
      my $href = $ww_at_depth[$depth];
      my $count_types = scalar keys %$href;
      print "depth $depth  $count_types\n";
    }
  }
  if (1) {
    my $pair_to_psum = sub {
      my ($pair) = @_;
      return sum(split /,/,$pair);
    };
    foreach my $depth (0 .. $#ww_at_depth) {
      print "depth $depth\n";
      my $href = $ww_at_depth[$depth];
      foreach my $pair (sort { $pair_to_psum->($a) <=> $pair_to_psum->($b) }
                        keys %$href) {
        my $count = $href->{$pair};
        my $psum = sum(map {bigint_class()->new($_)} split /,/,$pair);
        my $product = $psum * $count;
        print "  $count * $pair  $psum  = $product\n";
      }
    }
  }

  sub Graph_traverse_depth_first {
    my ($graph, %options) = @_;
    my $root = $options{'root'} // MyGraphs::Graph_tree_root($graph);
    my $subr = $options{'subr'} // sub {};
    my %seen;
    my @stack = [undef, $root, 0];
    while (@stack) {
      ### @stack
      my $elem  = pop @stack;
      my ($parent,$v,$depth) = @$elem;
      $seen{$v} = 1;
      my @children = grep {!$seen{$_}} $graph->neighbours($v);
      ### $v
      ### $depth
      ### @children
      if (@children) {
        push @stack, $elem, map {[$v,$_,$depth+1]} @children;
      } else {
        $subr->($parent, $v, $depth);
      }
    }
  }
  exit 0;
}
{
  # number of independent sets, by tree    -- WRONG
  # want 2, 17, 540497

  # should be
  #     P         Pwith    = Vwithout
  #     |         Pwithout = Vwith + Vwithout
  #     V
  #   / | \       Vwith    = prod(Cwithout)
  #  C  C  C      Vwithout = prod(Cwith + Cwithout)
  #                        = prod(Cany)
  #
  # Pwith    = prod(Cwith + Cwithout)
  #          = prod(Cany)
  #          = C1any^2 * C2any
  # Pwithout = prod(Cwith + Cwithout) + prod(Cwithout)
  # Pwithout = prod(Cany)             + prod(Cwithout)
  #          = C1any^2 * C2any + C1without^2*C2without

  # forbiddens
  # xVwith    = Cwith*Cwithout^2 + Cwith^2*Cwithout + Cwith^3     + xCany
  #           = prod(Cany) - prod(Cwithout)
  # xVwithout = 0                                                 + xCany
  #
  # total xVwith + xVwithout + Vwith + Vwithout == 2*prod(Cany)

  # vector(10,n,n--; 3^CountTernaryTwos(n))
  # vector(10,n,n--; fibonacci(n+3))
  # vector(10,n,n--; 3^CountTernaryTwos(n)*fibonacci(n+3))

  # vector(10,n,n+=2; 3^CountTernaryTwos(n)*fibonacci(n))
  # not in OEIS: 2, 3, 15, 24, 39, 189, 34, 55, 267, 144
  # vector(10,n,n--; 3^CountTernaryTwos(n)*fibonacci(n+3))
  # not in OEIS: 2, 3, 15, 8, 13, 63, 102, 165, 801, 144

  my %count_with;
  my %count_without;
  my $k = 2;
  my $graph = r5quad_area_tree_Graph($k-1);
  my $num_vertices = $graph->vertices;
  print "k=$k  num_vertices $num_vertices\n";
  my %seen;
  my $root = $graph->get_graph_attribute('root_vertex');
  my @pending = ($root);
  $count_with{$root} = 1;
  $count_without{$root} = 1;
  my $depth = 0;
  my @any_at_depth;
  my @totals_by_depth;
  my @totals_by_level;
  while (@pending) {
    my $total_with = 0;
    my $total_without = 0;
    my %values_seen;
    my @new_pending;

    foreach my $v (@pending) {
      $seen{$v} = 1;
      my $v_with = $count_with{$v};
      my $v_without = $count_without{$v};

      $values_seen{"$v_with,$v_without"}++;
      # $values_with{$v_without}++;
      # $values_any{$v_with + $v_without}++;

      $total_with += $v_with;
      $total_without += $v_without;
      foreach my $c ($graph->neighbours($v)) {
        next if $seen{$c};
        $count_with{$c}    =           $v_without;
        $count_without{$c} = $v_with + $v_without;
        push @new_pending, $c;
      }
    }
    my $total = $total_with + $total_without;

    my $width = scalar(@pending);
    print "depth=$depth  width $width  indsets $total_with + $total_without = $total\n";

    my $pair_to_psum = sub {
      my ($pair) = @_;
      return sum(split /,/,$pair);
    };
    foreach my $pair (sort { $pair_to_psum->($a) <=> $pair_to_psum->($b) }
                      keys %values_seen) {
      my $count = $values_seen{$pair};
      my $psum = sum(split /,/,$pair);
      my $product = $psum * $count;
      print "  $count * $pair  $psum  = $product\n";
    }

    push @totals_by_depth, $total;

    $depth++;
    @pending = @new_pending;
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(name => 'totals_by_depth',
                           array => \@totals_by_depth, verbose=>1);
  exit 0;
}

{
  # 2 middle  2 * 4 each = 8
  # 1 middle  4 * 2 each = 8
  # 0 middle  1
  require Image::Base::Text;
  my @vertices;
  my $graph = r5quad_turn_Graph(2, vertex_sequence=>\@vertices);
  print "$graph\n";

  my %C = (0 => 1,
           9 => 1,
           15 => 1,
           21 => 1);
  my @points = map {[split /,/]} @vertices;
  my $x_min = min(map{$_->[0]} @points);
  my $y_min = min(map{$_->[1]} @points);
  my $x_max = max(map{$_->[0]} @points);
  my $y_max = max(map{$_->[1]} @points);
  my $width = $x_max - $x_min + 1;
  my $height = $y_max - $y_min + 1;
  my @C_histogram = (0,0,0,0,0);
  foreach my $str ('111111111111111111111111',

                   '010101010101100101010101',
                   '010101010101101010010101',
                   '101001010101100101010101',
                   '101001010101101010010101',

                   '101010010101010101100101',
                   '101010010101010101101010',
                   '101010010101100101010101',
                   '101010010101101010010101',
                   '101010100101010101010101',
                   '101010100101010101100101',
                   '101010100101010101101010',
                   '101010101010010101010101',
                   '101010101010010101100101',
                   '101010101010010101101010',

                   # 3 outsides
                   '101010010101010101010101',

                   # 1 middle NE
                   '010101010101010101010101',
                   '101001010101010101010101',
                  ) {
    my $image = Image::Base::Text->new (-width => $width, -height => $height);
    my $count_C = 0;
    foreach my $i (0 .. length($str)-1) {
      my ($x,$y) = @{$points[$i]};
      $x -= $x_min;
      $y -= $y_min;
      $y = $height-1-$y;
      my $degree = $graph->vertex_degree($vertices[$i]);
      my $char;
      if (substr($str,$i,1)) {
        if ($C{$i}) {
          $char = 'C';
          $count_C++;
        } else {
          $char = '@';
        }
      } else {
        if ($C{$i}) {
          $char = '_';
        } else {
          $char = '.';
        }
      }
      $image->xy($x,$y, $char);
    }
    print "---------\n";
    print "$str   C=$count_C\n";
    $image->save('/dev/stdout');
    $C_histogram[$count_C]++;
  }
  print "\n";
  print join(' ',@C_histogram),"\n";
  exit 0;
}
{
  # Dpred with opposite turns
  # recurrence_guess([0,0,0,8,72,464,2640,14168,73752,377504,1913760])
  # vector(10,k,2*A(k))
  {
    my $max_k = 10;
    my $f = MyR5Dragon::DpredLeft_k_flat()
      ->intersect(MyR5Dragon::DpredRight_k_flat()
                  ->MyFLAT::digits_increment)
      ->MyFLAT::minimize;
    print join(',', MyFLAT::FLAT_count_contains($f,$max_k)),"\n";
  }
  {
    my $max_k = 10;
    my $f = MyR5Dragon::DpredRight_k_flat()
      ->intersect(MyR5Dragon::DpredLeft_k_flat()
                  ->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize;
    print join(',', MyFLAT::FLAT_count_contains($f,$max_k)),"\n";
  }
  exit 0;
}
{
  # pairs of 31 squares
  # 8, 72, 464, 2640
  # k=4 8

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (
                 # 3,
                 0 .. 8,
                ) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my $n_start = $n_lo;
    my $n_end   = $n_hi;
    my $count_pairs = 0;
    my %seen;
    foreach my $n ($n_start .. $n_end-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my ($sx,$sy) = xy_rotate_plus90($dx,$dy);
      my $pair = 0;
      foreach my $side ('L','R') {
        my $xo = $x + $dx + $sx;
        my $yo = $y + $dy + $sy;
        my ($x1,$y1, $x2,$y2) = rect_corners_sort ($x,$y, $xo,$yo);
        ### corner: "n=$n $side  $x1,$y1 to $x2,$y2  (from $x,$y $dx,$dy $sx,$sy)"
          my $type = MyR5Dragon::square_consecutive_type_xy($x1,$y1,
                                                            $n_start,$n_end);
          ### $type
          if ($type =~ /0011/) {
            $pair++;
          }
        if (! $seen{"$x1,$y1,$side"}++) {
        }
        ($sx,$sy) = (-$sx,-$sy);
      }
      if ($pair == 2) {
        ### both at: "n=$n"
        $count_pairs++;
      }
    }
    print "k=$k $count_pairs\n";
    push @values, $count_pairs;
  }
  if (@values > 4) {
    require Math::OEIS::Grep;
    Math::OEIS::Grep->search(array => \@values, verbose=>1);
  }
  exit 0;
}

{
  # R5 quad area tree Wiener among degree-4s
  foreach my $k (1 .. 5) {
    my $graph = r5quad_area_tree_Graph($k-1);
    print "k=$k  ",Graph_degree4_Wiener_index($graph),"\n";
  }
  exit;

  sub Graph_degree4_Wiener_index {
    my ($graph) = @_;
    my $total = 0;
    my $for = $graph->for_shortest_paths
      (sub {
         my ($t, $u,$v, $n) = @_;
         ### u: $graph->vertex_degree($u)
         ### v: $graph->vertex_degree($v)
         if ($graph->vertex_degree($u) == 4 && $graph->vertex_degree($v) == 4) {
           $total += $t->path_length($u,$v);
         }
       });
    return $total/2;
  }
}

{
  require Graph;
  require MyGraphs;
  my $s = 3;
  my $graph = Graph->new (undirected => 1);
  $graph->add_vertex(0);
  foreach my $k (1 .. 4) {
    $graph = MyGraphs::Graph_star_replacement($graph,$s);
    MyGraphs::Graph_view($graph);
  }
  exit;
}
{
  # r5quad turn graph diameter
  #          0 1  2  3  4  5  6
  # diameter   2,10,30,
  # edges 0,4,28,148,748,3748

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (0 .. 5) {
    my $graph = r5quad_turn_Graph($k);
    my $num_edges = $graph->edges;
    my $diameter = 0; # $graph->diameter || 0;
    print "k=$k edges=$num_edges  diameter=$diameter\n";
    push @values, $diameter;
  }
  shift @values;
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}
{
  # area tree degree counts
  #
  require MyGraphs;
  my @values;
  foreach my $k (0 .. 4) {
    my $graph = r5quad_area_tree_Graph($k);
    my @degrees = (0,0,0,0,0);
    foreach my $v ($graph->vertices) {
      $degrees[$graph->vertex_degree($v)]++;
    }
    print "k=$k ",join(', ',@degrees),"\n";
    push @values, $degrees[4];
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}
{
  # r5dragon graph diameter
  #          0 1 2 3  4  5  6
  # diameter   1,5,11,25,53
  # probably not:
  # A208739 1,2, 5,11,25,53,113,234,482,982,      2^n-partitions
  # A208739(k) = 2^k - numbpart(k);
  # vector(10,k,k--; A208739(k))

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (0..5) {
    my $graph = r5dragon_Graph($k);
    my $apsp = $graph->APSP_Floyd_Warshall;
    my $diameter      = $apsp->diameter;
    my $radius        = $apsp->diameter;
    my $shortest_path = $apsp->shortest_path;
    my $average_path  = $apsp->average_path_length;
    my $longest_path  = $apsp->longest_path;
    my @longest_path  = $apsp->longest_path;
    my @centre        = $apsp->center_vertices;
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
    print "k=$k  end $x_hi,$y_hi\n";
    print "  diameter=$diameter radius=$radius shortest=$shortest_path  avg=$average_path\n";
    print "  centre  ",join(' ',@centre),"\n";
    print "  longest=$longest_path [$longest_path[0] to $longest_path[-1]]\n";
    print "  longest ",join(' ',@longest_path),"\n";
    push @values, $diameter;
  }
  shift @values;
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}

{
  # r5quad convex hull samples
  my $zero = bigint_class()->new(0);
  my $one  = bigint_class()->new(1);
  my $points = [ [$zero,$zero],
                 [$one, $zero],
                 [$one, $one],
                 [$zero,$one] ];
  my $ex = $one;
  my $ey = $zero;
  ($points, $ex,$ey) = MyR5Dragon::quadhull_expand($points, $ex,$ey);

  print "% GP-DEFINE  QH_generated_samples = {[\n";
  my $max_k = 50;
  foreach my $k (0 .. $max_k) {
    my $comma = ($k==$max_k ? '' : ',');
    print "% GP-DEFINE    [ ",points_str_gp($points),"]$comma  /* k=$k */\n";
    ($points, $ex,$ey) = MyR5Dragon::quadhull_expand($points, $ex,$ey);
  }
  print "% GP-DEFINE  ]};\n";
  exit 0;
}
{
  # r5quad convex hull area increment pictures

  my $k = 5;
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $zero = bigint_class()->new(0);
  my $one  = bigint_class()->new(1);
  my $points = [ [0, 0],
                 [1, 0],
                 [1, 1],
                 [0, 1] ];
  my $ex = 1;
  my $ey = 0;

  foreach $k (0 .. $k) {
    ($points, $ex,$ey) = MyR5Dragon::quadhull_expand($points, $ex,$ey);
  }
  print "  % k = $k\n";
  print "  \\draw ";
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    ($x,$y) = mul($x,$y);
    print "($x,$y)--";
  }
  print "cycle\n  ";

  {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my $join = '';
    # foreach my $n (max(0,$pn-2) .. $pn+2) {
    foreach my $n ($n_lo .. $n_hi) {
      last unless $n <= $n_hi;
      my ($x,$y) = $path->n_to_xy($n);
      ($x,$y) = mul($x,$y);
      print "$join($x,$y)";
      $join = '--';
    }
    print "\n  ";
  }
  # foreach my $p (@$points) {
  #   my ($x,$y) = @$p;
  #   my $pn = $path->xy_to_n($x,$y) // next;
  # }


  ($points, $ex,$ey) = MyR5Dragon::quadhull_expand($points, $ex,$ey);
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    print "($x,$y)--";
  }
  print "cycle\n  ";

  {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k+1);
    my $join = '';
    # foreach my $n (max(0,$pn-2) .. $pn+2) {
    foreach my $n ($n_lo .. $n_hi) {
      last unless $n <= $n_hi;
      my ($x,$y) = $path->n_to_xy($n);
      print "$join($x,$y)";
      $join = '--';
    }
    print "\n  ";
  }
  print ";\n";
  exit 0;
}

{
  # postscript r5quad boundary

  my $mm_to_bp = 72/25.4;
  my $bp_to_mm = 1/$mm_to_bp;

  # A4 = 595, 841
  my $width  = 211 * $mm_to_bp;
  my $height = 297 * $mm_to_bp;
  my $border = .1;
  my $chamfer = .2;

  my $level = 3;
  if (@ARGV) { $level = shift @ARGV; }
  my $path = Math::PlanePath::R5DragonCurve->new;


  my $n_lo = 4*5**$level;
  my $n_hi = 8*5**$level;
  my ($x_lo,$y_lo) = $path->n_to_xy($n_lo);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

  my $z_hi = Math::Complex->new($x_hi,$y_hi);
  my $transform_factor = ~$z_hi * Math::Complex->new(0,1);
  my $transform_to_vertical = sub {
    my ($x,$y) = @_;
    my $z = Math::Complex->new($x,$y);
    $z *= $transform_factor;
    return ($z->Re, $z->Im);
  };

  my @points = ([$x_lo,$y_lo]);
  {
    my ($x,$y) = ($x_lo,$y_lo);
    my ($dx,$dy) = $path->n_to_dxdy($n_lo);
    my $n = $n_lo;
    do {
      ### at: "$x,$y  dxdy=$dx,$dy"
      $x += $dx;
      $y += $dy;
      push @points, [$x,$y];

      my @n_list = $path->xy_to_n_list($x,$y) or die;
      @n_list = grep {$_<=$n_hi} @n_list;

      if (@n_list==2 && ($n_list[0] < $n || turn($n_list[0])==1)) {
        ### double left, shift away: join(', ',@n_list)
        shift @n_list;
      }
      $n = $n_list[0];
      ($dx,$dy) = $path->n_to_dxdy($n_list[0]);

    } until ($x==$x_hi && $y==$y_hi);
  }

  @points = map {[ $transform_to_vertical->(@$_) ]} @points;
  ### @points

  my $x_min = min(map {$_->[0]} @points);
  my $x_max = max(map {$_->[0]} @points);
  my $y_min = min(map {$_->[1]} @points);
  my $y_max = max(map {$_->[1]} @points);

  my $x_size = $x_max - $x_min;
  my $y_size = $y_max - $y_min;
  my $length = hypot($x_hi,$y_hi);
  ### $x_size
  ### $y_size
  ### $length

  my $scale = min($width  / ($x_size||$y_size),
                  $height / ($y_size||$x_size));
  $scale *= (1-$border);
  my $segment_mm = $scale * $bp_to_mm;

  my $length_scaled = $length * $scale;
  my $length_mm  = $length_scaled * $bp_to_mm;
  print "X $x_min to $x_max is $x_size   Y $y_min to $y_max is $y_size\n";
  print "end length $length_scaled bp = $length_mm mm\n";
  print "segment $scale bp   $segment_mm mm\n";

  my $line_width = $scale*.1;

  my $x_middle = ($x_min + $x_max)/2;
  my $y_middle = ($y_min + $y_max)/2;
  print "curve middle $x_middle, $y_middle bp\n";

  my $mx = $width /2;
  my $my = $height/2;
  print "paper middle $mx $my bp\n";

  my $ox = $mx - $x_middle * $scale;
  my $oy = $my - $y_middle * $scale;
  print "paper origin $ox $oy bp\n";

  my $transform = sub {
    my ($x,$y) = @_;
    $x *= $scale;
    $y *= $scale;
    $x += $ox;
    $y += $oy;
    ### point: "$x, $y"
    return ($x,$y);
  };
  {
    my ($x,$y) = $transform->($x_middle,$y_middle);
    print "xy middle $x_middle, $y_middle  transformed $x, $y\n";
  }
  {
    my ($x1,$y1) = $transform->(0,0);
    my ($x2,$y2) = $transform->(1,0);
    my $seglen = hypot($x1-$x2,$y1-$y2);
    my $seglen_mm = $seglen * $bp_to_mm;
    print "transformed seglen $seglen bp = $seglen_mm mm\n";
  }
  {
    my ($x1,$y1) = $transform->(0,0);
    my ($x2,$y2) = $transform->($x_hi,$y_hi);
    my $endlen = hypot($x1-$x2,$y1-$y2);
    my $endlen_mm = $endlen * $bp_to_mm;
    print "transformed endlen $endlen bp = $endlen_mm mm\n";
  }

  my $ps_filename = '/tmp/x.ps';
  my $pdf_filename = '/tmp/x.pdf';

  if (1) {
    print "output\n";
    open my $fh, '>', $ps_filename or die;

    print $fh <<"HERE";
%!PS-Adobe-3.0
%%Title: R5 Dragon $level
%%LanguageLevel: 1
%%Creator: r5.pl
%%DocumentMedia: A4 595.27559 841.88976 0 ( ) ( )
%%Orientation: Portrait
%%Pages: 1
%%EndComments
%%BeginProlog
%%EndProlog
%%BeginSetup
<< /PageSize [ 595.27559 841.88976 ] /ImagingBBox null >> setpagedevice
%%EndSetup
%%Page: 1 1
%%BeginPageSetup
/pagelevel save def
%%EndPageSetup
$line_width setlinewidth
newpath
HERE

    {
      my ($x,$y) = @{$points[0]};
      ($x,$y) = $transform->($x,$y);
      print $fh "$x $y moveto\n";
    }
    foreach my $i (1 .. $#points) {
      my ($x,$y) = @{$points[$i]};
      ($x,$y) = $transform->($x,$y);
      print $fh "$x $y lineto\n";
    }

    print $fh <<"HERE";
stroke
%%PageTrailer
pagelevel restore
showpage
%%EOF
HERE
  }

  if (0) {
    require PostScript::Simple;
    my $ps = PostScript::Simple->new(papersize => "A4",
                                     eps => 0,
                                     units => "bp");
    $ps->newpage;
    $ps->setlinewidth($line_width);

    {
      my ($x1,$y1) = @{$points[0]};
      my ($x2,$y2) = @{$points[1]};
      ($x1,$y1) = $transform->($x1,$y1);
      ($x2,$y2) = $transform->($x2,$y2);
      $ps->line($x1,$y1, $x2,$y2);
    }
    foreach my $i (2 .. $#points) {
      my ($x,$y) = @{$points[$i]};
      ($x,$y) = $transform->($x,$y);
      $ps->linextend($x,$y);
    }
    $ps->output($ps_filename);
  }

  require IPC::Run;
  print "ps2pdf\n";
  IPC::Run::run(['ps2pdf12',$ps_filename,$pdf_filename]);

  print "qpdf\n";
  IPC::Run::run(['qpdf','--stream-data=uncompress',
                 $pdf_filename,
                 '/tmp/uncompressed.pdf.txt']);

  # push @ipc, IPC::Run::start(['gv','-presentation','-f',$ps_filename],'&');
  print "xpdf\n";
  push @ipc, IPC::Run::start(['xpdf',
                              '-z','50',
                              $pdf_filename], '2>',File::Spec->devnull, '&');

  IPC::Run::run(['pdfinfo',$pdf_filename],
                '>', \my $pdfinfo_out) or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;

  exit 0;
}

{
  # postscript R5 boundary

  my $mm_to_bp = 72/25.4;
  my $bp_to_mm = 1/$mm_to_bp;

  # A4 = 595, 841
  my $width  = 211 * $mm_to_bp;
  my $height = 297 * $mm_to_bp;
  my $border = .1;
  my $chamfer = .2;

  my $level = 3;
  if (@ARGV) { $level = shift @ARGV; }
  my $path = Math::PlanePath::R5DragonCurve->new;


  my ($n_lo, $n_hi) = $path->level_to_n_range($level);
  my ($x_lo,$y_lo) = $path->n_to_xy($n_lo);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

  my $z_hi = Math::Complex->new($x_hi,$y_hi);
  my $transform_factor = ~$z_hi * Math::Complex->new(0,1);
  my $transform_to_vertical = sub {
    my ($x,$y) = @_;
    my $z = Math::Complex->new($x,$y);
    $z *= $transform_factor;
    return ($z->Re, $z->Im);
  };

  my @points = ([$x_lo,$y_lo]);
  {
    my ($x,$y) = ($x_lo,$y_lo);
    my ($dx,$dy) = $path->n_to_dxdy($n_lo);
    my $n = $n_lo;
    do {
      ### at: "$x,$y  dxdy=$dx,$dy"
      $x += $dx;
      $y += $dy;
      push @points, [$x,$y];

      my @n_list = $path->xy_to_n_list($x,$y) or die;
      @n_list = grep {$_<=$n_hi} @n_list;

      if (@n_list==2 && ($n_list[0] < $n || turn($n_list[0])==1)) {
        ### double left, shift away: join(', ',@n_list)
        shift @n_list;
      }
      $n = $n_list[0];
      ($dx,$dy) = $path->n_to_dxdy($n_list[0]);

    } until ($x==$x_hi && $y==$y_hi);
    @points = reverse @points;
  }
  {
    my ($x,$y) = ($x_lo,$y_lo);
    my ($dx,$dy) = $path->n_to_dxdy($n_lo);
    my $n = $n_lo;
    do {
      ### at: "$x,$y  dxdy=$dx,$dy"
      $x += $dx;
      $y += $dy;
      push @points, [$x,$y];

      my @n_list = $path->xy_to_n_list($x,$y) or die;
      @n_list = grep {$_<=$n_hi} @n_list;

      if (@n_list==2 && ($n_list[0] < $n || turn($n_list[0])==-1)) {
        ### double left, shift away: join(', ',@n_list)
        shift @n_list;
      }
      $n = $n_list[0];
      ($dx,$dy) = $path->n_to_dxdy($n_list[0]);

    } until ($x==$x_hi && $y==$y_hi);
  }

  @points = map {[ $transform_to_vertical->(@$_) ]} @points;
  ### @points

  my $x_min = min(map {$_->[0]} @points);
  my $x_max = max(map {$_->[0]} @points);
  my $y_min = min(map {$_->[1]} @points);
  my $y_max = max(map {$_->[1]} @points);

  my $x_size = $x_max - $x_min;
  my $y_size = $y_max - $y_min;
  my $length = hypot($x_hi,$y_hi);
  ### $x_size
  ### $y_size
  ### $length

  my $scale = min($width  / ($x_size||$y_size),
                  $height / ($y_size||$x_size));
  $scale *= (1-$border);
  my $segment_mm = $scale * $bp_to_mm;

  my $length_scaled = $length * $scale;
  my $length_mm  = $length_scaled * $bp_to_mm;
  print "X $x_min to $x_max is $x_size   Y $y_min to $y_max is $y_size\n";
  print "end length $length_scaled bp = $length_mm mm\n";
  print "segment $scale bp   $segment_mm mm\n";

  my $line_width = $scale*.1;

  my $x_middle = ($x_min + $x_max)/2;
  my $y_middle = ($y_min + $y_max)/2;
  print "curve middle $x_middle, $y_middle bp\n";

  my $mx = $width /2;
  my $my = $height/2;
  print "paper middle $mx $my bp\n";

  my $ox = $mx - $x_middle * $scale;
  my $oy = $my - $y_middle * $scale;
  print "paper origin $ox $oy bp\n";

  my $transform = sub {
    my ($x,$y) = @_;
    $x *= $scale;
    $y *= $scale;
    $x += $ox;
    $y += $oy;
    ### point: "$x, $y"
    return ($x,$y);
  };
  {
    my ($x,$y) = $transform->($x_middle,$y_middle);
    print "xy middle $x_middle, $y_middle  transformed $x, $y\n";
  }
  {
    my ($x1,$y1) = $transform->(0,0);
    my ($x2,$y2) = $transform->(1,0);
    my $seglen = hypot($x1-$x2,$y1-$y2);
    my $seglen_mm = $seglen * $bp_to_mm;
    print "transformed seglen $seglen bp = $seglen_mm mm\n";
  }
  {
    my ($x1,$y1) = $transform->(0,0);
    my ($x2,$y2) = $transform->($x_hi,$y_hi);
    my $endlen = hypot($x1-$x2,$y1-$y2);
    my $endlen_mm = $endlen * $bp_to_mm;
    print "transformed endlen $endlen bp = $endlen_mm mm\n";
  }

  my $ps_filename = '/tmp/x.ps';
  my $pdf_filename = '/tmp/x.pdf';

  if (1) {
    print "output\n";
    open my $fh, '>', $ps_filename or die;

    print $fh <<"HERE";
%!PS-Adobe-3.0
%%Title: R5 Dragon $level
%%LanguageLevel: 1
%%Creator: r5.pl
%%DocumentMedia: A4 595.27559 841.88976 0 ( ) ( )
%%Orientation: Portrait
%%Pages: 1
%%EndComments
%%BeginProlog
%%EndProlog
%%BeginSetup
<< /PageSize [ 595.27559 841.88976 ] /ImagingBBox null >> setpagedevice
%%EndSetup
%%Page: 1 1
%%BeginPageSetup
/pagelevel save def
%%EndPageSetup
$line_width setlinewidth
newpath
HERE

    {
      my ($x,$y) = @{$points[0]};
      ($x,$y) = $transform->($x,$y);
      print $fh "$x $y moveto\n";
    }
    foreach my $i (1 .. $#points) {
      my ($x,$y) = @{$points[$i]};
      ($x,$y) = $transform->($x,$y);
      print $fh "$x $y lineto\n";
    }

    print $fh <<"HERE";
stroke
%%PageTrailer
pagelevel restore
showpage
%%EOF
HERE
  }

  if (0) {
    require PostScript::Simple;
    my $ps = PostScript::Simple->new(papersize => "A4",
                                     eps => 0,
                                     units => "bp");
    $ps->newpage;
    $ps->setlinewidth($line_width);

    {
      my ($x1,$y1) = @{$points[0]};
      my ($x2,$y2) = @{$points[1]};
      ($x1,$y1) = $transform->($x1,$y1);
      ($x2,$y2) = $transform->($x2,$y2);
      $ps->line($x1,$y1, $x2,$y2);
    }
    foreach my $i (2 .. $#points) {
      my ($x,$y) = @{$points[$i]};
      ($x,$y) = $transform->($x,$y);
      $ps->linextend($x,$y);
    }
    $ps->output($ps_filename);
  }

  require IPC::Run;
  print "ps2pdf\n";
  IPC::Run::run(['ps2pdf12',$ps_filename,$pdf_filename]);

  print "qpdf\n";
  IPC::Run::run(['qpdf','--stream-data=uncompress',
                 $pdf_filename,
                 '/tmp/uncompressed.pdf.txt']);

  # push @ipc, IPC::Run::start(['gv','-presentation','-f',$ps_filename],'&');
  print "xpdf\n";
  push @ipc, IPC::Run::start(['xpdf',
                              '-z','50',
                              $pdf_filename], '2>',File::Spec->devnull, '&');

  IPC::Run::run(['pdfinfo',$pdf_filename],
                '>', \my $pdfinfo_out) or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;

  exit 0;
}


{
  # postscript chamfered

  # end sqrt(5)^0 = 1 seglen 1
  # end sqrt(5)^1 = 2.236
  # end sqrt(5)^2 = 5 seglen
  # endpoint 200mm/5^3=125 is 1.6mm each
  # endpoint 200/sqrt(6)^7 is 0.378mm each

  my $mm_to_bp = 72/25.4;
  my $bp_to_mm = 1/$mm_to_bp;

  # A4 = 595, 841
  my $width  = 211 * $mm_to_bp;
  my $height = 297 * $mm_to_bp;
  # $width  = 595 * $mm_to_bp;
  # $height = 841 * $mm_to_bp;
  my $border = .1;
  my $chamfer = .2;

  my $level = 3;
  if (@ARGV) { $level = shift @ARGV; }
  my $path = Math::PlanePath::R5DragonCurve->new;


  my ($n_lo, $n_hi) = $path->level_to_n_range($level);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
  my $z_hi = Math::Complex->new($x_hi,$y_hi);
  my $transform_factor = ~$z_hi * Math::Complex->new(0,1);
  my $transform_to_vertical = sub {
    my ($x,$y) = @_;
    my $z = Math::Complex->new($x,$y);
    $z *= $transform_factor;
    return ($z->Re, $z->Im);
  };

  # require Geometry::AffineTransform;
  # my $affine = Geometry::AffineTransform->new;
  # $affine->translate(1/3, 2/3);
  # ($x,$y) = $affine->transform($x,$y);

  my @points;
  {
    my ($x,$y) = $path->n_to_xy($n_lo);
    push @points, [$x,$y];
  }
  foreach my $n ($n_lo+1 .. $n_hi-1) {
    foreach my $c (-$chamfer, $chamfer) {
      my ($x,$y) = $path->n_to_xy($n + $c);
      push @points, [$x,$y];
    }
  }
  {
    my ($x,$y) = $path->n_to_xy($n_hi);
    push @points, [$x,$y];
  }
  @points = map {[ $transform_to_vertical->(@$_) ]} @points;
  ### @points

  my $x_min = min(map {$_->[0]} @points);
  my $x_max = max(map {$_->[0]} @points);
  my $y_min = min(map {$_->[1]} @points);
  my $y_max = max(map {$_->[1]} @points);

  my $x_size = $x_max - $x_min;
  my $y_size = $y_max - $y_min;
  my $length = hypot($x_hi,$y_hi);
  ### $x_size
  ### $y_size
  ### $length

  my $scale = min($width  / ($x_size||$y_size),
                  $height / ($y_size||$x_size));
  $scale *= (1-$border);
  my $segment_mm = $scale * $bp_to_mm;

  my $length_scaled = $length * $scale;
  my $length_mm  = $length_scaled * $bp_to_mm;
  print "X $x_min to $x_max is $x_size   Y $y_min to $y_max is $y_size\n";
  print "end length $length_scaled bp = $length_mm mm\n";
  print "segment $scale bp   $segment_mm mm\n";

  my $line_width = $scale*.1;

  my $x_middle = ($x_min + $x_max)/2;
  my $y_middle = ($y_min + $y_max)/2;
  print "curve middle $x_middle, $y_middle bp\n";

  my $mx = $width /2;
  my $my = $height/2;
  print "paper middle $mx $my bp\n";

  my $ox = $mx - $x_middle * $scale;
  my $oy = $my - $y_middle * $scale;
  print "paper origin $ox $oy bp\n";

  my $transform = sub {
    my ($x,$y) = @_;
    $x *= $scale;
    $y *= $scale;
    $x += $ox;
    $y += $oy;
    ### point: "$x, $y"
    return ($x,$y);
  };
  {
    my ($x,$y) = $transform->($x_middle,$y_middle);
    print "xy middle $x_middle, $y_middle  transformed $x, $y\n";
  }
  {
    my ($x1,$y1) = $transform->(0,0);
    my ($x2,$y2) = $transform->(1,0);
    my $seglen = hypot($x1-$x2,$y1-$y2);
    my $seglen_mm = $seglen * $bp_to_mm;
    print "transformed seglen $seglen bp = $seglen_mm mm\n";
  }
  {
    my ($x1,$y1) = $transform->(0,0);
    my ($x2,$y2) = $transform->($x_hi,$y_hi);
    my $endlen = hypot($x1-$x2,$y1-$y2);
    my $endlen_mm = $endlen * $bp_to_mm;
    print "transformed endlen $endlen bp = $endlen_mm mm\n";
  }

  my $ps_filename = '/tmp/x.ps';
  my $pdf_filename = '/tmp/x.pdf';

  if (1) {
    print "output\n";
    open my $fh, '>', $ps_filename or die;

    print $fh <<"HERE";
%!PS-Adobe-3.0
%%Title: R5 Dragon $level
%%LanguageLevel: 1
%%Creator: r5.pl
%%DocumentMedia: A4 595.27559 841.88976 0 ( ) ( )
%%Orientation: Portrait
%%Pages: 1
%%EndComments
%%BeginProlog
%%EndProlog
%%BeginSetup
<< /PageSize [ 595.27559 841.88976 ] /ImagingBBox null >> setpagedevice
%%EndSetup
%%Page: 1 1
%%BeginPageSetup
/pagelevel save def
%%EndPageSetup
$line_width setlinewidth
newpath
HERE

    {
      my ($x,$y) = @{$points[0]};
      ($x,$y) = $transform->($x,$y);
      print $fh "$x $y moveto\n";
    }
    foreach my $i (1 .. $#points) {
      my ($x,$y) = @{$points[$i]};
      ($x,$y) = $transform->($x,$y);
      print $fh "$x $y lineto\n";
    }

    print $fh <<"HERE";
stroke
%%PageTrailer
pagelevel restore
showpage
%%EOF
HERE
  }

  if (0) {
    require PostScript::Simple;
    my $ps = PostScript::Simple->new(papersize => "A4",
                                     eps => 0,
                                     units => "bp");
    $ps->newpage;
    $ps->setlinewidth($line_width);

    {
      my ($x1,$y1) = @{$points[0]};
      my ($x2,$y2) = @{$points[1]};
      ($x1,$y1) = $transform->($x1,$y1);
      ($x2,$y2) = $transform->($x2,$y2);
      $ps->line($x1,$y1, $x2,$y2);
    }
    foreach my $i (2 .. $#points) {
      my ($x,$y) = @{$points[$i]};
      ($x,$y) = $transform->($x,$y);
      $ps->linextend($x,$y);
    }
    $ps->output($ps_filename);
  }

  require IPC::Run;
  print "ps2pdf\n";
  IPC::Run::run(['ps2pdf12',$ps_filename,$pdf_filename]);

  print "qpdf\n";
  IPC::Run::run(['qpdf','--stream-data=uncompress',
                 $pdf_filename,
                 '/tmp/uncompressed.pdf.txt']);

  # push @ipc, IPC::Run::start(['gv','-presentation','-f',$ps_filename],'&');
  print "xpdf\n";
  push @ipc, IPC::Run::start(['xpdf',
                              '-z','50',
                              $pdf_filename], '2>',File::Spec->devnull, '&');

  IPC::Run::run(['pdfinfo',$pdf_filename],
                '>', \my $pdfinfo_out) or die;
  $pdfinfo_out =~ s/^Jav.*\n//m;
  print $pdfinfo_out;

  exit 0;
}


{
  # r5dragon convex hull samples
  my $points = [ [0,0], [1,0] ];
  my $ex = 1;
  my $ey = 0;

  print "% GP-DEFINE  hull_generated_samples = {[\n";
  my $max_k = 13;
  foreach my $k (0 .. $max_k) {
    my $comma = ($k==$max_k ? '' : ',');
    print "% GP-DEFINE    [ ",points_str_gp($points),"]$comma  /* k=$k */\n";
    ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
  }
  print "% GP-DEFINE  ]};\n";
  exit 0;
}

{
  # r5dragon graph endpoint distance

  # not in OEIS: 1, 5, 11, 25, 53, 115, 265, 573, 1203

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (0..8) {
    my $graph = r5dragon_Graph($k);
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
    my $u = '0,0';
    my $v = "$x_hi,$y_hi";
    $graph->has_vertex($u) or die;
    $graph->has_vertex($v) or die;
    my @path = $graph->SP_Dijkstra($u,$v);
    ### @path
    my $dist = $#path;
    print "$dist,\n";
    push @values, $dist;
  }
  shift @values;
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}




{
  # area tree Wiener index
  #
  require MyGraphs;
  my @W;
  my @Wc;
  foreach my $k (0 .. 4) {
    my $graph = r5quad_area_tree_Graph($k);
    my $num_vertices = $graph->vertices;
    my $diameter = $graph->diameter || 0;
    my @diameter = MyGraphs::Graph_tree_diameter_path($graph);
    my $W = MyGraphs::Graph_Wiener_index($graph);
    my $Wc = MyGraphs::Graph_Wiener_part_at_vertex($graph, $diameter[0]);
    print "k=$k W=$W Wc=$Wc diam=$diameter  vertices $num_vertices\n";
    push @W, $W;
    push @Wc, $Wc;
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@W, verbose=>1);
  Math::OEIS::Grep->search(array => \@Wc, verbose=>1);
  exit 0;

}
{
  # quad by base
  # (1+2*I)^2 == -3+4*I
  # (-1+I)*(1+2*I) == -3 - I
  exit 0;
}
{
  # turn graph when arms off cycle differ

  # k=1       single unit square
  #   0  1
  #
  # star of squares,
  # k=2
  #   0
  #   1  0,4         1-component with 1 distinct, 4 of arms
  #   2
  #   3
  #   4  0,1         4-components with 1 distinct, 1 of centre
  #
  # k=3
  #   0
  #   1  0,12        1-component with 1 distinct, 4 of arms
  #   2  0,0,8       2-components both distinct, 8 of in-between
  #   3
  #   4  0,1,4       4-components 1-distinct, 1 of centre
  #                  4-components 2-distinct, 4 of centre of arm star
  #
  # k=4
  #   0
  #   1  0,52
  #   2  0,0,48
  #   3
  #   4  0,1,16,8

  my $rep = 0;
  require MyGraphs;
  foreach my $k (1 .. 4) {
    print "k=$k\n";
    my $graph = r5quad_turn_Graph($k);
    my @count;
    my @count_distinct;
    MyGraphs::Graph_find_all_4cycles
        ($graph,
         callback => sub {
           my @cycle = @_;
           my %cycle;
           @cycle{@cycle} = ();  # hash slice
           my @neighbours = (map {$graph->neighbours($_)} @cycle);
           my @components = grep {! exists $cycle{$_}} @neighbours;
           # @cycle
           # @neighbours
           # @components
           my $g = $graph->copy;
           $g->delete_vertices(@cycle);
           $rep++;
           Graph_is_different_components($g,@components) or die;
           @components = map {
             $g->has_vertex($_) or die;
             $g->subgraph ([$g->all_reachable($_)])
           } @components;
           # my @component_num_vertices = map {scalar $_->vertices} @components;
           # # @component_num_vertices

           my $num_components = scalar @components;
           $count[$num_components]++;
           my @distinct_components = @components;
         OUTER: for (my $i = 1; $i <= $#distinct_components; ) {
             foreach my $j (0 .. $i-1) {
               if (MyGraphs::Graph_is_isomorphic($distinct_components[$i],
                                                 $distinct_components[$j])) {
                 # same, delete
                 # diag "same $j $i";
                 splice @distinct_components, $i,1;
                 next OUTER;
               }
             }
             # all different, look at next
             $i++;
           }
           my $num_distinct_components = scalar @distinct_components;
           $count_distinct[$num_components]->[$num_distinct_components]++;
           # diag "$num_components (",join(' ',@components),
           #   ") distinct  $num_distinct_components (",
           #   join(' ',@distinct_components), ")";
           # if ($rep == 2) {
           #   MyGraphs::Graph_view($g);
           #   exit;
           # }
           # diag "$num_components distinct $num_distinct_components";
           # if ($num_components==4 && $num_distinct_components==3) {
           #   MyGraphs::Graph_view($components[0]);
           #   MyGraphs::Graph_view($components[1]);
           #   MyGraphs::Graph_view($components[2]);
           # }
         });
    # @count = map {$_||0} @count;
    # @count_distinct = map {$_||0} @count_distinct;
    foreach my $i (0 .. $#count_distinct) {
      my $aref = $count_distinct[$i];
      @$aref = map {$_||0} @$aref;
      print "  $i  ",join(',', @$aref),"\n";
    }
  }
  exit;

  # return true if the connected components containing $u and $v are isomorphic
  sub Graph_component_is_isomorphic {
    my ($graph, $u, $v) = @_;
    return MyGraphs::Graph_is_isomorphic
      ($graph->subgraph([$graph->connected_component_by_vertex($u)]),
       $graph->subgraph([$graph->connected_component_by_vertex($v)]));
  }

  # ($graph, $v1, $v2, ...)
  # return true if all given vertices are in different connected components
  sub Graph_is_different_components {
    my $graph = shift;
    my $u = shift;
    foreach my $v (@_) {
      if ($graph->same_connected_components($u,$v)) { return 0; }
    }
    return 1;
  }
}


{
  # FLAT difference
  require FLAT;
  my $f = FLAT::Regex->new('foo | bar')->as_dfa;
  my $t = FLAT::Regex->new('bar |f|o')->as_dfa;
  print "$f\n";
  print "$t\n";
  $f = $f->difference($t);
  print $f->is_empty ? "empty\n" : "not empty\n";

  require MyFLAT;
  MyFLAT::FLAT_show_acyclics($f);
  exit 0;
}
{
  # FLAT difference
  require FLAT;
  require MyFLAT;
  my $f = FLAT::Regex->new('(1|2)*')->as_dfa;
  my $t = FLAT::Regex->new('2*')->as_dfa;
  # print "$f\n";
  # print "$t\n";
  $f = $f->symdiff($t);  # f or t but not both
  $f = $f->as_dfa;
  MyFLAT::FLAT_show_breadth($f);
  exit 0;
}
{
  # Dpred high to low
  my $f = Dpred_k_flat();
  $f = $f->intersect(TurnLpred_flat());
  MyFLAT::view($f->MyFLAT::minimize);
  exit 0;
}
{
  # Dpred by right boundary

  require MyFLAT;
  my $f1 = MyFLAT::aref_to_FLAT_DFA (\@Rpred_state_table,
                                     accepting_list => [2,3,4]); # Y,Z,non
  my $f2 = MyFLAT::aref_to_FLAT_DFA (\@Rpred_state_table,
                                     accepting_list => [1,2,4]); # X,Y,non
  $f1 = $f1->MyFLAT::concat(FLAT::Regex->new('1 0*'));
  $f2 = $f2->MyFLAT::concat(FLAT::Regex->new('2 0*'));
  my $DpredLeft = $f1->union($f2)->as_dfa;
  $DpredLeft = $DpredLeft->MyFLAT::minimize;
  $DpredLeft->{'name'} = 'DpredLeft';

  FLAT_check_is_subset($DpredLeft, TurnLpred_flat());
  $DpredLeft->{'name'} = 'DpredLeft';

  my $DpredRight = $DpredLeft->MyFLAT::FLAT_rename
    (symbols_map => {0=>4,4=>0, 1=>3,3=>1});
  $DpredRight->{'name'} = 'DpredRight';

  my $Dpred = $DpredLeft->union($DpredRight)->as_dfa;
  $Dpred = $Dpred->MyFLAT::minimize;
  $Dpred->{'name'} = 'Dpred';

  my $DpredLeft_LtoH = $DpredLeft->MyFLAT::reverse;
  $DpredLeft_LtoH = $DpredLeft_LtoH->MyFLAT::minimize;
  $DpredLeft_LtoH->{'name'} = 'DpredLeft LtoH';

  MyFLAT::view($Dpred);
  # exit;
  # MyFLAT::view($DpredLeft);
  # MyFLAT::view($DpredLeft_LtoH);
  # MyFLAT::view(Rpred_flat());
  # MyFLAT::view(Rpred_flat()->MyFLAT::reverse->MyFLAT::minimize);

  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $k (0 .. 10) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      @n_list = grep {$_ < $n_hi} @n_list;
      my $n5 = sprintf '%0*s', $k, cnv($n,10,5);

      {
        my $d_by_path = @n_list==2 && turn($n)==-1;
        my $d_by_flat = $DpredRight->contains($n5) ?1:0;
        if ($d_by_path != $d_by_flat) {
          print "n=$n $n5 bad Right flat=$d_by_flat",
            " path=$d_by_path (",join(',',@n_list),")\n";
        }
      }
      {
        my $d_by_path = @n_list==2 && turn($n)==1;
        my $d_by_flat = $DpredLeft->contains($n5) ?1:0;
        if ($d_by_path != $d_by_flat) {
          print "n=$n $n5 bad Left flat=$d_by_flat",
            " path=$d_by_path (",join(',',@n_list),")\n";
        }
      }
    }
  }
  exit 0;
}

{
  # Rpred and Lpred both boundary
  require MyFLAT;
  my $f = Rpred_flat()->intersect(Lpred_k_flat());
  $f = $f->MyFLAT::reverse;
  $f = $f->MyFLAT::minimize;
  MyFLAT::view($f);
  exit 0;
}


{
  # other n state transitions

  # digit a of n means z/$dir == 0,1+I,I,1,2*i
  # which means z/$other_dir == [0,1+I,I,1,2*i]*$dir/$other_dir
  # apply(unpoint_z_to_low_digit,[0,1+I,I,1,2*I]*I)   == [0, 3, 4, 2, 1]
  # apply(unpoint_z_to_low_digit,[0,1+I,I,1,2*I]*I^2) == [0, 2, 1, 4, 3]
  # apply(unpoint_z_to_low_digit,[0,1+I,I,1,2*I]*I^3) == [0, 4, 3, 1, 2]
  # apply(unpoint_z_to_low_digit,[0,1+I,I,1,2*I]*I^3) == [0, 4, 3, 1, 2]
  # apply(z->(real(z)+2*imag(z))%5, unpoint_z_to_low_digit_vector)
  # apply(z->(real(z)+2*imag(z))%5, digitpoint_vector)

  sub xy_mod_b {
    my ($x,$y) = @_;
    return ($x + 2*$y) % 5;
  }

  BEGIN {
    my @unpoint_xy_to_low_digit = (0,3,2,1,4);
    sub unpoint_xy_to_low_digit {
      my ($x,$y) = @_;
      return $unpoint_xy_to_low_digit[xy_mod_b($x,$y)];
    }
  }

  my @low_digit_to_x = (0,1,0,1,0);
  my @low_digit_to_y = (0,1,1,0,2);
  foreach my $digit (0 .. $#low_digit_to_x) {
    unpoint_xy_to_low_digit($low_digit_to_x[$digit],$low_digit_to_y[$digit])
      == $digit or die;
  }

  require MyFSM;
  my %table;
  my $state_to_name = sub {
    my ($state) = @_;
    my ($dx,$dy,$dir,$other_dir, $prev_flip) = @$state;
    my $ddir = ($other_dir - $dir) % 4;
    my $parity = ($dx+$dy)%2 ? 'odd' : 'even';
    return "$dx,$dy $dir,$other_dir ";
    return "$dx,$dy $dir,$other_dir ddir=$ddir";
    return "$dx,$dy $dir,$other_dir ddir=$ddir  $prev_flip";

    # my ($z_delta,$dir_delta) = @$state;
    # my $out = ''; # ($bit ^ ($delta_x+$delta_y)) & 1;
    # my $flip = ''; # ($delta*$ap) % 3;
    # # if ($flip==2) { $flip = -1; }
    # return "$delta, dirsign=$dirsign flip $flip";
  };
  # [delta, dirsign]
  my @pending = ([0,0,0,2, 'none'],
                 # [1,1,0,2, 'none'],
                 # [2,2,0,2, 'none'],
                 # [3,3,0,2, 'none'],
                );
  my %initials = map {$state_to_name->($_) => 1} @pending;
  my %accepting;
  my $transition_use_new_digit = 0;
  my %seen;
  my %state_to_digit_deltas;
  while (@pending) {
    my $state = shift @pending;
    my $from = $state_to_name->($state);
    next if $seen{$from}++;

    print "$from\n";
    my ($dx,$dy,$dir,$other_dir) = @$state;
    if ($dx == 0 && $dy == 0 && $dir==$other_dir) {
      print "  is accepting\n";
      $accepting{$from} = "same";
    }

    foreach my $a (0 .. 4) {
      my ($dx,$dy,$dir,$other_dir) = @$state;
      ### $a
      ### dir(a): dir($a)
      ### $dx
      ### $dy
      ### $dir
      ### $other_dir

      my $z_over_d_x = $low_digit_to_x[$a];
      my $z_over_d_y = $low_digit_to_y[$a];
      ### $z_over_d_x
      ### $z_over_d_y

      my ($zx,$zy) = xy_rotate_plus90($z_over_d_x,$z_over_d_y, $dir);
      ### $zx
      ### $zy

      my $other_zx = $zx + $dx;
      my $other_zy = $zy + $dy;
      ### $other_zx
      ### $other_zy

      my ($other_z_over_other_d_x,$other_z_over_other_d_y)
        = xy_rotate_minus90($other_zx,$other_zy, $other_dir);
      ### $other_z_over_other_d_x
      ### $other_z_over_other_d_y

      my $other_a = unpoint_xy_to_low_digit($other_z_over_other_d_x,
                                            $other_z_over_other_d_y);
      ### $other_a
      ### dir(other_a): dir($other_a)

      $dir = ($dir - dir($a)) % 4;
      $other_dir = ($other_dir - dir($other_a)) % 4;
      ### new dir: $dir
      ### new other dir: $other_dir

      my $ax = $digit_to_x[$a];
      my $ay = $digit_to_y[$a];
      ### $ax
      ### $ay
      ($ax,$ay) = xy_rotate_plus90($ax,$ay, $dir);
      ### rotated ...
      ### $ax
      ### $ay
      $zx -= $ax;
      $zy -= $ay;
      ### $zx
      ### $zy
      ### div ...
      ($zx,$zy) = div($zx,$zy);
      ### $zx
      ### $zy

      my $other_ax = $digit_to_x[$other_a];
      my $other_ay = $digit_to_y[$other_a];
      ($other_ax,$other_ay) = xy_rotate_plus90($other_ax,$other_ay, $other_dir);
      $other_zx -= $other_ax;
      $other_zy -= $other_ay;
      ($other_zx,$other_zy) = div($other_zx,$other_zy);

      $dx = $other_zx - $zx;
      $dy = $other_zy - $zy;

      my $a_diff = ($other_a - $a) % 5;
      my $flip = ($a_diff == 0 ? "same" : "flip +$a_diff");
      my $new_state = [$dx,$dy,$dir,$other_dir, $flip];
      my $to = $state_to_name->($new_state);
      $table{$from}->{$transition_use_new_digit ? $other_a : $a} = $to;
      print "  $a -> $to  as old=$a new=$other_a $flip\n";
      # $state_to_digit_deltas{$from} .= ",$a_diff";
      $state_to_digit_deltas{$from} .= ",$other_a";

      push @pending, $new_state;
    }
  }
  print scalar(keys %table)," states\n";

  my $fsm = MyFSM->new(table => \%table,
                       initials => \%initials,
                       accepting => \%accepting,
                       title => 'N Other Transitions'
                       . ($transition_use_new_digit ? ", New Digit" : ""),
                       flow => 'west',
                      );
  foreach my $state (keys %table) {
    $fsm->rename_state($state, $state.' a'.$state_to_digit_deltas{$state});
  }
  # my @accepting = keys %accepting;
  # foreach my $i (1 .. $#accepting) {
  #   $fsm->merge_state($accepting[0],$accepting[1]);
  # }

  # $fsm = $fsm->digit_delta;
  $fsm->show;
  $fsm->simplify (verbose => 0);

  #  $fsm = $fsm->reverse;
  # $fsm->simplify (verbose => 1);
  # $fsm->renumber_states;

  # $fsm->print_tikz;
  $fsm->view;
  exit 0;
}


{
  # other

  my @low_digit_to_x = (0,1,0,1,0);
  my @low_digit_to_y = (0,1,1,0,2);
  foreach my $digit (0 .. $#low_digit_to_x) {
    unpoint_xy_to_low_digit($low_digit_to_x[$digit],$low_digit_to_y[$digit])
      == $digit or die;
  }

  my $n = 17;
  my $n5 = cnv($n,10,5);
  my @n_digits = digit_split_lowtohigh($n,5);
  push @n_digits, 0;

  my $path = Math::PlanePath::R5DragonCurve->new;
  my ($x,$y) = $path->n_to_xy($n);
  my @n_list = $path->xy_to_n_list($x,$y);
  my @other_n_digits;

  my $dx = 0;
  my $dy = 0;
  my $dir = dir($n);
  my $other_dir = ($dir + 2) % 4;

  foreach my $a (@n_digits) {
    ### $a
    ### dir(a): dir($a)
    ### $dx
    ### $dy
    ### $dir
    ### $other_dir

    my $z_over_d_x = $low_digit_to_x[$a];
    my $z_over_d_y = $low_digit_to_y[$a];
    ### $z_over_d_x
    ### $z_over_d_y

    my ($zx,$zy) = xy_rotate_plus90($z_over_d_x,$z_over_d_y, $dir);
    ### $zx
    ### $zy

    my $other_zx = $zx + $dx;
    my $other_zy = $zy + $dy;
    ### $other_zx
    ### $other_zy

    my ($other_z_over_other_d_x,$other_z_over_other_d_y)
      = xy_rotate_minus90($other_zx,$other_zy, $other_dir);
    ### $other_z_over_other_d_x
    ### $other_z_over_other_d_y

    my $other_a = unpoint_xy_to_low_digit($other_z_over_other_d_x,
                                          $other_z_over_other_d_y);
    ### $other_a
    ### dir(other_a): dir($other_a)
    push @other_n_digits, $other_a;

    $dir = ($dir - dir($a)) % 4;
    $other_dir = ($other_dir - dir($other_a)) % 4;
    ### new dir: $dir
    ### new other dir: $other_dir

    my $ax = $digit_to_x[$a];
    my $ay = $digit_to_y[$a];
    ### $ax
    ### $ay
    ($ax,$ay) = xy_rotate_plus90($ax,$ay, $dir);
    ### rotated ...
    ### $ax
    ### $ay
    $zx -= $ax;
    $zy -= $ay;
    ### $zx
    ### $zy
    ### div ...
    ($zx,$zy) = div($zx,$zy);
    ### $zx
    ### $zy

    my $other_ax = $digit_to_x[$other_a];
    my $other_ay = $digit_to_y[$other_a];
    ($other_ax,$other_ay) = xy_rotate_plus90($other_ax,$other_ay, $other_dir);
    $other_zx -= $other_ax;
    $other_zy -= $other_ay;
    ($other_zx,$other_zy) = div($other_zx,$other_zy);

    $dx = $other_zx - $zx;
    $dy = $other_zy - $zy;

    my $a_diff = ($other_a - $a) % 5;
    my $flip = ($a_diff == 0 ? "same" : "flip +$a_diff");
    print "  $a -> $dx,$dy,$dir,$other_dir  as old=$a new=$other_a $flip\n";
  }

  my $got_other_n = digit_join_lowtohigh(\@other_n_digits,5);
  print "other ",join(',',@other_n_digits)," = $got_other_n",
    "  want ",join(',',@n_list),"\n";
  exit 0;
}


{
  # hullsaround fully enclosed common segments

  &fBpred_hullsaround_init(verbose=>2);
  my $int = $hullsaround_segs_all;
  foreach my $state ($fNonBpred_hullsaround_flat->get_accepting) {
    my $conf = $fBpred_hullsaround_state_to_conf[$state];
    $conf->print;
    $int = $int->intersection($conf);
  }
  $int->print;
  exit 0;
}
{
  require MyGraphs;
  my $k = 4;
  my $turn_graph = r5quad_turn_Graph($k);
  $turn_graph->is_undirected or die;
  for (;;) {
    my $area_tree  = r5quad_area_tree_Graph($k-1);
    $area_tree->is_undirected or die;

    foreach my $xy (1, 0) {
      my $replacement = MyGraphs::Graph_cycle_replacement($area_tree, 4,
                                                          vertices_between=>1);
      $replacement->is_undirected or die;

      if (MyGraphs::Graph_is_isomorphic($replacement, $turn_graph)) {
        print "ok  xy=$xy\n";
      } else {
        print "bad  xy=$xy\n";
        my $r_num_vertices = $replacement->vertices;
        my $t_num_vertices = $turn_graph->vertices;
        print "  num_vertices  $r_num_vertices $t_num_vertices\n";
        my $r_num_edges = $replacement->edges;
        my $t_num_edges = $turn_graph->edges;
        print "  num_edges     $r_num_edges $t_num_edges\n";
      }

      # with or without the xy vertex naming
      $area_tree->set_graph_attribute('vertex_name_type',undef);
    }
  }

  # my $s1 = join(',',MyGraphs::Graph_degree_sequence($turn_graph));
  # my $s2 = join(',',MyGraphs::Graph_degree_sequence($replacement));
  # print "$s1\n";
  # print "$s2\n";
  # is ($s1,$s2);
  # print MyGraphs::graph6_str_to_canonical(MyGraphs::Graph_to_graph6_str($turn_graph));
  # print MyGraphs::graph6_str_to_canonical(MyGraphs::Graph_to_graph6_str($replacement));
  # MyGraphs::Graph_view($turn_graph);
  # MyGraphs::Graph_view($area_tree);
  # MyGraphs::Graph_view($replacement);

  exit;
}

{
  &fBpred_hullsaround_init(verbose=>2);
  {
    my $f = $fRpred_hullsaround_flat->MyFLAT::minimize;
    MyFLAT::view($f);
  }
  {
    my $f = $fLpred_hullsaround_flat->MyFLAT::minimize;
    MyFLAT::view($f);
  }
  exit 0;
}
{
  # hullsaround segments

  $B_squares_segs->print;
  $hullsaround_segs_outer->print;
  $hullsaround_segs_inner->print;
  $hullsaround_segs_all->print;

  MyR5DragonConf->initial->print;
  foreach my $digit (0 .. 4) {
    my $s = MyR5DragonConf->initial;
    $s->expand($digit)->print;
  }

  foreach my $digit (0 .. 4) {
    $hullsaround_segs_all->expand($digit)->print;
  }
  exit 0;
}
{
  # convex hull width
  # Y 0.5769885917447919200071754
  # X 0.1769885917447919200071754

  # angle
  # arg(1+2*I)*180/Pi
  # 1.10714871779409050301706 \\ radians A105199 arctan(2)
  # 63.4349488229220106484278

  # upper bound
  # suminf(i=1,1/sqrt(5)^i)
  # 1/(1-1/sqrt(5))-1
  # (1/5+1/sqrt(5))/(4/5)
  # (1+sqrt(5))/4
  # 0.809016994374947424102293 \\ A019863  sin(54deg)  3/20 of circle
  #
  # end
  # suminf(i=2,1/sqrt(5)^i)
  # (5+sqrt(5))/20
  # 0.361803398874989484820458  \\ close to phi


  my $points = [ [to_bigint(0),to_bigint(0)], [to_bigint(1),to_bigint(0)] ];
  my $ex = to_bigint(1);
  my $ey = to_bigint(0);

  my $prev_max_x = 0;
  my $prev_max_y = 0;
  foreach my $k (0 .. 200) {
    my $count = scalar(@$points);
    print "k=$k count $count  ending $ex,$ey\n";

    my $max = 0;
    my @max_i = 0;
    my $max_x = 0;
    my $max_y = 0;
    foreach my $i (0 .. $#$points) {
      my $p = $points->[$i];
      my ($x,$y) = @$p;
      # (x+I*y)*conj(ex+ey*I) == (x*ex+y*ey) + (x*-ey+y*ex)*I
      my $wx = $x*$ex + $y*$ey;
      my $wy = $x*-$ey + $y*$ex;
      $wx = $wx;
      $wy = $wy;
      my $m = -$wy;
      if ($m > $max) {
        $max = $m; @max_i = ();
        $max_x = $x;
        $max_y = $y;
      }
      if ($m == $max) { push @max_i, $i; }
    }
    my $decimals = 35;
    $max *= to_bigint(10)**$decimals;
    $max /= to_bigint(5)**$k;
    printf "  max at i=".join(',',@max_i)." = %0*s\n", $decimals, $max;

    ($prev_max_x,$prev_max_y) = mul($prev_max_x,$prev_max_y);
    my $dx = $max_x - $prev_max_x;
    my $dy = $max_y - $prev_max_y;
    print "  coords   $max_x,$max_y  incr $dx,$dy";
    if (($dx&1)==0 && ($dy&1)==0) {
      $dx /= 2;
      $dy /= 2;
      print " half $dx,$dy";
    }
    my $dir = dxdy_to_dir4($dx,$dy);
    print " dir=$dir\n";

    ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
    $prev_max_x = $max_x;
    $prev_max_y = $max_y;
  }
  exit 0;
}
{
  # convex hull boundary length sqrt(5) part

  # 4*3*sqrt5^0 + 2*2*sqrt5^1   \\ k=3
  # 4*3*sqrt5^0 + 4*2*sqrt5^1 + 2*2*sqrt5^2 \\ k==4
  # HB5(k) = sum(j=0,k-2, if(j==0,4*3, j<k-2, 4*2, 2*2) * sqrt5^j)
  # vector(12,k,k+=3; HB5(k))
  # HB5(k) = 2*(3/5 + 1/sqrt5)*(sqrt5^k - 5)
  # 2*(3/5 + 1/sqrt5) *1.0
  # 2.09442719099991587856366
  my %powers;
  {
    print "powers ";
    my $x = to_bigint(1);
    my $y = to_bigint(0);
    foreach my $k (0 .. 40) {
      print " $y:$x";
      foreach (1 .. 4) {
        $powers{"$y:$x"} = $k;
        ($x,$y) = xy_rotate_plus90($x,$y);
      }
      ($x,$y) = mul($x,$y);
    }
    print "\n";
  }

  my $points = [ [to_bigint(0),to_bigint(0)], [to_bigint(1),to_bigint(0)] ];
  my $ex = to_bigint(1);
  my $ey = to_bigint(0);

  foreach my $k (0 .. 10) {
    print "k=$k\n";
    my $diffs = points_differences($points);
    my $total_integer = to_bigint(0);
    my $total_sqrt5 = to_bigint(0);
    my $non_inexact = 0;
    my @power_counts;
    foreach my $p (@$diffs) {
      my ($dx,$dy) = @$p;
      my $g = gcd($dx,$dy);
      $dx /= $g;
      $dy /= $g;
      if (defined (my $power = $powers{"$dy:$dx"})) {
        print "  power $g sqrt5**$power\n";
        $power_counts[$power]++;
        if ($power & 1) {
          $power >>= 1;
          $total_sqrt5 += $g*to_bigint(5)**$power;
        } else {
          $power >>= 1;
          $total_integer += $g*to_bigint(5)**$power;
        }
      } else {
        print "  non  $dy:$dx  g=$g\n";
        $non_inexact += hypot($dx->numify,$dy->numify)*$g->numify;
      }
    }
    my $inexact = $total_integer->numify + sqrt(5)*$total_sqrt5->numify;
    my $f = $inexact / (sqrt(5)**$k);
    print "  total $total_integer + sqrt5 * $total_sqrt5  = $inexact  $f\n";

    @power_counts = map {$_//0} @power_counts;
    print "  pows ",join(',',@power_counts),"\n";

    my $non_f = $non_inexact / (sqrt(5)**$k);
    print "  total non $non_inexact  $non_f\n";


    ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
  }
  exit 0;
}

{
  # convex hull slopes
  my %powers;
  {
    print "powers ";
    my $x = to_bigint(1);
    my $y = to_bigint(0);
    foreach my $k (0 .. 40) {
      print " $y:$x";
      foreach (1 .. 4) {
        $powers{"$y:$x"} = 1;
        ($x,$y) = xy_rotate_plus90($x,$y);
      }
      ($x,$y) = mul($x,$y);
    }
    print "\n";
  }


  my $points = [ [to_bigint(0),to_bigint(0)], [to_bigint(1),to_bigint(0)] ];
  my $ex = to_bigint(1);
  my $ey = to_bigint(0);

  foreach my $k (0 .. 30) {
    my $diffs = points_differences($points);
    my @gcds;
    my $slopes = [ map { my ($dx,$dy) = @$_;
                         my $g = gcd($dx,$dy);
                         push @gcds, $g;
                         $dx /= $g;
                         $dy /= $g;
                         [$dx,$dy] } @$diffs ];
    my $len = scalar(@$slopes);
    my $non = 0;
    foreach my $p (@$slopes) {
      my ($dx,$dy) = @$p;
      my $g = gcd($dx,$dy);
      $dx /= $g;
      $dy /= $g;
      if (! $powers{"$dy:$dx"}) {
        $non++;
      }
    }
    print "k=$k ($non/$len non)  ";

    foreach my $i (0 .. $#$slopes) {
      if ($i >= 15) {
        print " ...";
        my $p = $slopes->[-1];
        my ($dx,$dy) = @$p;
        print " $dy:$dx";
        if (! $powers{"$dy:$dx"}) {
          print "*";
        }
        last;
      }
      my $p = $slopes->[$i];
      my ($dx,$dy) = @$p;
      print " $dy:$dx";
      if (! $powers{"$dy:$dx"}) {
        print "*";
      }
    }
    print "\n";
    if (points_contains($points,0,0)) {
      print "  (inc origin)\n";
    }
    print "  ",join(',',@gcds),"\n";

    ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
    $k++;
  }
  exit 0;

  sub points_contains {
    my ($points, $x,$y) = @_;
    foreach my $p (@$points) {
      if ($x==$p->[0] && $y==$p->[1]) {
        return 1;
      }
    }
    return 0;
  }

  sub points_differences {
    my ($points, $x,$y) = @_;
    return [ map { my $p1 = $points->[$_];
                   my $p2 = $points->[($_+1) % scalar(@$points)];
                   [ $p2->[0] - $p1->[0],
                     $p2->[1] - $p1->[1] ]
                 } 0 .. $#$points ];
  }
}

{
  # convex hull vertex originating parts
  # k=4  has 1..3 of 0 and 1..3 of 1       from among 0..4
  #
  # k=5  has 1..5 of 0 and 2..4 of 1       from among 0..5
  # k=6  has 2..7 of 0 and 3..6 of 1       from among 0..7
  # k=7  has 3..9 of 0 and 4..8 of 1       from among 0..9
  #    makes 0..6          7..11
  # k=8  has 4..11 of 0 and 5..10 of 1     from among 0..11
  #    makes 0..7          8..13
  # k=9  has 4..12 of 0 and 5..11 of 1     from among 0..13
  # k=10 has 5..14 of 0 and 6..13 of 1     from among 0..15
  my $points = [ [0,0], [1,0] ];
  my $ex = 1;
  my $ey = 0;

  my $k = 0;
  foreach (1 .. 8) {
    ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
    $k++;
  }
  my $count = scalar(@$points);
  print "k=$k count $count  ending $ex,$ey\n";
  print points_str($points),"\n";
  my %name;
  foreach my $i (0 .. $#$points) {
    my $p = $points->[$i];
    foreach my $digit (0 .. 4) {
      my ($x,$y) = unfold (@$p, $ex,$ey, $digit);
      $name{"$x,$y"} = "$digit.P$i";
    }
  }

  ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
  $k++;
  print points_str($points),"\n";
  foreach my $p (@$points) {
    print $name{"$p->[0],$p->[1]"},"  at $p->[0],$p->[1]\n";
  }
  exit 0;
}
{
  # convex hull vertex count
  my $points = [ [to_bigint(0),to_bigint(0)], [to_bigint(1),to_bigint(0)] ];
  my $ex = to_bigint(1);
  my $ey = to_bigint(0);
  foreach my $k (0 .. 100) {
    my $count = scalar(@$points);
    print "k=$k count $count  ending $ex,$ey\n";
    ($points, $ex,$ey) = hull_expand($points, $ex,$ey)
  }
  exit 0;
}

{
  # convex hull
  my $points = [ [0,0], [1,0] ];
  my @P1;
  my $ex = 1;
  my $ey = 0;
  foreach my $k (0 .. 10) {
    push @P1, $points->[0];
    my $count = scalar(@$points);
    print "count $count  ",points_str($points),"  ending $ex,$ey\n";
    ($points, $ex,$ey) = hull_expand($points, $ex,$ey)
  }

  say points_str_gp(\@P1);

  my $bk = Math::Complex->new($ex,$ey);
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    my $z = Math::Complex->new($x,$y);
    $z /= $bk;
    printf "(%.5f,%.5f)node[my dot]--", $z->Re, $z->Im;
  }
  print "cycle;\n";

  my $b = Math::Complex->new(1,2);
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    my $z = Math::Complex->new($x,$y);
    $z /= $bk;
    $z /= $b;
    printf "(%.5f,%.5f)--", $z->Re, $z->Im;
  }
  print "cycle;\n";

  my $z2 = Math::Complex->new(1,1)/$b;
  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    my $z = Math::Complex->new($x,$y);
    $z /= $bk;
    $z /= $b;
    $z = $z2-$z;
    printf "(%.5f,%.5f)--", $z->Re, $z->Im;
  }
  print "cycle;\n";

  foreach my $p (@$points) {
    my ($x,$y) = @$p;
    my $z = Math::Complex->new($x,$y);
    $z /= $bk;
    $z /= $b;
    $z *= Math::Complex->new(0,1);
    $z += 1/$b;
    printf "(%.5f,%.5f)--", $z->Re, $z->Im;
  }
  print "cycle;\n";

  exit 0;
}
{
  # convex hull by points
  # count 6  0,0 1,0 1,3 -3,4 -4,4 -4,1
  # count 8  1,0 1,3 -3,5 -6,5 -12,-2 -12,-5 -8,-7 -5,-7

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @P1;
  foreach my $k (1 .. 8) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($ex,$ey) = $path->n_to_xy($n_hi);
    my $points = [ map {[$path->n_to_xy($_)]} $n_lo .. $n_hi ];

    if (@$points >= 5) {
      require Math::Geometry::Planar;
      my $planar = Math::Geometry::Planar->new;
      $planar->points($points);
      $planar = $planar->convexhull2;
      $points = $planar->points;
    }

    hull_points_rotate_first_horizontal($points);

    push @P1, $points->[0];
    my $count = scalar(@$points);
    print "count $count  ",points_str($points),"\n";

    my $bk = Math::Complex->new($ex,$ey);
    foreach my $p (@$points) {
      my ($x,$y) = @$p;
      my $z = Math::Complex->new($x,$y);
      $z /= $bk;
      printf "(%.5f,%.5f)node[my dot]--", $z->Re, $z->Im;
    }
    print "cycle;\n";
  }

  say points_str_gp(\@P1);
  exit 0;
}





{
  # join centroid in join square number
  #
  # join square number counted from unfold point
  # JS(k)=6,20,60,182,546,1640,4920,14762
  #   ternary 2020, 20202
  #   A081251 g(x)=2/((1-x)*(1+x)*(1-3*x))
  #           a(n) = a(n-2) + 2*3^(n)
  #           a(n) = 2*A033113(n+1)    A033113=ternary 1010101
  # JS(k) = (3^k - 2 + (-1)^k)/4
  # vector(10,k,JS(k))
  #
  # N segment = base5 4242424  or k even 4240,4241,4242
  # GJseg(k) = if(k==0,1, 5*GJseg(k-1) - (-1)^(k-1)) + 3;
  # GJseg(k) = 1/12*(11*5^k - 2*(-1)^k - 9);
  # vector(10,k,digits(GJseg(k),5))

  # k=3  jxy=-10,-6
  #   n=114 [424]  j=10 [20][101]  nlen=1
  # k=4  jxy=0,-24
  #   n=572 [4240,4241,4242]  j=36 [121][1100]  nlen=3
  # k=5  jxy=48,-22
  #   n=2864 [42424]  j=116 [431][11022]  nlen=1
  # k=6  jxy=92,74
  #   n=14322 [424240,424241,424242]  j=358 [2413][111021]  nlen=3
  # k=7  jxy=-58,258
  #   n=71614 [4242424]  j=1086 [13321][1111020]  nlen=1
  # k=8  jxy=-576,144
  #   n=358072 [42424240,42424241,42424242]  j=3272 [101042][11111012]  nlen=3
  # k=9  jxy=-864,-1006
  #   n=1790364 [424242424]  j=9832 [303312][111111011]  nlen=1
  # k=10  jxy=1148,-2734
  #   n=8951822 [4242424240,4242424241,4242424242]  j=29514 [1421024][1111111010]  nlen=3
  # k=11  jxy=6614,-438
  #   n=44759114 [42424242424]  j=88562 [10313222][11111111002]  nlen=1
  # k=12  jxy=7488,12792
  #   n=223795572 [424242424240,424242424241,424242424242]  j=265708 [32000313][111111111001]  nlen=3
  # k=13  jxy=-18096,27770
  #   n=1118977864 [4242424242424]  j=797148 [201002043][1111111111000]  nlen=1
  # k=14  jxy=-73636,-8422

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
 K: foreach my $k (1 .. 8) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($bx,$by) = $path->n_to_xy($n_hi);
    ### $n_hi

    my ($jx,$jy) = GJ_xy($k);
    if ($jx < 0) { $jx -= 1; }
    if ($jy < 0) { $jy -= 1; }
    $jx = int($jx);
    $jy = int($jy);

    print "k=$k  jxy=$jx,$jy\n";
    my $prev_mx = 0;
    my $prev_my = 0;
    my $j = 0;
    for (my $n = $n_hi-1; $n >= $n_lo; $n--) {
      ### $n
      next unless $path->_UNDOCUMENTED__n_segment_is_left_boundary($n,$k);
      # next unless $n_is_join->($n);

      my ($x1,$y1) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);

      my $x2 = $x1 + $dx;
      my $y2 = $y1 + $dy;
      ($dx,$dy) = xy_rotate_plus90($dx,$dy);

      my $x3 = $x2 + $dx;
      my $y3 = $y2 + $dy;
      ($dx,$dy) = xy_rotate_plus90($dx,$dy);

      my $x4 = $x3 + $dx;
      my $y4 = $y3 + $dy;

      my $mx = min($x1,$x2,$x3,$x4);
      my $my = min($y1,$y2,$y3,$y4);
      ### mxy: "$mx $my"

      my @n_list = ($n,
                    $path->xyxy_to_n_either($x2,$y2, $x3,$y3) // (),
                    $path->xyxy_to_n_either($x3,$y3, $x4,$y4) // (),
                    $path->xyxy_to_n_either($x4,$y4, $x1,$y1) // ());
      @n_list = sort {$a<=>$b} @n_list;
      my @n_list_above = grep {$_ >= $n_hi} @n_list;
      @n_list = grep {$_ < $n_hi} @n_list;
      ### n_list: join(' ',@n_list)
      ### above : join(' ',@n_list_above)

      @n_list_above > 0 || next; # not a join square
      (@n_list ==1 || @n_list==3) or die;

      if ($mx == $jx && $my == $jy) {
        my @n5_list = map {cnv($_,10,5)} @n_list;
        my $n_list = join(',',@n_list);
        my $n5_list = join(',',@n5_list);
        my $nlen = scalar(@n_list);
        # $n_list
        my $j5 = cnv($j,10,5);
        my $j3 = cnv($j,10,3);
        print "  n=$n_list[-1] [$n5_list]  j=$j [$j5]_5 [$j3]_3  nlen=$nlen\n";

         push @values, $n_list[-1];
        # push @values, $j;
        next K;
      }

      if ($mx != $prev_mx || $my != $prev_my) {
        $j++;
      }
      ($prev_mx,$prev_my) = ($mx,$my);
    }
  }

  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;

  # my $n_is_join = sub {
  #   my ($n) = @_;
  #   my ($x,$y) = $path->n_to_xy($n);
  #   my @n_list = $path->xy_to_n_list($x,$y);
  #   @n_list = grep {$_<=2*$n_hi} @n_list;
  #   ### @n_list
  #   my $below = grep {$_ < $n_hi} @n_list;
  #   my $above = grep {$_ > $n_hi} @n_list;
  #   return ($below == 1 && $above == 1);
  # };
}

{
  # join squares in directions
  #
  # k=1  1 0 0 0 = 1                1 0 0 0  = 1
  # k=2  1 2 0 1 = 4                1 2 0 1  = 4
  # k=3  4 3 4 2 = 13               2 1 2 0  = 5
  # k=4  9 11 9 11 = 40             0 2 0 2  = 4
  # k=5  32 29 31 29 = 121          3 0 2 0  = 5
  # k=6  90 93 89 92 = 364          1 4 0 2  = 7
  # k=7  275 272 275 271 = 1093     4 1 4 0  = 9
  # k=8  818 822 818 822 = 3280     0 4 0 4  = 8

  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (1 .. 9) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($k);
    my ($bx,$by) = $path->n_to_xy($n_hi);

    my @count = (0,0,0,0, 0,0,0,0);
    my $j = 0;
    my %seen;
    foreach my $n ($n_lo .. $n_hi-1) {
      ### $n
      my ($x1,$y1) = $path->n_to_xy($n);
      my ($dx1,$dy1) = $path->n_to_dxdy($n);

      my $x2 = $x1 + $dx1;
      my $y2 = $y1 + $dy1;
      my ($dx2,$dy2) = xy_rotate_plus90($dx1,$dy1);

      my $x3 = $x2 + $dx2;
      my $y3 = $y2 + $dy2;
      my ($dx3,$dy3) = xy_rotate_plus90($dx2,$dy2);

      my $x4 = $x3 + $dx3;
      my $y4 = $y3 + $dy3;
      my ($dx4,$dy4) = xy_rotate_plus90($dx3,$dy3);

      my $mx = min($x1,$x2,$x3,$x4);
      my $my = min($y1,$y2,$y3,$y4);
      ### mxy: "$mx $my"
      next if $seen{"$mx,$my"}++;

      my $n2 = $path->xyxy_to_n_either($x2,$y2, $x3,$y3) // next;
      my $n3 = $path->xyxy_to_n_either($x3,$y3, $x4,$y4) // next;
      my $n4 = $path->xyxy_to_n_either($x4,$y4, $x1,$y1) // next;

      my @v = (0,
               $n2 >= $n_hi ? 1 : 0,
               $n3 >= $n_hi ? 1 : 0,
               $n4 >= $n_hi ? 1 : 0);
      my $other_sides = sum(@v);
      next if $other_sides == 0;
      my $sides = 4 - $other_sides;
      ($sides == 1 || $sides == 3) or die "sides=$sides";
      my $single_v = ($sides == 1 ? 0 : 1);
      my $i = 0;
      for ( ; $i <= $#v; $i++) { last if $v[$i] == $single_v; }
      my $d = ($i + dxdy_to_dir4($dx1,$dy1)) % 4;
      if ($sides == 3) { $d += 4; }

      $count[$d]++;
    }

    my $total = sum(@count);
    print "k=$k  ";
    foreach my $i (0 .. 7) {
      if ($i == 4) { print "  "; }
      print " ",$count[$i];
    }
    print " = $total ";
    foreach my $i (0 .. 3) { print " ",$count[($i+2)%4] - $count[$i+4]; }
    print "\n";

    $total == J($k) or die;
    push @values, $count[0];
  }

  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}

{
  # JEQ square at end of join

  #                   *  *--*
  #                   |  |  |
  #                   *--*--*      direction rotate +90 each
  #         *            |
  #         ^            *--*
  #         |               |
  #   *---->*            *--*
  #
  # J(k) = (3^k-1)/2;
  # vector(10,k, J(k) == 3*J(k-1)+1)    /* 1 new each level */
  #

  my $path = Math::PlanePath::R5DragonCurve->new;
  K: foreach my $k (0 .. 8) {
    print "k=$k\n";
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      next unless $path->_UNDOCUMENTED__n_segment_is_left_boundary($n,$k);
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      next unless @n_list >= 2;
      my $n2 = $n_list[1];
      next if $n2 >= 5*$n_hi;
      if ($n2 >= $n_hi) {
        my ($dx,$dy) = $path->n_to_dxdy($n-1);
        my $dir4 = dxdy_to_dir4($dx,$dy);
        my ($jx,$jy) = JEQ($k);
        print "  touch at $x,$y   $dx,$dy dir=$dir4   $jx,$jy\n";
        next K;
      }
    }
  }
  exit 0;

  sub JEQ {
    my ($k) = @_;
    my ($x,$y) = (1,0);
    my ($dx,$dy) = (-1,0);
    for (1 .. $k) {
      ($x,$y) = mul($x,$y);
      $x += $dx;
      $y += $dy;
      ($dx,$dy) = xy_rotate_plus90($dx,$dy);
    }
    return ($x,$y);
  }
}

{
  # EpredL in base-5
  foreach my $n (0 .. 5**5) {
    if (EpredL_by_digits_low($n)) {
      my $n5 = cnv($n,10,5);
      print "$n [$n5]\n";
    }
  }
  exit 0;
}

{
  # segment enclosure -- check low to high
  #

  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $want (1, -1) {
    if ($want == 1) { print "EpredR\n"; } else { print "EpredL\n"; }
    foreach my $n (0 .. 19) {
      my $path_state = n_to_enc_state_by_path($path, $n);
      my $path_enc_RL = enc_state_to_RL($path_state);
      print $path_enc_RL == $want ? 1 : 0, ",";
      if ($n % 5 == 4) { print " \\; "; }
    }
    print "\n";
    my $count = 0;
    for (my $n = 0; $count < 10; $n++) {
      my $path_state = n_to_enc_state_by_path($path, $n);
      my $path_enc_RL = enc_state_to_RL($path_state);
      if ($path_enc_RL == $want) {
        $count++;
        print "$n,";
      }
    }
    print "\n";
    print "\n";
  }

  my $denc = n_to_encRL_by_digits_low(713);
  ### $denc

  my $bad = 0;
  my $good = 0;
  foreach my $n (0 .. 5**8) {
    my $path_state = n_to_enc_state_by_path($path, $n);
    my $path_enc_RL = enc_state_to_RL($path_state);
    my $denc_RL = n_to_encRL_by_digits_low($n);
    if ($path_enc_RL == $denc_RL) {
      $good++;
    } else {
      my $n5 = cnv($n,10,5);
      print "n=$n [$n5]  path [s$path_state]=$path_enc_RL  digits $denc_RL\n";
      die "too much bad" if ++$bad > 10;
    }
  }
  print "ok  $good\n";
  exit 0;

  sub EpredR_by_digits_low {
    my ($n) = @_;

    {
      my @digits = reverse digit_split_lowtohigh($n,5); # high to low
      @digits = grep {$_ != 1} @digits;
      my $low = $digits[-1] || 0;
      if ($low == 2 || $low == 3) { return 0; }
      foreach my $i (reverse 0 .. $#digits-1) {
        if ($digits[$i] == 3 && $digits[$i+1] == 0) { return 1; }
        if ($digits[$i] == 4 && $digits[$i+1] == 0) { return 1; }
        if ($digits[$i] == 2 && $digits[$i+1] == 2) { return 0; }
        if ($digits[$i] == 3 && $digits[$i+1] == 2) { return 0; }
        if ($digits[$i] == 2 && $digits[$i+1] == 4) { return 0; }
        if ($digits[$i] == 3 && $digits[$i+1] == 4) { return 0; }
      }
      return 0;
    }
    {
      my $str = cnv($n,10,5);
      $str =~ s/1//g;
      $str =~ s/.*(30|40)/0/ or return 0;
      if ($str =~ /22|23|24|32|33|34/) { return 0; }
      if ($str =~ /[23]$/) { return 0; }
      return 1;
    }
    {
      for (;;) {
        # non-0s
        while (my $digit = _divrem_mutate($n,5)) {
          if ($digit==2 || $digit==3) { return 0; }
        }
        # non-2s
        while ((my $digit = _divrem_mutate($n,5)) != 2) {
          if ($digit==3 || $digit==4) { return 1; }
          if ($n == 0) { return 0; }
        }
      }
    }
  }
  sub EpredL_by_digits_low {
    my ($n) = @_;
    {
      my @digits = reverse digit_split_lowtohigh($n,5); # high to low
      @digits = grep {$_ != 3} @digits;
      my $low = $digits[-1] || 0;
      if ($low == 0 || $low == 1) { return 0; }
      foreach my $i (reverse 0 .. $#digits-1) {
        # non-4s can have 0,1,2  with 2 below
        if ($digits[$i] == 1 && $digits[$i+1] == 2) { return 1; } # 12
        if ($digits[$i] == 2 && $digits[$i+1] == 2) { return 1; } # 22
        if ($digits[$i] == 1 && $digits[$i+1] == 0) { return 1; } # 10
        if ($digits[$i] == 2 && $digits[$i+1] == 0) { return 1; } # 20
        # non-2s can have 0,1,4  with 4 below
        if ($digits[$i] == 0 && $digits[$i+1] == 4) { return 0; }
        if ($digits[$i] == 1 && $digits[$i+1] == 4) { return 0; }
      }
      return 0;
    }
    {
      # cf Lpred
      #             04,
      # 10, 11, 12, 14,
      # 20, 21, 22
      # any 11 or 21 has 12 or 10 below
      my $str = cnv($n,10,5);
      $str =~ s/3//g;
      $str =~ s/.*(10|12|20|22)/substr($1,1)/e or return 0;
      if ($str =~ /04|14/) { return 0; }
      if ($str =~ /[01]$/) { return 0; }
      return 1;
    }
    {
      for (;;) {
        # non-2s
        while ((my $digit = _divrem_mutate($n,5)) != 2) {
          if ($digit==0 || $digit==1) { return 0; }
          if ($n == 0) { return 0; }
        }
        # non-4s
        while ((my $digit = _divrem_mutate($n,5)) != 4) {
          if ($digit==1 || $digit==2) { return 1; }
          if ($n == 0) { return 0; }
        }
      }
    }
  }

  sub n_to_encRL_by_digits_low {
    my ($n) = @_;
    ### n_to_encRL_by_digits_low(): $n

    {
      my $R = EpredR_by_digits_low($n);
      my $L = EpredL_by_digits_low($n);
      if ($R && $L) { die; }
      if ($R) { return 1; }
      if ($L) { return -1; }
      return 0;
    }

    {
      if (LowestNonFour($n) <= 1) {
        # right
        #       1        0
        #       |        |
        #     1,4    2  0,1
        #       MM <--- MM
        #       |  --->  |
        #   2,3 |   0    | 3,4
        #      non      right
        #
        my $seek = 0;
        while ($n) {
          my $digit = _divrem_mutate($n,5);
          if ($digit == $seek) { $seek ^= 2; next; }
          if ($seek == 0 && ($digit == 2 || $digit == 3)) { return 0; }
          if ($seek == 2 && ($digit == 3 || $digit == 4)) { return 1; }
        }
        return 0;

      } else {
        ### low 2,3 lefts ...
        #
        #       3        2
        #       |        |
        #     0,1    4  0,3
        #       MM <--- MM
        #       |  --->  |
        #   3,4 |   2    | 1,2
        #      non      left
        #
        # 1203
        my $digit;
        do {
          $digit = _divrem_mutate($n,5);
        } while ($digit == 4);
        ($digit == 2 || $digit == 3) or die;
        my $seek = ($digit == 2 ? 4 : 2);

        while ($n) {
          $digit = _divrem_mutate($n,5);
          if ($digit == $seek) { $seek ^= 6; next; }
          if ($seek == 4 && ($digit == 1 || $digit == 2)) { return -1; }
          if ($seek == 2 && ($digit == 0 || $digit == 1)) { return 0; }
        }
        return 0;
      }
    }

    # {
    #   #
    #   # [34] [01]* 0$  right
    #   # [34] [01]+ 0.  right
    #   # [23] [14]* 1$   none
    #   # [23] [14]* 2   none
    #
    #   my $str = cnv($n,10,5);
    #   ### $str
    #   my $p1 = -1;
    #   if ($str =~ /.*([34])[01]*(0)/) {
    #     $p1 = $+[0];
    #   }
    #   my $p2 = -1;
    #   if ($str =~ /.*([23])[14]*(2|1$)/) {
    #     $p2 = $+[0];
    #   }
    #   ### $str
    #   ### $p1
    #   ### $p2
    #   if ($p1 >= 0 && $p1 > $p2) {
    #     ### yes, right ...
    #     return 1;
    #   }
    #   return 0;
    # }
    # {
    #   #
    #   # low 3
    #   # [12] [03]* 2$    left
    #   # [12] [03]* 2.    left
    #   # [01] [34]* 3$    none
    #   # [01] [34]* 4.    none
    #   #
    #   # vv
    #   # 120423
    #   #   ^^^^
    #   #
    #   # 12042
    #   #
    #   my $str = cnv($n,10,5);
    #   ### $str
    #   my $p1 = -1;
    #   if ($str =~ /.*([12])[03]*(2)/) {
    #     $p1 = $+[0];
    #   }
    #   my $p2 = -1;
    #   if ($str =~ /.*([01])[34]*(3$|4.)/) {
    #     $p2 = $+[0];
    #   }
    #   ### $str
    #   ### $p1
    #   ### $p2
    #   if ($p1 >= 0 && $p1 >= $p2) {
    #     ### yes, left ...
    #     return -1;
    #   }
    #   return 0;
    # }
  }
}

{
  # segment enclosure run lengths which occur

  # Epred

  # EpredR
  # ones=3  379..381 [112001]..[112010]

  my @seen_ones;
  my @seen_zeros;
  my $prev = 0;
  my $ones = 0;
  my $ones_start = 0;
  my $ones_end = 0;
  my $zeros = 0;
  my $zeros_start = 0;
  my $zeros_end = 0;
  for (my $n = 0; ; $n++) {
    my $enc = n_to_encRL_by_digits_low($n);
    # my $enc = EpredR($n);
    if ($prev) {
      if ($enc) {
        # 1,1
        $ones++;
        $ones_end = $n;
      } else {
        # 1,0
        unless ($seen_ones[$ones]++ > 3) {
          my $start5 = cnv($ones_start,10,5);
          my $end5 = cnv($ones_end,10,5);
          print "n=$n  ones=$ones  $ones_start..$ones_end [$start5]..[$end5]\n";
        }
        $zeros = 1;
        $zeros_start = $zeros_end = $n;
      }
    } else {
      if ($enc) {
        # 0,1
        unless ($seen_zeros[$zeros]++ > 3) {
          my $start5 = cnv($zeros_start,10,5);
          my $end5 = cnv($zeros_end,10,5);
          print "n=$n        zeros=$zeros  $zeros_start..$zeros_end [$start5]..[$end5]\n";
        }
        $ones = 1;
        $ones_start = $ones_end = $n;
      } else {
        # 0,0
        $zeros++;
        $zeros_end = $n;
      }
    }
    $prev = $enc;
  }
  print "\n";
  exit 0;

  sub EpredR {
    my ($n) = @_;
    return n_to_encRL_by_digits_low($n) == 1 ? 1 : 0;
  }
}

{
  # boundary segment numbers samples

  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $n (0 .. 25) {
    next unless $path->_UNDOCUMENTED__n_segment_is_right_boundary($n);
    print "$n, ";
  }
  print "\n";

  foreach my $n (0 .. 72) {
    next unless $path->_UNDOCUMENTED__n_segment_is_left_boundary($n);
    print "$n, ";
  }
  print "\n";
  exit 0;
}

{
  # right and left enc tables and reversals graphs

  foreach my $side ('right','left') {
    my $enc_table = ($side eq 'right' ? \@right_enc_table : \@left_enc_table);
    my $table = {};
    foreach my $state (0 .. $#$enc_table) {
      my $row = $enc_table->[$state];
      foreach my $digit (0 .. $#$row) {
        my $new_state = $row->[$digit];
        $table->{$state}->{$digit} = $new_state;
      }
    }

    require MyFSM;
    my $fsm = MyFSM->new
      (table => $table,
       initial => 0,
       accepting => { 3 => $side },
      );
    $fsm->show;

    $fsm = $fsm->reverse;
    $fsm->show;
    $fsm->view;

    $fsm->simplify;
    $fsm->show;
    $fsm->view;
  }
  exit 0;
}
{
  # segment enclosure combination states -- low to high by mappings graph

  my @identity = (0 .. 14);
  my @digit_map;
  foreach my $digit (0..4) {
    foreach my $state (0 .. 14) {
      $digit_map[$digit][$state] = $enc_table[$state][$digit];
    }
    my $decided = state_map_decided($digit_map[$digit]);
    print "digit map $digit  ",join(',',@{$digit_map[$digit]}),"  $decided\n";

    foreach my $state (0 .. 14) {
      print enc_state_to_RL($digit_map[$digit][$state]),",";
    }
    print "\n";
  }


  sub state_map_product {
    my ($s, $t) = @_;
    ### state_map_product ...
    ### $s
    ### $t
    my @new;
    foreach my $state (0 .. 14) {
      my $n = $state;
      $n = $s->[$n];
      $n = $t->[$n];
      $new[$state] = $n;
    }
    return \@new;
  }
  sub state_map_decided {
    my ($s) = @_;
    my %seen;
    foreach my $state (@$s) {
      $seen{enc_state_to_RL($state)} = 1;
    }
    my $any_non  = $seen{0};
    my $any_left = $seen{-1};
    my $any_right = $seen{1};
    return ($any_non && ! $any_left && ! $any_right ? " =non"
            : $any_left && ! $any_non && ! $any_right ? " =left"
            : $any_right && ! $any_non && ! $any_left ? " =right"
            : '');
  }
  sub state_map_string {
    my ($s) = @_;
    return join(',',@$s) . state_map_decided($s);

  }
  sub state_map_uniqs {
    my ($s) = @_;
    my %s;
    foreach my $state (@$s) {
      $s{$state} = 1;
    }
    delete $s{0};
    return join(',', sort {$a<=>$b} keys %s);
  }

  my %seen;
  my @m;
  sub state_map_to_number {
    my ($s) = @_;
    my $key = state_map_string($s);
    my $num = $seen{$key};
    if (! defined $num) {
      push @m, $s;
      $num = $#m;
      $seen{$key} = $num;
    }
    return $num;
  }

  my @pending = (\@identity);
  while (@pending) {
    my $s = shift @pending;
    my $num = state_map_to_number($s);
    my $key = state_map_string($s);
    my $uniq = state_map_uniqs($s);
    my $decided = state_map_decided($s);
    printf "m%-2d %-37s [%s]\n", $num, $key, $uniq;
    next if $decided;

    foreach my $digit (0..4) {
      my $t = state_map_product($digit_map[$digit], $s);
      my $tkey = state_map_string($t);
      my $already = '';
      if (defined $seen{$tkey}) {
        $already = '  (seen)';
      } else {
        push @pending, $t;
      }
      my $tnum = state_map_to_number($t);
      print "  $digit -> m$tnum$already\n";
    }
    # print "pending $#pending\n";
  }
  print "total maps ",scalar(@m),"\n";

  # Graph
  require Graph::Easy;
  my $graph = Graph::Easy->new(timeout => 9999);
  my $node_number = 0;
  my %nodes;
  my %edges;
  foreach my $i (0 .. $#m) {
    my $s = $m[$i];
    my $sdecided = state_map_decided($s);
    next if $sdecided;
    my $sname = state_map_name($i);
    foreach my $digit (0..4) {
      my $t = state_map_product($digit_map[$digit], $s);
      my $tdecided = state_map_decided($t);
      if ($tdecided) {
        my $node = ($nodes{"m$i.$tdecided"}
                    ||= do {
                      my $node = $graph->add_node($node_number++);
                      $node->set_attribute('label',$tdecided);
                      $node
                    });
        my $edge = ($edges{"m$i.$tdecided"}
                    ||= do {
                      $graph->add_edge($sname, $node, '');
                    });
        my $label = $edge->attribute('label');
        $label = join(',',non_empty($label),$digit);
        $edge->set_attribute('label',$label);
      } else {
        my $ti = state_map_to_number($t);
        my $tname = state_map_name($ti);
        my $edge = ($edges{"m$i.m$ti"}
                    ||= do {
                      $graph->add_edge($sname, $tname, '');
                    });
        my $label = $edge->attribute('label');
        $label = join(',',non_empty($label),$digit);
        $edge->set_attribute('label',$label);
      }
    }
  }
  sub non_empty {
    my ($str) = @_;
    if (! defined $str || $str eq '') { return (); }
    return $str;
  }
  sub state_map_name {
    my ($i) = @_;
    my $s = $m[$i];
    my $decided = state_map_decided($s);
    my $zero = ($decided ? '' : "[$s->[0]] ");
    return "m$i -- $zero".state_map_uniqs($s).$decided;
  }

  {
    my $graphviz = $graph->as_graphviz();
    require File::Slurp;
    File::Slurp::write_file('/tmp/terdragon-enc.dot', $graphviz);
    require IPC::Run;
    IPC::Run::run(['dot','-Tps','/tmp/terdragon-enc.dot'],'>','/tmp/terdragon-enc.ps');
    # IPC::Run::run(['dotty','/tmp/terdragon-enc.dot'],'&');
    IPC::Run::run(['gv','--scale=.4','/tmp/terdragon-enc.ps'],'&');
  }

  #print $graph->as_ascii;

  print "type ",$graph->type,"\n";
  print "is_simple: ",$graph->is_simple ? "yes\n" : "no\n";
  print "roots: ",join(' ', map{$_->name} $graph->source_nodes), "\n";

  exit 0;
}
{
  # base 5 delete all 1s
  #
  my @values;
  foreach my $n (5 .. 60) {
    my $n5 = cnv($n,10,5);
    $n5 =~ s/1//g;
    $n5 =~ s/^0+//;
    if ($n5 eq '') { $n5 = 0; }
    my $value = cnv($n5,5,10);
    push @values, $value;
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}
{
  # base 5 with no 1s = A023725
  #
  my @values;
  for (my $n = 5; @values < 40; $n++) {
    my $n5 = cnv($n,10,5);
    next if $n5 =~ /1/;
    push @values, $n;
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}



{
  # left enclosure by high to low
  #
  # *====>*   s0
  #
  # |
  # |
  # *====>*   s1
  #
  # *-----
  # |
  # |
  # *====>*   s2
  #
  # *-----*
  # |     |
  # |     |
  # *====>*   s3

  BEGIN {
    @left_enc_table = ([0,1,2,0,0],  # s0
                       [1,2,3,1,0],
                       [1,2,3,2,0],  # s2
                       [1,2,3,3,3]);
  }
  sub n_to_left_enc_state_high_to_low {
    my ($n) = @_;
    my $state = 0;
    foreach my $digit (reverse digit_split_lowtohigh($n,5)) {
      $state = $left_enc_table[$state][$digit];
    }
    return $state;
  }

  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $n (0 .. 5**8) {
    my $state_path = n_to_left_enc_state_by_path($path,$n);
    my $state_HtoL = n_to_left_enc_state_high_to_low($n);
    if ($state_path != $state_HtoL) {
      die;
    }
  }
  exit 0;

  sub n_to_left_enc_state_by_path {
    my ($path, $n) = @_;
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);
    my ($x2,$y2) = ($x+$dx, $y+$dy);
    my ($dxL,$dyL) = xy_rotate_plus90($dx,$dy);
    return
      (path_xyxy_is_traversed_either($path, $x,$y, $x+$dxL,$y+$dyL, $n) ? 1 : 0)
      + (path_xyxy_is_traversed_either($path, $x2,$y2, $x2+$dxL,$y2+$dyL, $n) ? 1 : 0)
      + (path_xyxy_is_traversed_either($path, $x+$dxL,$y+$dyL, $x2+$dxL,$y2+$dyL, $n) ? 1 : 0);
  }
}

{
  # right enclosure by high to low
  #
  # *====>*   s0
  #
  # *====>*   s1
  # |
  # |
  #
  # *====>*   s2
  # |
  # |
  # *-----
  #
  # *====>*   s3
  # |     |
  # |     |
  # *-----*

  BEGIN {
    @right_enc_table = ([0,0,0,1,2],  # s0
                        [3,1,0,1,2],
                        [3,2,0,1,2],  # s2
                        [3,3,1,2,3]);
  }
  sub n_to_right_enc_state_high_to_low {
    my ($n) = @_;
    my $state = 0;
    foreach my $digit (reverse digit_split_lowtohigh($n,5)) {
      $state = $right_enc_table[$state][$digit];
    }
    return $state;
  }

  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $n (0 .. 5**8) {
    my $state_path = n_to_right_enc_state_by_path($path,$n);
    my $state_HtoL = n_to_right_enc_state_high_to_low($n);
    if ($state_path != $state_HtoL) {
      die;
    }
  }
  exit 0;

  sub n_to_right_enc_state_by_path {
    my ($path, $n) = @_;
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);
    my ($x2,$y2) = ($x+$dx, $y+$dy);
    my ($dxR,$dyR) = xy_rotate_minus90($dx,$dy);
    return
      (path_xyxy_is_traversed_either($path, $x,$y, $x+$dxR,$y+$dyR, $n) ? 1 : 0)
      + (path_xyxy_is_traversed_either($path, $x2,$y2, $x2+$dxR,$y2+$dyR, $n) ? 1 : 0)
      + (path_xyxy_is_traversed_either($path, $x+$dxR,$y+$dyR, $x2+$dxR,$y2+$dyR, $n) ? 1 : 0);
  }
}

{
  my $fsm = {};
  foreach my $state (0 .. $#enc_table) {
    my $row = $enc_table[$state];
    foreach my $digit (0 .. $#$row) {
      my $new_state = $row->[$digit];
      $fsm->{$state}->{$digit} = $new_state;
    }
  }

  require MyFSM;
  print "states: ",join(',',sort {$a cmp $b} MyFSM::fsm_states($fsm)),"\n";
  print "edges:  ",join(',',sort {$a cmp $b} MyFSM::fsm_edges($fsm)),"\n";
  MyFSM::fsm_print($fsm);
  print "\n";

  $fsm = MyFSM::fsm_reverse($fsm);
  print "reversal\n";
  MyFSM::fsm_print($fsm);
  print "\n";
  # MyFSM::fsm_replace($fsm, '10,10,10,10,10,10,10,10','right');

 STATE: foreach my $state (MyFSM::fsm_states($fsm)) {
    next STATE unless $state =~ /^(right|(12|13|14)(,(12|13|14))*)$/;
    foreach my $edge (MyFSM::fsm_edges($fsm)) {
      my $new_state = $fsm->{$state}->{$edge};
      next STATE unless $new_state =~ /^(right|(12|13|14)(,(12|13|14))*)$/;
    }
    MyFSM::fsm_replace($fsm, $state,'right');
  }
  MyFSM::fsm_print($fsm);
  print "\n";

 STATE: foreach my $state (MyFSM::fsm_states($fsm)) {
    next STATE unless $state =~ /^(left|(9|10|11)(,(9|10|11))*)$/;
    foreach my $edge (MyFSM::fsm_edges($fsm)) {
      my $new_state = $fsm->{$state}->{$edge};
      next STATE unless $new_state =~ /^(left|(9|10|11)(,(9|10|11))*)$/;
    }
    MyFSM::fsm_replace($fsm, $state,'left');
  }
  MyFSM::fsm_print($fsm);
  print "\n";

 STATE: foreach my $state (MyFSM::fsm_states($fsm)) {
    next STATE unless $state =~ /^(non|([0-8])(,([0-8]))*)$/;
    foreach my $edge (MyFSM::fsm_edges($fsm)) {
      my $new_state = $fsm->{$state}->{$edge};
      next STATE unless $new_state =~ /^(non|([0-8])(,([0-8]))*)$/;
    }
    MyFSM::fsm_replace($fsm, $state,'non');
  }
  MyFSM::fsm_print($fsm);
  print "\n";

  MyFSM::simplify($fsm);
  exit 0;
}


{
  # boundary hanging squares = 1/2*(3^(k-1) - 1) each side
  # R[k] = 2* R[k-1] +    U[k-1] + 2
  # U[k] =    R[k-1] + 2* U[k-1] + 2
  # R[0] = 0    U[0] = 0
  # R[1] = 0    U[1] = 0
  # R[2] = 1    U[2] = 0
  # R[3] = 4    U[3] = 3
  # R[4] = 13   U[4] =
  #
  # my(r=1,u=0);for(k=0,8,print(r," ",u); my(r2=2*r+u+2,u2=r+2*u+2); r=r2;u=u2)
  #
  # R[k] = 1/2*(3^(k-1) - 1)
  # U[k] = 1/2*(3^(k-1) - 3) = R[k] - 1
  # vector(8,k,1/2*(3^(k-1) - 3))
  #
  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $level (0 .. 8) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($level);

    my $count = 0;
    my %seen;
  N: foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my $mx = $x;
      my $my = $y;
      foreach (1 .. 4) {
        if ($x <= $mx && $y <= $my) {
          $mx = $x;
          $my = $y;
        }
        if (! xyxy_is_boundary($x,$y, $x+$dx,$y+$dy, $level, $n_hi)) {
          next N;
        }
        $x += $dx;
        $y += $dy;
        ($dx,$dy) = xy_rotate_plus90($dx,$dy);
      }
      next if $seen{"$mx,$my"}++;

      if ($level <= 2) {
        print "  hanging n=$n at $mx,$my\n"
      }
      $count++;
    }
    print "k=$level  $count\n";
    push @values, $count;
  }

  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values);
  exit 0;

  sub xyxy_is_boundary {
    my ($x1,$y1, $x2,$y2, $level, $n_limit) = @_;
    my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2) // return 0;
    if ($n >= $n_limit) { return 0; }
    return $path->_UNDOCUMENTED__n_segment_is_right_boundary($n)
      || $path->_UNDOCUMENTED__n_segment_is_left_boundary($n, $level);
  }
}



{
  # enclosure by high to low
  BEGIN {
    @enc_table = (
              [0,1,4,2,6], [1,4,9,3,6], [12,3,4,2,6], [13,5,9,3,6], [1,4,9,5,6],
              [13,5,9,5,6], [12,7,4,2,6], [13,8,9,3,6], [13,8,9,5,6],
              [1,4,9,10,11], [13,5,9,10,11], [13,8,9,10,11],
              [12,13,5,6,12], [13,14,10,7,12], [13,14,10,8,12],
             );
  }
  sub n_to_enc_state_high_to_low {
    my ($n) = @_;
    my $state = 0;
    foreach my $digit (reverse digit_split_lowtohigh($n,5)) {
      $state = $enc_table[$state][$digit];
    }
    return $state;
  }
  sub n_to_encRL_high_to_low {
    my ($n) = @_;
    return enc_state_to_RL(n_to_enc_state_high_to_low($n));
  }
  sub enc_state_to_RL {
    my ($state) = @_;
    if ($state >= 12) { return 1; }
    if ($state >= 9) { return -1; }
    return 0;
  }

  my $path = Math::PlanePath::R5DragonCurve->new;
  foreach my $n (0 .. 5**5) {
    my $state_path = n_to_enc_state_by_path($path,$n);
    my $state_HtoL = n_to_enc_state_high_to_low($n);
    if ($state_path != $state_HtoL) {
      die;
    }
  }
  exit 0;
}

{
  # enclosure state configurations
  my @config;
  BEGIN {
    @config = (0b000000,  # s0
               0b100000,
               0b000100,
               0b100100,  # s3
               0b110000,
               0b110100,
               0b000110,  # s6
               0b100110,
               0b110110,

               0b111000,  # s9
               0b111100,
               0b111110,

               0b000111,  # s12
               0b100111,
               0b110111,
              );
  }
  sub n_to_enc_state_by_path {
    my ($path, $n) = @_;
    my ($x,$y) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);

    my ($dxR,$dyR) = xy_rotate_minus90($dx,$dy);
    my ($dxL,$dyL) = xy_rotate_plus90($dx,$dy);

    my $bits = 0;
    if (path_xyxy_is_traversed_either($path, $x,$y, $x+$dxL,$y+$dyL, $n)) {
      $bits |= 0b100000;
    }
    if (path_xyxy_is_traversed_either($path, $x,$y, $x+$dxR,$y+$dyR, $n)) {
      $bits |= 0b000100;
    }

    my $x2 = $x + $dx;
    my $y2 = $y + $dy;
    if (path_xyxy_is_traversed_either($path, $x2,$y2, $x2+$dxL,$y2+$dyL, $n)) {
      $bits |= 0b001000;
    }
    if (path_xyxy_is_traversed_either($path, $x2,$y2, $x2+$dxR,$y2+$dyR, $n)) {
      $bits |= 0b000001;
    }

    if (path_xyxy_is_traversed_either($path, $x+$dxL,$y+$dyL, $x2+$dxL,$y2+$dyL, $n)) {
      $bits |= 0b010000;
    }
    if (path_xyxy_is_traversed_either($path, $x+$dxR,$y+$dyR, $x2+$dxR,$y2+$dyR, $n)) {
      $bits |= 0b000010;
    }

    return enc_bits_to_state($bits)
      // die "oops n=$n bits=" . sprintf('%06b',$bits);
  }
  sub enc_bits_to_state {
    my ($bits) = @_;;
    foreach my $state (0 .. $#config) {
      if ($config[$state] == $bits) {
        return $state;
      }
    }
    return undef;
  }


  my @expand = ('a00ddd',  # 0
                '1a0def',
                '11af00',  # 2
                'abc1f0',
                'ccc11f',  # 4
               );
  my %letter_to_bitpos = (a => 5,
                          b => 4,
                          c => 3,
                          d => 2,
                          e => 1,
                          f => 0);
  my $bitpos_max = 5;
  my @table;
  foreach my $state (0 .. $#config) {
    my $bits = $config[$state];
    foreach my $digit (0 .. 4) {
      my $new_bits = 0;
      my $expand = $expand[$digit];
      foreach my $i (0 .. $bitpos_max) {
        my $letter = substr($expand,$i,1);
        my $bit;
        if ($letter eq '0') {
          $bit = 0;
        } elsif ($letter eq '1') {
          $bit = 1;
        } else {
          my $bitpos = $letter_to_bitpos{$letter} // die;
          $bit = $bits & (1 << $bitpos) ? 1 : 0;
        }
        $new_bits |= ($bit << (5-$i));
      }
      my $new_state = enc_bits_to_state($new_bits);
      $new_state
        // die "state=$state digit=$digit new_bits not found: " . sprintf('%06b',$new_bits);
      $table[$state][$digit] = $new_state;
    }
  }

  {
    my @done;
    my @reached;
    my @pending = (0);
    while (@pending) {
      my $state = pop @pending;
      next if $done[$state]++;
      foreach my $digit (0 .. 4) {
        my $new_state = $table[$state][$digit];
        $reached[$new_state] = 1;
        push @pending, $new_state;
      }
    }
    foreach my $state (0 .. $#config) {
      if (! $reached[$state]) {
        print "state $state not reached\n";
      }
    }
  }

  require MyPrintwrap;
  MyPrintwrap::printwrap_indent("   ");
  foreach my $state (0 .. $#table) {
    if ($state == 9 || $state == 12) { MyPrintwrap::printwrap_newline(); }
    my $str = ' [' . join(',',@{$table[$state]}) . '],';
    MyPrintwrap::printwrap($str);
  }
  print "\n";

  require MyPrintwrap;
  MyPrintwrap::printwrap_indent("   ");
  foreach my $state (0 .. $#table) {
    my $semi = $state == $#table ? '' : ';';
    print "% Test-Pari-DEFINE                       ",
      join(',',@{$table[$state]}), $semi,
      "\n";
  }
  print "\n";


  # Graph
  require Graph::Easy;
  my $graph = Graph::Easy->new(timeout => 9999);
  foreach my $state (0 .. $#config) {
    my $node = $graph->add_node("s$state");
  }
  my %edges;
  foreach my $state (0 .. $#config) {
    foreach my $digit (0 .. 4) {
      my $new_state = $table[$state][$digit];
      my $edge = ($edges{"s$state.s$new_state"}
                  ||= do {
                    $graph->add_edge("s$state", "s$new_state", '');
                  });
      my $label = $edge->attribute('label');
      $label = join(',',non_empty($label),$digit);
      $edge->set_attribute('label',$label);
    }
  }
  {
    my $graphviz = $graph->as_graphviz();
    require File::Slurp;
    File::Slurp::write_file('/tmp/terdragon-enc.dot', $graphviz);
    require IPC::Run;
    IPC::Run::run(['dot','-Tps','/tmp/terdragon-enc.dot'],'>','/tmp/terdragon-enc.ps');
    # IPC::Run::run(['dotty','/tmp/terdragon-enc.dot'],'&');
    IPC::Run::run(['gv','--scale=.5','/tmp/terdragon-enc.ps'],'&');
  }
  print "type ",$graph->type,"\n";
  print "is_simple: ",$graph->is_simple ? "yes\n" : "no\n";
  print "roots: ",join(' ', map{$_->name} $graph->source_nodes), "\n";

  exit 0;
}


{
  # boundary segments in direction
  #
  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  my (@v0,@v1,@v2,@v3,@total);
  foreach my $level (0 .. 8) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($level);
    my @count = (0,0,0,0);
    foreach my $n ($n_lo .. $n_hi-1) {
      next unless $path->_UNDOCUMENTED__n_segment_is_right_boundary($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my $dir4 = dxdy_to_dir4($dx,$dy);
      $count[$dir4]++;
    }
    print "level=$level  $count[0] $count[1] $count[2] $count[3]\n";

    push @v0, $count[0];
    push @v1, $count[1];
    push @v2, $count[2];
    push @v3, $count[3];
    push @total, sum(@count);
  }

  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@v0, name=>'v0');
  Math::OEIS::Grep->search(array => \@v1, name=>'v1');
  Math::OEIS::Grep->search(array => \@v2, name=>'v2');
  Math::OEIS::Grep->search(array => \@v3, name=>'v3');
  Math::OEIS::Grep->search(array => \@total, name=>'total');
  exit 0;
}

{
  # middle segment direction
  # (-i)^(d+2k) = (-i)^d*(-i)^2k
  # (-i)^2k = (-1)^k
  # (-I)^d*(-b)^k + (I)^d*(-conj(b))^k
  # (-1)^(2k) = 1
  #
  # 1,1,5,37,153,761,3965
  # 0,2,8,26,160,802,3848
  # half 0, 1, 4, 13, 80, 401, 1924
  #
  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $level (0 .. 8) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($level);
    my $n_mid = ($n_hi-1)/2;
    my ($dx,$dy) = $path->n_to_dxdy($n_mid);
    print "k=$level  $dx,$dy\n"
  }
  exit 0;
}

{
  # moment of inertia, vertices
  #
  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (0 .. 6) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($k);
    my ($ex,$ey) = $path->n_to_xy($n_hi);
    my $mx = $ex/2;
    my $my = $ey/2;

    my $I = 0;
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      my $rx = $mx - $x;
      my $ry = $my - $y;
      my $r_squared = $rx*$rx + $ry*$ry;
      $I += $r_squared;  # with mass=1
    }

    # $I /= 5**$k;
    # $I *= 2;
    # $I -= 1;
    # $I /= 10;
    my $f = vertices_unit_masses($k);
    print "k=$k  I=$I   f=$f\n";
    push @values, $I;
  }
  print join(', ', map{(2*$_).'/2'} @values),"\n";
  print join(', ', map{"\\frac{".(2*$_)."}{2}"} @values),"\n";

  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;

  sub vertices_unit_masses {
    my ($k) = @_;
    return (3*25**$k + 2*5**$k + 5)/20;
  }
}

{
  # moment of inertia, midpoint of segments
  #
  # A125833 = "333..33" base 5
  #         = 3*(5^k-1)/4
  # with unit masses at segment midpoints
  #   IS(k) = 3/20 * 5^k * (5^k - 1)
  #         = 3/20 * 5^k*5^k - 3/20 * 5^k
  # shape with I=3/20 in each segment would be 3/20*25^k
  #
  # line = 1/12
  # barbell = 2* 1/2*(1/2)^2 = 1/4
  # 1/12*m + (1-m)*1/4 = 3/20
  #
  my $path = Math::PlanePath::R5DragonCurve->new;
  my @values;
  foreach my $k (0 .. 9) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($k);
    my ($ex,$ey) = $path->n_to_xy($n_hi);
    my $mx = $ex/2;
    my $my = $ey/2;

    my $I = 0;
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      $x += $dx/2;
      $y += $dy/2;  # segment midpoint
      my $rx = $mx - $x;
      my $ry = $my - $y;
      my $r_squared = $rx*$rx + $ry*$ry;
      $I += $r_squared;  # with mass=1
    }

    #    $I /= 5**($k-1);
    my $f = midpoint_unit_masses($k);
    print "k=$k  I=$I   f=$f\n";
    push @values, $I;
  }
  print join(', ', @values),"\n";
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;

  sub midpoint_unit_masses {
    my ($k) = @_;
    return 5**$k * (5**$k - 1) * 3 /20;
  }
}


{
  # boundary segments in direction
  #
  #
  unshift @INC, "$FindBin::Bin/../../path/xt";
  require MyOEIS;
  require Math::OEIS::Grep;
  my $path = Math::PlanePath::R5DragonCurve->new;
  my (@v0,@v1,@v2,@v3,@total);
  foreach my $level (0 .. 8) {
    (my $n_lo, my $n_hi) = $path->level_to_n_range($level);

    my $points = MyOEIS::path_boundary_points ($path, $n_hi,
                                               side => 'right');
    my @count = (0,0,0,0);
    foreach my $i (1 .. $#$points) {
      my $p1 = $points->[$i-1];
      my $p2 = $points->[$i];
      my ($x1,$y1) = @$p1;
      my ($x2,$y2) = @$p2;
      my $dx = $x2-$x1;
      my $dy = $y2-$y1;
      my $dir = dxdy_to_dir4($dx,$dy);
      $count[$dir]++;
    }
    print "level=$level  $count[0] $count[1] $count[2] $count[3]\n";

    push @v0, $count[0];
    push @v1, $count[1];
    push @v2, $count[2];
    push @v3, $count[3];
    push @total, sum(@count);
  }

  Math::OEIS::Grep->search(array => \@v0, name=>'v0');
  Math::OEIS::Grep->search(array => \@v1, name=>'v1');
  Math::OEIS::Grep->search(array => \@v2, name=>'v2');
  Math::OEIS::Grep->search(array => \@v3, name=>'v3');
  Math::OEIS::Grep->search(array => \@total, name=>'total');
  exit 0;
}
{
  # rectangular extents relative to endpoints

  require Geometry::AffineTransform;
  foreach my $k (0 .. 10) {
    my $path = Math::PlanePath::R5DragonCurve->new;
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
    my $affine = Geometry::AffineTransform->new;
    my $scale = 1 / hypot($y_hi,$x_hi);
    $affine->scale($scale, $scale);
    $affine->rotate(- rad2deg(atan2($y_hi,$x_hi)));

    my $x_min = 0;
    my $y_min = 0;
    my $x_max = 0;
    my $y_max = 0;
    foreach my $n ($n_lo .. $n_hi) {
      my ($x,$y) = $path->n_to_xy($n);
      ($x,$y) = $affine->transform($x,$y);
      $x_min = min($x_min, $x);
      $x_max = max($x_max, $x);
      $y_min = min($y_min, $y);
      $y_max = max($y_max, $y);
    }
    print "k=$k  $x_min $x_max   $y_min $y_max\n";
  }
  exit 0;
}

#------------------------------------------------------------------------------
# functions

sub LowestNonFour {
  my ($n) = @_;
  my $low;
  do {
    $low = _divrem_mutate($n,5);
  } while ($low == 4);
  return $low;
}

sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

END {
  foreach my $h (@ipc) {
    $h->finish;
  }
}
