/* Copyright 2017 Kevin Ryde

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3, or (at your option) any later
   version.

   This file is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this file.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Some R5 dragon graph stuff using Nauty. */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/naugroup.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* GP-DEFINE  read("r5dragon-defines.gp"); */


/* ------------------------------------------------------------------------- */
/* Misc Diagnostics */

void
writeautom (int *p, int n)
{
  int i;
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}

void
show_edges (graph *g, int m,int n)
{
  set *row = (set*) g;
  int from, to;
  printf("edges");
  for (from=0; from < n-1; from++, row+=m) {
    for (to=from+1; to < n; to++) {
      if (ISELEMENT(row,to))
        printf(" %d-%d", from,to);
    }
  }
  putchar('\n');
}


/* ------------------------------------------------------------------------- */
/* Misc Helpers */

/* Change (x,y) to (-y,x) which is a rotation by +90 degrees. */
#define R5DRAGON_XY_ROTATE_PLUS90(x,y)            \
  do {                                          \
    int R5DRAGON_XY_ROTATE_PLUS90__tmp = (x);     \
    (x) = -(y);                                 \
    (y) = R5DRAGON_XY_ROTATE_PLUS90__tmp;         \
  } while (0)

/* Change (x,y) to (y,-x) which is a rotation by -90 degrees. */
#define R5DRAGON_XY_ROTATE_MINUS90(x,y)           \
  do {                                          \
    int R5DRAGON_XY_ROTATE_MINUS90__tmp = (x);    \
    (x) = (y);                                  \
    (y) = - R5DRAGON_XY_ROTATE_MINUS90__tmp;      \
  } while (0)

static int dir4_to_dx[] = { 1,0,-1,0 };
static int dir4_to_dy[] = { 0,1,0,-1 };

/* ------------------------------------------------------------------------- */
/* R5 Dragon Sequence Values */

void
r5dragon_overflow (void)
{
  fprintf(stderr, "r5dragon overflow\n");
  abort();
}

/* r5dragon_RQ(k) returns the number of right boundary squares on R5 dragon
   curve k.
*/
int
r5dragon_RQ (int k)
{
  double d = pow(3,k);
  int RQ = d;
  if (RQ < d) { r5dragon_overflow(); }
  return RQ;
}

/* r5dragon_R(k) returns the right boundary length of R5 dragon curve k.
 */
int
r5dragon_R (int k)
{
  double d = 2*pow(3,k) - 1;
  int R = d;
  if (R < d) { r5dragon_overflow(); }
  return R;
}

/* r5dragon_P(k) returns the number distinct points visited by R5 dragon k.
   P(k) = 1/2*(5^k + 1) + 3^k;
 */
int
r5dragon_P (int k)
{
  double d = (pow(5,k) + 1)/2 + pow(3,k);
  int P = d;
  if (P < d) { r5dragon_overflow(); }
  return P;
}

/* r5quad_QP(k) returns the number of distinct points visited by R5 quad k.
 */
int
r5quad_TP (int k)
{
  abort();
}

/* QAdiameter(k) returns the diameter of R5 quad area tree k.
   QAdiameter(k) = 3^k - 1
 */
int
r5quad_QAdiameter(int k)
{
  double d = pow(3,k) - 1;
  int QAdiameter = d;
  if (QAdiameter < d) { r5dragon_overflow(); }
  return QAdiameter;
}

/* r5quad_QAW(k) returns the Wiener index of R5 quad area
   tree k.
   QAW(k) = 1/35 * 5^k * (10*15^k - 7*5^k - 3);
*/
unsigned long long
r5quad_QAW(int k)
{
  double d = (10*pow(15,k) - 7*pow(5,k) - 3) * pow(5,k) / 35.0;
  unsigned long long W = d;
  if (W < d) { r5dragon_overflow(); }
  return W;
}
/* r5quad_QATW(k) returns the terminal Wiener index of R5 quad area tree k.
   QATW(k) = 8/175*75^k + 38/125*25^k + 69/175*5^k - 1 for k>=1
           = (40*75^k + 266*25^k + 345*5^k)/875 - 1 for k>=1
   lcm([175,125,175]) == 875
*/
unsigned long long
r5quad_QATW(int k)
{
  double d;
  unsigned long long W;
  if (k == 0) return 0;
  d = (40*pow(75,k) + 266*pow(25,k) + 345*pow(5,k)) / 875 - 1;
  W = d;
  if (W < d) { r5dragon_overflow(); }
  return W;
}

/* JAdiameter(k) returns the diameter of R5 quad area tree k.
   JAdiameter(k) = 3^k - 1
 */
int
r5quad_JAdiameter(int k)
{
  if (k==0) return 0;
  double d = (pow(3,k) - 3)/2;
  int JAdiameter = d;
  if (JAdiameter < d) { r5dragon_overflow(); }
  return JAdiameter;
}

/* return the lowest non-0 base-5 digit of n */
int
LowestNon0 (int n)
{
  assert (n >= 1);
  while (n) {
    int r = n % 5;
    if (r) return(r);
    n /= 5;
  }
  return 0;
}
/* return the lowest non-4 base-5 digit of n */
int
LowestNon4 (int n)
{
  assert (n >= 0);
  for (;;) {
    int d = n % 5;
    if (d != 4) return d;
    n /= 5;
  }
}

/* Return the turn in  R5 dragon curve at point n, being +1 left -1 right.
   The first point with a turn is n=1.  */
int
r5dragon_turn (int n)
{
  return (LowestNon0(n) <= 2 ? 1 : -1);
}
/* Return true if the R5 dragon curve turns right at point n.
   The first point with a turn is n=1.  */
int
r5dragon_turn_is_right (int n)
{
  return (r5dragon_turn(n) > 0);
}

/* Return true if the R5 dragon curve turns right after segment n.
   The first segment is n=0.  */
int
r5dragon_next_turn_is_right (int n)
{
  return (LowestNon4(n) >= 2);
}

/* ------------------------------------------------------------------------- */
/* R5 Dragon Curve Graph */

/* r5dragon_graph(g, m,n, k) sets graph g to dragon curve k.
   n should be the number of vertices r5dragon_P(k).
   g must have enough room for size m,n.

   The current implementation follows the curve by its turn sequence and
   does a linear search for preceding x,y to notice a point revisited.  This
   is not very efficient but is easy and is fast enough since any subsequent
   real work on the graph takes much longer.
*/
int r5dragon_graph_INTERNAL_last_v;
static void
r5dragon_graph_INTERNAL(graph *g, int m,int n, int start,int end)
{
  struct { int x; int y; } *v_to_xy;
  int i;
  int from = -1;
  int dx = 1;
  int dy = 0;
  int xy_upto = 0;
  int x = 0;
  int y = 0;
  DEBUG(printf("r5dragon_graph_INTERNAL n=%d  start=%d end=%d\n", n,start,end));

  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("r5dragon_graph()");

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  for (i = 1; i <= end; i++) {
    int to;
    x += dx;
    y += dy;
    if (i >= start) {
      DEBUG(printf(" i=%d xy= %d,%d\n", i, x,y));
      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == x && v_to_xy[to].y == y) {
          goto GOT_TO;
        }
      }
      DEBUG(printf(" new location\n"));
      to = xy_upto++;
      if (to >= n) {
        fprintf(stderr, "r5dragon_graph() oops, at i=%d (end=%d) to=%d xy=%d,%d exceeds size n=%d\n",
                i, end, to, x,y, n);
        abort();
      }
      v_to_xy[to].x = x;
      v_to_xy[to].y = y;
    GOT_TO:
      if (from >= 0) {
        DEBUG(printf(" edge %d to %d at %d,%d\n", from,to, x,y));
        ADDONEEDGE(g, from, to, m);
      }
      from = to;
      r5dragon_graph_INTERNAL_last_v = to;
    } else {
      DEBUG(printf(" skip i=%d xy= %d,%d\n", i, x,y));
    }

    if (r5dragon_turn(i) > 0)
      R5DRAGON_XY_ROTATE_PLUS90(dx,dy);       /* left */
    else
      R5DRAGON_XY_ROTATE_MINUS90(dx,dy);      /* right */
  }

  if (xy_upto != n) {
    fprintf(stderr, "oops, dragon graph start=%d end=%d xy_upto=%d but vertices n=%d\n",
            start,end, xy_upto, n);
    abort();
  }
  FREES(v_to_xy);
}
void
r5dragon_graph(graph *g, int m, int n, int k)
{
  r5dragon_graph_INTERNAL(g, m,n, 0, 1<<k);
}

/* ------------------------------------------------------------------------- */
/* R5 Quad Area Tree */

/* Modify vpar by star replacement.
   1 -> 1,2,3,4,5    5*n-4,...5*n */
void
star_replacement_vpar (int *vpar, int n)
{
  DYNALLSTAT(int,arm,arm_sz);
  DYNALLOC1(int,arm,arm_sz,n+1,"try");
  memset(arm, '\0', (n+1)*sizeof(int));
  int v;
  for (v = n; v >= 1; v--) {
    int p = vpar[v];
    DEBUG(printf("  rep v=%d p=%d\n", v, p));
    assert (arm[p] >= -2);
    vpar[5*v-4] = 5*p + (arm[p]--);
    vpar[5*v-3] = 5*v-4;    /* 5*v-3 middle */
    vpar[5*v-2] = 5*v-3;
    vpar[5*v-1] = 5*v-3;
    vpar[5*v]   = 5*v-3;
  }
}

/* r5quad_area_tree_n(k) returns the number of vertices in R5 quad area
   tree k.  This is simply 5^k.

   r5quad_area_tree(g, m,n) sets graph g to R5 quad area tree k.
   n should be the number of vertices as per r5quad_area_tree_n(k).
   g should be already allocated to size m,n.

   The R5 quad area tree is quite sparse, since each vertex degree <= 3,
   so usually r5quad_area_tree_sg() below is better, when sparsegraph
   suits.
*/
int
r5quad_area_tree_n(int k)
{
  double d = pow(5,k);
  int n = d;
  if (n < d) { r5dragon_overflow(); }
  return n;
}
void
r5quad_area_tree(graph *g, int m, int n, int k)
{
  int m_in,n_in;
  graph *g_in, *g_out;
  DYNALLSTAT(graph, gtmp,gtmp_sz);

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  DEBUG(printf("r5quad_area_tree() %d,%d %d\n", m,n,k));

  DYNALLOC2 (graph, gtmp,gtmp_sz,m,n,"malloc");
  g_in  = (k & 1 ? gtmp : g);   /* k=0 is g */
  g_out = (k & 1 ? g : gtmp);
  n_in = 1;
  m_in = SETWORDSNEEDED(n_in);
  EMPTYGRAPH(g_in, m_in,n_in);

  while (k-- > 0) {
    int n_out = 5*n_in;
    int m_out = SETWORDSNEEDED(n_out);
    DEBUG(printf("at k=%d  m_in,n_in=%d,%d  g_in=%p\n", k, m_in,n_in, g_in));
    /* show_edges(g_in,m_in,n_in); */

    nautyextra_star_replacement(g_in, m_in,n_in, g_out,m_out, 5);
    { graph *g_swap = g_in; g_in = g_out; g_out = g_swap; }
    n_in = n_out;
    m_in = m_out;
  }

  if (g_in != g) { abort(); }
  DYNFREE(gtmp,gtmp_sz);
}

void
r5quad_area_tree_sg (sparsegraph *sg, int k)
{
  SG_DECL(sgtmp);
  sparsegraph *sg_in, *sg_out;

  SG_INIT(sgtmp);
  sg_in  = (k & 1 ? &sgtmp : sg);   /* k=0 is g */
  sg_out = (k & 1 ? sg : &sgtmp);

  /* k=0 single vertex no edges */
  SG_ALLOC(*sg_in, 1, 0, "r5quad_area_tree_sg()");
  sg_in->nv = 1;
  sg_in->nde = 0;
  sg_in->d[0] = 0;
  sg_in->v[0] = 0;

  while (k-- > 0) {
    DEBUG(printf("at k=%d  n_in=%d\n", k, sg_in->nv));
    /* show_edges_sg(sg_in); */

    nautyextra_star_replacement_sg(sg_in, sg_out, 5);
    { sparsegraph *g_swap = sg_in; sg_in = sg_out; sg_out = g_swap; }
  }

  if (sg_in != sg) { abort(); }
  SG_FREE(sgtmp);
}

void
r5quad_area_tree_vpar (int *vpar, int k)
{
  int n = 1;
  vpar[1] = 0;  /* k=0 tree single vertex */
  int i;
  for (i = 1; i <= k; i++) {
    star_replacement_vpar(vpar,n);
    n *= 5;
  }
}

/* ------------------------------------------------------------------------- */
/* R5 Join Area Tree */

/* r5dragon_join_area_tree_n(k) returns the number of vertices in R5 join
   area tree k.  This is (5^k - 1)/4.

   r5dragon_join_area_tree_vpar(vpar,k) sets vpar[1..n] to R5 join area
   tree k.
*/
int
r5dragon_join_area_tree_n(int k)
{
  double d = (pow(5,k) - 1)/4;
  int n = d;
  if (n < d) { r5dragon_overflow(); }
  return n;
}
void
r5dragon_join_area_tree_vpar(int *vpar, int k)
{
  int n = 0;  /* k=0 empty */
  int i;
  for (i = 1; i <= k; i++) {
    star_replacement_vpar(vpar,n);
    n *= 5;
    int p = n;
    vpar[++n] = p;
  }
}

/*
  row start
  not in OEIS: 1,2,3,6,7,8,11,14,17,26
  not in OEIS: 0,1,2,5,6,7,10,13,16,25

  row end
  not in OEIS: 1,2,5,6,7,10,13,16,25
  not in OEIS: 0,1,4,5,6, 9,12,15,24


                  1
                  |
                  2
                / | \
               3  4  5
                  |
                  6
                  |
                  7
                / | \
              8   9  10
             /    |    \
           11    12    13
           /      |      \
         14      15       16
        /|\      /|\      /|\
     17 18 19 20 21 22 23 24 25



                    1
                    |
                    2
                 /  |  \
               10  11  12
                    |
                   20
                    |
                   21
                 /  |  \
               22  100  101
              /     |     \
            102    110    111
           /        |       \
         112       120     121
        /|\        /|\     /|\
    122 200 201  202 210 211  212 220 221



                    0                           ternary 0-based
                    |
                    1
                 /  |  \
               02  10  11
                    |
                   12
                    |
                   21
                 /  |  \
               22  100  101
              /     |     \
            102    110    111
           /        |       \
         112       120     121
        /|\        /|\     /|\
    122 200 201  202 210 211  212 220 221


                0
                |
                1
             /  |  \
            2   3   4
                |
               10
                |
               11
             /  |  \
           12  13   14
          /     |     \
        20     21     22
         /      |       \
      23       24      30
       /|\     /|\     /|\
    31 32 33  34 40 41  42 43 44
 */



/* ------------------------------------------------------------------------- */

int
r5quad_turn_graph_n(int k)
{
  return pow(5,k) - 1;
}
void
r5quad_turn_graph(graph *g, int m,int n, int k)
{
  struct { int x; int y; } *v_to_xy;
  int xy_upto = 0;
  DEBUG(printf("r5quad_turn_graph()  n=%d k=%d\n", n,k));

  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("r5quad_turn_graph()");

  nauty_check(WORDSIZE, m,n, NAUTYVERSIONID);
  EMPTYGRAPH(g, m,n);

  int i_lo = 4*pow(5,k);
  int i_hi = 8*pow(5,k);
  DEBUG(printf(" i_lo %d i_hi %d\n", i_lo,i_hi));
  int x = 0;
  int y = 0;
  int dx = 1;
  int dy = 0;
  int from = -1;

  int i;
  for (i = i_lo; i < i_hi-1; ) {
    int to;
    x += dx;
    y += dy;

    i++;
    int turn = r5dragon_turn(i);
    DEBUG(printf(" at i=%d turn=%d xy= %d,%d  dxdy=%d,%d\n",
                 i, turn, x,y, dx,dy));
    if (turn > 0) {
      R5DRAGON_XY_ROTATE_PLUS90(dx,dy);       /* left */
      continue;
    } else {
      R5DRAGON_XY_ROTATE_MINUS90(dx,dy);      /* right */
    }

    for (from = 0; from < xy_upto; from++) {
      if (v_to_xy[from].x == x && v_to_xy[from].y == y) {
        goto GOT_FROM;
      }
    }
    from = xy_upto++;
    DEBUG(printf("  new location from=%d\n", from));
    assert (from < n);
    v_to_xy[from].x = x;
    v_to_xy[from].y = y;

    int d;
    for (d = 0; d < 4; d++) {
      int odx = dir4_to_dx[d];
      int ody = dir4_to_dy[d];
      int to_x = x + odx;
      int to_y = y + ody;

      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == to_x && v_to_xy[to].y == to_y) {
          DEBUG(printf("  edge %d (%d,%d) to %d (%d,%d)\n",
                       from,x,y, to,to_x,to_y));
          ADDONEEDGE(g, from, to, m);
          break;
        }
      }
    }

  GOT_FROM: ;
  }

  DEBUG(printf(" end xy=%d,%d\n", x,y));
  assert (x==0);
  assert (y==1);
  assert (xy_upto == n);
  FREES(v_to_xy);
}

void
r5quad_turn_graph_sg(sparsegraph *sg, int k)
{
  struct { int x; int y; } *v_to_xy;
  int xy_upto = 0;
  DEBUG(printf("r5quad_turn_graph_sg()  k=%d\n", k));

  int n = r5quad_turn_graph_n(k);  /* num vertices */
  v_to_xy = ALLOCS(n, sizeof(v_to_xy[0]));
  if (v_to_xy == NULL) alloc_error("r5quad_turn_graph_sg()");

  sg->nv = 0;
  sg->nde = 0;
  SG_ALLOC(*sg, n, 6*pow(5,k-1)-2, "r5quad_turn_graph_sg");

  int i_lo = 4*pow(5,k);
  int i_hi = 8*pow(5,k);
  DEBUG(printf(" i_lo %d i_hi %d\n", i_lo,i_hi));
  int x = 0;
  int y = 0;
  int dx = 1;
  int dy = 0;
  int from = -1;

  int i;
  for (i = i_lo; i < i_hi-1; ) {
    int to;
    x += dx;
    y += dy;

    i++;
    int turn = r5dragon_turn(i);
    DEBUG(printf(" at i=%d turn=%d xy= %d,%d  dxdy=%d,%d\n",
                 i, turn, x,y, dx,dy));
    if (turn > 0) {
      R5DRAGON_XY_ROTATE_PLUS90(dx,dy);       /* left */
      continue;
    } else {
      R5DRAGON_XY_ROTATE_MINUS90(dx,dy);      /* right */
    }

    for (from = 0; from < xy_upto; from++) {
      if (v_to_xy[from].x == x && v_to_xy[from].y == y) {
        goto GOT_FROM;
      }
    }
    from = xy_upto++;
    DEBUG(printf("  new location from=%d\n", from));
    assert (from < n);
    v_to_xy[from].x = x;
    v_to_xy[from].y = y;

    int d;
    for (d = 0; d < 4; d++) {
      int odx = dir4_to_dx[d];
      int ody = dir4_to_dy[d];
      int to_x = x + odx;
      int to_y = y + ody;

      for (to = 0; to < xy_upto; to++) {
        if (v_to_xy[to].x == to_x && v_to_xy[to].y == to_y) {
          DEBUG(printf("  edge %d (%d,%d) to %d (%d,%d)\n",
                       from,x,y, to,to_x,to_y));
          nautyextra_add_edge_sg(sg, from,to);
          break;
        }
      }
    }

  GOT_FROM: ;
  }

  DEBUG(printf(" end xy=%d,%d\n", x,y));
  assert (x==0);
  assert (y==1);
  assert (xy_upto == n);
  FREES(v_to_xy);
}

/* ------------------------------------------------------------------------- */

void
try_r5quad_area_tree_dense_sparse_isomorphic (void)
{
  DYNALLSTAT(graph, qa,qa_sz);
  SG_DECL(qa_sg);
  SG_DECL(qa_sg_from_dense);
  int n,m,k;

  SG_INIT(qa_sg);
  SG_INIT(qa_sg_from_dense);

  for (k = 0; k < 8; k++) {
    printf("k=%d\n", k);

    n = r5quad_area_tree_n(k);
    m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,qa,qa_sz,m,n,"malloc");
    r5quad_area_tree(qa, m,n, k);

    printf("  made as sg\n");
    r5quad_area_tree_sg(&qa_sg, k);
    nautyextra_print_sg(&qa_sg);

    printf("  from dense\n");
    nauty_to_sg(qa, &qa_sg_from_dense, m,n);
    nautyextra_print_sg(&qa_sg_from_dense);

    if (nautyextra_is_multiedged_sg(&qa_sg)) {
      printf("  multiedged\n");
      abort();
    }
    if (nautyextra_loopcount_sg(&qa_sg)) {
      printf("  selfloop\n");
      abort();
    }
    if (nautyextra_is_undirected_sg(&qa_sg)) {
      printf("  undirected\n");
      abort();
    }

    printf("  check isomorphic\n");
    if (nautyextra_is_isomorphic_sg(&qa_sg, &qa_sg_from_dense)) {
      printf("  same\n");
    }

    printf("\n");
  }
  SG_FREE(qa_sg);
  SG_FREE(qa_sg_from_dense);
}

void
try_r5quad_area_tree (void)
{
  static DEFAULTOPTIONS_SPARSEGRAPH(options);
  statsblk stats;

  DYNALLSTAT(int, lab,    lab_sz);
  DYNALLSTAT(int, ptn,    ptn_sz);
  DYNALLSTAT(int, orbits, orbits_sz);
  DYNALLSTAT(int, map,    map_sz);
  SG_DECL(ta);
  SG_DECL(tacanon);
  SG_DECL(tacanon2);
  int num_vertices;

  int k = 1;

  options.getcanon = TRUE;
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;

  printf("make tree\n");
  r5quad_area_tree_sg(&ta, k);
  nautyextra_print_sg(&ta);

  num_vertices = ta.nv;
  DYNALLOC1(int, lab,    lab_sz,    num_vertices, "try");
  DYNALLOC1(int, ptn,    ptn_sz,    num_vertices, "try");
  DYNALLOC1(int, orbits, orbits_sz, num_vertices, "try");
  DYNALLOC1(int, map,    map_sz,    num_vertices, "try");

  printf("\nlabel ta\n");
  sparsenauty(&ta,lab,ptn,orbits,&options,&stats,&tacanon);
  sortlists_sg(&ta);

  {
    int num_edges = ta.nde;
    int m = SETWORDSNEEDED(num_vertices);
    printf("fcanon m=%d\n", m);
    /* int loops = loopcount(&ta, m, num_vertices); */
    /* printf("loops %d\n", loops); */
    SG_ALLOC(tacanon2, num_vertices, num_edges, "tacanon2");
    fcanonise_inv_sg(&ta, m, num_vertices, &tacanon2,
                     NULL,  /* fmt colours */
                     NULL,  /* distances_sg,  */ /* invarproc */
                     1,     /* mininvarlevel */
                     1,     /* maxinvarlevel */
                     3,     /* invararg */
                     0);    /* digraph */
    sortlists_sg(&tacanon2);
  }

  putgraph_sg(stdout, &ta, 79);  /* line length */

  printf("\n");
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&ta);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&tacanon);
  writeline(stdout,SPARSE6_HEADER);
  writes6_sg(stdout,&tacanon2);

  printf("\n");
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&ta);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&tacanon);
  writeline(stdout,GRAPH6_HEADER);
  writeg6_sg(stdout,&tacanon2);

  printf("aresame  %d %d",
         aresame_sg(&ta,&tacanon),
         aresame_sg(&tacanon,&tacanon2));


  printf ("\nGenerators\n");
  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  printf ("num orbits    %d\n", stats.numorbits);
  printf ("errstatus     %d\n", stats.errstatus);
  printf ("numnodes      %lu\n", stats.numnodes);
  printf ("numbadleaves  %lu\n", stats.numbadleaves);
  printf ("maxlevel      %d\n", stats.maxlevel);

  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);
      allgroup(group,writeautom);
    }
  }
}

void
try_QAdiameter (void)
{
  DYNALLSTAT(graph, qa, qa_sz);
  int k;

  for (k=0; k < 7; k++) {
    int QAdiameter = r5quad_QAdiameter(k);
    int n = r5quad_area_tree_n(k);
    int m = SETWORDSNEEDED(n);
    int radius, diameter;

    DYNALLOC2(graph,qa,qa_sz,m,n,"malloc");
    r5quad_area_tree(qa, m,n, k);
    diamstats(qa, m,n, &radius, &diameter);

    printf("k=%d n=%d  radius %d  diameter %d func=%d\n",
           k,n, radius, diameter, QAdiameter);
    /* show_edges (qa,m,n); */
  }
}

void
try_r5quad_area_tree_sg_Wiener (void)
{
  SG_DECL(qa);
  int k;

  SG_INIT(qa);
  for (k=0; k < 7; k++) {
    unsigned long long QAW_calc;
    unsigned long long QAW_formula = r5quad_QAW(k);

    r5quad_area_tree_sg(&qa, k);
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index_sg(&qa, 0, &info);
    QAW_calc = info.W;

    printf("k=%d  Wiener index sg %Ld %Ld\n", k, QAW_calc, QAW_formula);
  }
  SG_FREE(qa);
}

void
try_r5quad_area_tree_terminal_Wiener (void)
{
  SG_DECL(qa);
  int k;

  SG_INIT(qa);
  for (k=0; k < 9; k++) {
    unsigned long long QATW_calc;
    unsigned long long QATW_formula = r5quad_QATW(k);

    r5quad_area_tree_sg(&qa, k);
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index_sg(&qa, 1, &info);
    QATW_calc = info.W;

    printf("k=%d  Wiener index sg %Ld %Ld\n", k, QATW_calc, QATW_formula);
  }
  SG_FREE(qa);
}

void
try_r5quad_area_tree_Wiener (void)
{
  DYNALLSTAT(graph, qa, qa_sz);
  int k;

  for (k=0; k < 7; k++) {
    int n = r5quad_area_tree_n(k);
    int m = SETWORDSNEEDED(n);
    unsigned long long QAW_calc;
    unsigned long long QAW_formula = r5quad_QAW(k);
    struct nautyextra_Wiener_info info;

    DYNALLOC2(graph,qa,qa_sz,m,n,"malloc");
    r5quad_area_tree(qa, m,n, k);
    nautyextra_Wiener_index(qa, m,n, 0, &info);
    QAW_calc = info.W;

    printf("k=%d  Wiener index %Ld %Ld\n", k, QAW_calc, QAW_formula);
  }
}

void
try_sequences (void)
{
  int n, k;
  printf("turn right n=1: ");
  for (n=1; n < 10; n++) {
    printf(" %d,", r5dragon_turn_is_right(n));
  }
  printf("\n");

  for (k=0; k < 10; k++) {
    int R = r5dragon_R(k);
    int P = r5dragon_P(k);
    int TP = r5quad_TP(k);
    printf("k=%d  R=%d P=%d TP=%d\n", k, R, P, TP);
  }
}
void
test (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);

#if 0
  {
    static const int vpar[] = {
      0,
      0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 2, 2, 20, 20,
      22, 20, 23, 5, 26, 26, 28, 26, 29, 5, 8, 8, 11, 11, 45, 39, 46,
      46, 46, 40, 58, 55, 40, 40, 36, 44, 35, 45, 45, 37, 49, 50, 50,
      50, 38, 54, 55, 55, 61, 61, 61, 63, 61, 64, 14, 76, 70, 77, 77,
      77, 71, 89, 86, 71, 71, 67, 75, 66, 76, 76, 68, 80, 81, 81, 81,
      69, 85, 86, 86, 92, 92, 92, 94, 92, 95, 14
    };
    int n = sizeof(vpar)/sizeof(vpar[0]) - 1;
    unsigned long long  got = nautyextra_indsets_count_vpar(vpar,n);
    unsigned long long  want = 3069893676516396957361ULL;
    printf("%Lu\n", got);
  }
#endif
  {
    static const int turn_samples[] = {
      1,1,-1,-1,1,1,1,-1,-1,1,1,1,-1,-1,-1,1,1,-1,-1,-1,1,1,-1,-1
    };
    int i;
    for (i=0; i < sizeof(turn_samples)/sizeof(turn_samples[0]); i++) {
      {
        int want = turn_samples[i];
        int n = i+1;
        int got = r5dragon_turn(n);
        assert(got==want);
      }
      {
        int want = turn_samples[i] < 0;
        int got = r5dragon_next_turn_is_right(i);
        assert(got==want);
      }
    }
  }

  {
    /* r5quad_area_tree_sg() vs r5quad_area_tree_vpar() */
    int k;
    for (k=0; k <= 4; k++) {
      r5quad_area_tree_sg(&sg, k);

      int n = r5quad_area_tree_n(k);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      r5quad_area_tree_vpar(vpar,k);
      /* nautyextra_print_vpar(vpar,n); */
      /* nautyextra_view_vpar (vpar,n); */

      if (! nautyextra_is_isomorphic_sg_vpar(&sg,vpar,n)) {
        abort();
      }
    }
  }

  {
    /* r5dragon_join_area_tree_vpar() */
    int k;
    for (k=0; k <= 6; k++) {
      int n = r5dragon_join_area_tree_n(k);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      r5dragon_join_area_tree_vpar(vpar,k);

      {
        /* JAdiameter */
        int by_formula = r5quad_JAdiameter(k);
        int by_vpar    = nautyextra_diameter_vpar(vpar,n);
        if (by_vpar == -1) by_vpar = 0;
        if (by_formula != by_vpar) {
          printf ("r5dragon_join_area_tree_vpar() k=%d\n", k);
          printf ("  JAdiameter by_formula %d by_vpar %d\n",
                  by_formula, by_vpar);
          nautyextra_view_vpar (vpar,n);
          abort();
        }
      }
    }
  }
}

void
try_r5dragon_diameter (void)
{
  DYNALLSTAT(graph, g,g_sz);
  int k;

  for (k=0; k < 7; k++) {
    int n = r5quad_area_tree_n(k);
    int m = SETWORDSNEEDED(n);
    int radius, diameter;

    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    r5quad_area_tree(g, m,n, k);
    diamstats(g, m,n, &radius, &diameter);

    printf("k=%d n=%d  radius %d  diameter %d\n",
           k,n, radius, diameter);
    /* show_edges (g,m,n); */
  }
}

/*
  diameter 2,10,30,86,250,738
  radius   2, 6,16,44,126,370
  W        8,1340,116700,9021000,679480000,50936462500
  ecc S    2, 9, 28, 83, 246, 733,  2192
*/
void
try_r5quad_turn_diameter (void)
{
  DYNALLSTAT(graph, g,g_sz);
  int k;
  for (k=1; k < 8; k++) {
    int n = r5quad_turn_graph_n(k);

    printf("k=%d n=%d\n", k,n);

    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    r5quad_turn_graph(g, m,n, k);

    /* if (k <= 2) nautyextra_view (g,m,n); */

    int radius=0, diameter=0;
    diamstats(g, m,n, &radius, &diameter);
    printf("    radius %d  diameter %d\n", radius, diameter);

    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index(g, m,n, 0, &info);
    unsigned long long QTW_calc = info.W;

    printf("    W %llu\n", QTW_calc);
  }
}
void
try_r5quad_turn_diameter_sg (void)
{
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=1; k < 8; k++) {
    int n = r5quad_turn_graph_n(k);

    printf("k=%d n=%d\n", k,n);
    r5quad_turn_graph_sg(&sg, k);
    assert (! (sg.nde&1));

    printf("    edges %d  ecc start %d\n",
           sg.nde/2, nautyextra_eccentricity_of_vertex_sg(&sg,0,NULL));

    /* if (k <= 2) nautyextra_view_sg (sg); */

    /* int radius=0, diameter=0; */
    /* diamstats_sg(sg, &radius, &diameter); */
    /* printf("    radius %d  diameter %d\n", radius, diameter); */

    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index_sg(&sg, 0, &info);
    unsigned long long QTW_calc = info.W;
    printf("    W %llu\n", QTW_calc);

    DYNALLSTAT(int,dist,dist_sz);
    DYNALLOC1(int,dist,dist_sz,n,"try");
    distvals(&sg, 0, dist, n);
    unsigned long long wC = 0;
    int i;
    for (i=0; i<n; i++) wC += dist[i];
    printf("    wC %llu\n", wC);
  }
  SG_FREE(sg);
}

/* not in OEIS: 2, 17, 540497
 */
void
try_r5quad_area_tree_indsets (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k < 8; k++) {
    int n = r5quad_area_tree_n(k);

    printf("k=%d n=%d\n", k,n);
    r5quad_area_tree_sg(&sg, k);

    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    nautyextra_sg_to_vpar(&sg,vpar,0);
    unsigned long long count = nautyextra_indsets_count_vpar(vpar,n);

    printf("    indsets %Lu\n", count);
  }
  SG_FREE(sg);
}

/* not in OEIS: 2, 96, 31747188880
 */
void
try_r5quad_area_tree_indomsets (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k < 8; k++) {
    int n = r5quad_area_tree_n(k);

    printf("k=%d n=%d\n", k,n);
    r5quad_area_tree_sg(&sg, k);

    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    nautyextra_sg_to_vpar(&sg,vpar,0);
    unsigned long long count_indomsets
      = nautyextra_indomsets_count_vpar(vpar,n);

    printf("    indomsets %Lu\n", count_indomsets);
  }
  SG_FREE(sg);
}

/*
  1  5  25  125
  0, 4, 17,  81, 401, 2001, 10001, 50001, 250001, 1250001, 6250001

  initial exceptions [0, 4]
  QAindnum(k) = if(k==0,0, k==1,4, 16*5^(k-2) + 1);
  16/25 of the 5^k vertices

  4*17 == 68
  81-68 == 13 \\ 17-4
*/
void
try_r5quad_area_tree_indnum (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k < 12; k++) {
    int n = r5quad_area_tree_n(k);

    printf("k=%d n=%d\n", k,n);
    r5quad_area_tree_sg(&sg, k);

    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    nautyextra_sg_to_vpar(&sg,vpar,0);
    int indnum = nautyextra_indnum_vpar(vpar,n);

    printf("    indnum %d\n", indnum);
  }
  SG_FREE(sg);
}
/*
  k= 0  1  2    3         4
     1  5 25  125
     1, 1, 1,  16, 16777216
  GP-Test  16777216 == 2^24
*/
void
try_r5quad_area_tree_indnum_count (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k < 12; k++) {
    int n = r5quad_area_tree_n(k);

    printf("k=%d n=%d\n", k,n);
    r5quad_area_tree_sg(&sg, k);

    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    nautyextra_sg_to_vpar(&sg,vpar,0);
    struct nautyextra_indnum_and_count info;
    nautyextra_indnum_and_count_vpar(vpar,n, &info);

    printf("    indnum %d count %Lu\n", info.indnum, info.count);
  }
  SG_FREE(sg);
}

/* k      1  2   3
   n      4 24 124
   indnum 2 12
   count  2 17

   17 = 2 even outers either way * 2 opposite parity
*/
void
try_r5quad_turn_indnum (void)
{
  nautyextra_want_progress = 1;
  DYNALLSTAT(graph, g,g_sz);
  int k;
  for (k=2; k <= 2; k++) {
    int n = r5quad_turn_graph_n(k);
    printf("k=%d n=%d\n", k,n);
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g,g_sz,m,n,"malloc");
    r5quad_turn_graph(g, m,n, k);

    struct nautyextra_indnum_and_count info;
    nautyextra_indnum_and_count(g, m,n, &info);

    printf("    indnum %d count %Lu\n", info.indnum, info.count);
  }
}

/*
  vertices 1  5  25  125
  domnum   1  1   5   25 125 625 3125 15625 78125 390625 1953125
  A000351 5^k
  count    1  1   1
*/
void
try_r5quad_area_tree_domnum (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  SG_DECL(sg);
  SG_INIT(sg);
  int k;
  for (k=0; k < 10; k++) {
    r5quad_area_tree_sg(&sg, k);
    int n = sg.nv;
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    nautyextra_sg_to_vpar(&sg,vpar,0);

    if (1) {
      struct nautyextra_domnum_and_count  info;
      nautyextra_domnum_and_count_sg(&sg, &info);
      printf("n=%d  domnum %d  count %Lu\n", n, info.domnum, info.count);
    }
    if (0) {
      int domnum = nautyextra_domnum_vpar(vpar,n);
      printf("%d,", domnum);
    }
    if (0) {
      /* matchnum 0,1,8,44,224,1124,5624,28124,140624,703124,
         .        0 1 2
         A198768 = 9*5^k - 1
      */
      int matchnum = nautyextra_matchnum_vpar(vpar,n);
      printf("%d,", matchnum);
    }
    if (0) {
      /* count 0,4,513,1699990476912,
         .     0 1   2             3
         factor(1699990476912)
         GP-Test  1699990476912 == 2^4 * 13^4 * 3^12 * 7
         1699990476912^(1/5^3)     \\ 1.252 689 888 56
      */
      struct nautyextra_matchnum_and_count info;
      nautyextra_matchnum_and_count_vpar(vpar,n, &info);
      printf("%Lu,", info.count);
    }
  }
  printf("\n");
  SG_FREE(sg);
}

void
try_r5dragon_join_area_tree (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  int k;
  for (k=0; k <= 10; k++) {
    int n = r5dragon_join_area_tree_n(k);
    /* printf ("k=%d n=%d\n", k, n); */
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    r5dragon_join_area_tree_vpar(vpar,k);
    /* nautyextra_print_vpar(vpar,n); */
    /* nautyextra_view_vpar (vpar,n); */

    if (0) {
      unsigned long long count = nautyextra_indsets_count_vpar(vpar,n);
      printf("    indsets %Lu\n", count);
    }
    if (0) {
      unsigned long long count = nautyextra_indomsets_count_vpar(vpar,n);
      printf("    maximal indsets %Lu\n", count);
    }

    if (0) {
      int indnum = nautyextra_indnum_vpar(vpar,n);
      printf("    indnum %d\n", indnum);
    }
    if (0) {
      struct nautyextra_indnum_and_count info;
      nautyextra_indnum_and_count_vpar(vpar,n, &info);
      printf("    indnum %d count %Lu\n", info.indnum, info.count);
    }
    if (0) {
      int domnum = nautyextra_domnum_vpar(vpar,n);
      printf("    domnum %d\n", domnum);
    }
    /* n=0  domnum 0  count 1 */
    /* n=1  domnum 1  count 1 */
    /* n=6  domnum 2  count 2 */
    /* n=31  domnum 7  count 2 */
    if (1) {
      SG_DECL(sg);
      SG_INIT(sg);
      nautyextra_vpar_to_sg (vpar,n, &sg, 0);  /* undirected */
      struct nautyextra_domnum_and_count  info;
      nautyextra_domnum_and_count_sg(&sg, &info);
      printf("n=%d  domnum %d  count %Lu\n", n, info.domnum, info.count);
      SG_FREE(sg);
    }
    if (0) {
      /* matchnum 0,0,2,11,56,281,1406,7031,35156,175781,878906
         .        0 1 2
         A198769 = 1/4*(9*5^k - 1)
      */
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      int matchnum = nautyextra_matchnum_vpar(vpar,n);
      printf("%d,", matchnum);
    }

    if (0) {
      unsigned long long count = nautyextra_minimal_domsets_count_vpar(vpar,n);
      printf("    minimal domsets %Lu\n", count);
    }
  }
}

/* minimal domsets
   1, 2, 306, 26215298260416
   GP-Test  306 == 2 * 3^2 * 17
   GP-Test  26215298260416 == 2^6 * 3^4 * 563 * 8982173
*/
void
try_r5dragon_quad_area_tree (void)
{
  DYNALLSTAT(int,vpar,vpar_sz);
  int k;
  for (k=0; k <= 10; k++) {
    int n = r5quad_area_tree_n(k);
    /* printf ("k=%d n=%d\n", k, n); */
    DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
    r5quad_area_tree_vpar(vpar,k);
    /* nautyextra_print_vpar(vpar,n); */
    /* nautyextra_view_vpar (vpar,n); */

    if (0) {
      unsigned long long count = nautyextra_indsets_count_vpar(vpar,n);
      printf("    indsets %Lu\n", count);
    }
    if (1) {
      /* factor(31747188880)
         not in OEIS: 1, 2, 96, 31747188880
       */
      unsigned long long count = nautyextra_indomsets_count_vpar(vpar,n);
      printf("    maximal indsets %Lu\n", count);
    }

    if (0) {
      int indnum = nautyextra_indnum_vpar(vpar,n);
      printf("    indnum %d\n", indnum);
    }
    if (0) {
      struct nautyextra_indnum_and_count info;
      nautyextra_indnum_and_count_vpar(vpar,n, &info);
      printf("    indnum %d count %Lu\n", info.indnum, info.count);
    }
    if (0) {
      int domnum = nautyextra_domnum_vpar(vpar,n);
      printf("    domnum %d\n", domnum);
    }
    if (0) {
      /* matchnum 0,0,2,11,56,281,1406,7031,35156,175781,878906
         .        0 1 2
         A198769 = 1/4*(9*5^k - 1)
      */
      DYNALLOC1(int,vpar,vpar_sz,n+1,"try");
      int matchnum = nautyextra_matchnum_vpar(vpar,n);
      printf("%d,", matchnum);
    }

    if (0) {
      unsigned long long count = nautyextra_minimal_domsets_count_vpar(vpar,n);
      printf("    minimal domsets %Lu\n", count);
    }
  }
}

/* ------------------------------------------------------------------------- */

int
main (void)
{
  setbuf(stdout,NULL);
  setbuf(stderr,NULL);
  test ();

  try_r5quad_turn_diameter_sg();
  return 0;
  try_r5dragon_quad_area_tree();
  try_r5dragon_join_area_tree();
  try_r5quad_area_tree_domnum();
  try_r5quad_area_tree_indomsets();
  try_r5quad_turn_indnum();
  try_r5quad_area_tree_indnum();
  try_r5quad_area_tree_indnum_count();
  try_r5quad_area_tree_indsets();
  try_r5quad_area_tree_dense_sparse_isomorphic();
  try_r5quad_area_tree();
  try_r5quad_area_tree_terminal_Wiener();
  try_r5quad_area_tree_sg_Wiener();
  try_r5quad_area_tree_Wiener();
  try_QAdiameter();
  try_sequences();
  try_r5dragon_diameter();
  try_r5quad_turn_diameter();
  return 0;
}
