#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use Math::PlanePath::R5DragonCurve;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyR5Dragon ':all';

plan tests => 418;

# uncomment this to run the ### lines
# use Smart::Comments;


#------------------------------------------------------------------------------
# dir4_flat()

{
  my $bad = 0;
  my $n_hi = 5**4;
 DIR: foreach my $want_dir4 (0 .. 3) {
    my $f = dir4_flat($want_dir4);
    foreach my $n (0 .. $n_hi) {
      my $str = Math::BaseCnv::cnv($n,  10,5);
      foreach my $zeros (0 .. 2) {
        my $str = ('0' x $zeros) . $str;

        my $by_f   = $f->contains($str) ? 1 : 0;
        my $by_dir = (dir($n) % 4) == $want_dir4 ? 1 : 0;
        if ($by_f != $by_dir) {
          print "$f->{name} at N=$n $str got flat=$by_f want dir=$by_dir\n";
          last DIR if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "dir4_flat() vs dir()");
}

foreach my $want_dir4 (0 .. 3) {
  my $f = dir4_flat($want_dir4);
  my $rev = $f->MyFLAT::reverse($f);
  ok ($f->equals($rev),
      "dir4_flat() same on reverse digits");
}


#------------------------------------------------------------------------------
# hullstouch

foreach my $digit (0 .. 4) {
  # $hullstouch_segs->expand($digit)->print;
  ok( $hullstouch_segs->is_subset( $hullstouch_segs->expand($digit) ),
      "$hullstouch_segs->{name} expand digit=$digit");
}

&fBpred_hullstouch_init(verbose=>2);
is ($fRpred_hullstouch_flat->num_states, 18);
is ($fLpred_hullstouch_flat->num_states, 18);
is ($fBpred_hullstouch_flat->num_states, 18);

is (scalar(@{$fBpred_hullstouch_state_enclosed_lists[0]}), 5);
is (scalar(@{$fBpred_hullstouch_state_enclosed_lists[1]}), 5);
is (scalar(@{$fBpred_hullstouch_state_enclosed_lists[2]}), 1); 

{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullstouch_state_table,
     name => 'fNonRpred_hullstouch',
     accepting_list => $fBpred_hullstouch_state_enclosed_lists[0],
    );
  is ($f->MyFLAT::num_accepting, 5);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 0);
  is (scalar(@{$info{'states'}}), 0,
      'hullstouch no further eventually accepting');
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullstouch_state_table,
     name => 'fNonRpred_hullstouch',
     accepting_list => $fBpred_hullstouch_state_enclosed_lists[1],
    );
  is ($f->MyFLAT::num_accepting, 5);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 0);
  is (scalar(@{$info{'states'}}), 0,
      'hullstouch no further eventually accepting');
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullstouch_state_table,
     name => 'fNonRpred_hullstouch',
     accepting_list => $fBpred_hullstouch_state_enclosed_lists[2],
    );
  is ($f->MyFLAT::num_accepting, 1);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 0);
  is (scalar(@{$info{'states'}}), 0,
      'hullstouch no further eventually accepting');
}

is ($fNonRpred_hullstouch_flat->MyFLAT::num_accepting, 5);
is ($fNonLpred_hullstouch_flat->MyFLAT::num_accepting, 5);
is ($fNonBpred_hullstouch_flat->MyFLAT::num_accepting, 1);

is ($fRpred_hullstouch_flat->num_states, 18);  # not minimized
is (NonBpred_k_flat()->num_states, 18);        # is same as Bpred
is (Bpred_k_flat()->num_states, 18);
is (Bpred_k_flat()->MyFLAT::minimize->num_states, 18);

ok ($fRpred_hullstouch_flat->equals(Rpred_flat()),
    'fRpred hullstouch = Rpred');
ok ($fLpred_hullstouch_flat->equals(Lpred_k_flat()));
ok ($fBpred_hullstouch_flat->equals(Bpred_k_flat()));

ok ($fNonRpred_hullstouch_flat->equals(NonRpred_flat()));
ok ($fNonLpred_hullstouch_flat->equals(NonLpred_k_flat()));
ok ($fNonBpred_hullstouch_flat->equals(NonBpred_k_flat()));


#------------------------------------------------------------------------------
# EpredR,EpredL three consecutive

{
  my $f = EpredR_flat();
  for (1 .. 2) {
    $f = EpredR_flat()->intersect($f->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize(verbose=>1);
  }
  ok ($f->equals(EpredR3_flat()));

  # MyFLAT::FLAT_show_breadth($f, 8);
  my $max_k = 12;
  is(join(',', MyFLAT::FLAT_count_contains($f,$max_k)),
     join(',', map {$_<2 ? 0 : 2*AR($_-2)} 0 .. $max_k));
}
{
  my $f = EpredL_flat();
  for (1 .. 2) {
    $f = EpredL_flat()->intersect($f->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize(verbose=>1);
  }
  ok ($f->equals(EpredL3_flat()));

  # MyFLAT::FLAT_show_breadth($f, 8);
  my $max_k = 12;
  is(join(',', MyFLAT::FLAT_count_contains($f,$max_k)),
     join(',', map {$_<2 ? 0 : AR($_-1)} 0 .. $max_k));
}


#------------------------------------------------------------------------------
# EpredSix runs

{
  my $f = Epred_flat();
  for (1 .. 5) {
    $f = Epred_flat()->intersect($f->MyFLAT::digits_increment(add=>-1))
      ->MyFLAT::minimize(verbose=>1);
  }
  ok ($f->equals(EpredSix_flat()));

  # MyFLAT::FLAT_show_breadth($f, 8);
  my $max_k = 12;
  is(join(',', MyFLAT::FLAT_count_contains($f,$max_k)),
     join(',', map {$_<2 ? 0 : AR($_-2)} 0 .. $max_k));


  # n+8 is 7 terms n to n+6 inclusive
  $f = Epred_flat()->intersect($f->MyFLAT::digits_increment(add=>-1));
  ok($f->is_empty);
}


#------------------------------------------------------------------------------

{
  # intersection both fRpred and fLpred is only f=0 and f=1
  fBpred_hullsaround_init();

  # eventually non-accepting
  my $got = $fRpred_hullsaround_flat->intersect($fLpred_hullsaround_flat);
  $got = $got->complement;
  $got = $got->MyFLAT::eventually_accepting;
  $got = $got->complement;

  my $want = FLAT::Regex->new('0*|4*')->as_dfa;
  ok ($got->equals($want));
}

#------------------------------------------------------------------------------
# TurnRuns3 count by state machine

foreach my $side ('L','R') {
  my $bad = 0;
  my $f  = ($side eq 'L' ? TurnLpred_flat()   : TurnRpred_flat());
  my $f3 = ($side eq 'L' ? TurnLLLpred_flat() : TurnRRRpred_flat());
  foreach my $n (1 .. 125) {
    my $n5        = Math::BaseCnv::cnv($n,  10,5);
    my $n5_plus_1 = Math::BaseCnv::cnv($n+1,10,5);
    my $n5_plus_2 = Math::BaseCnv::cnv($n+2,10,5);
    my $by_three = $f->contains($n5)
      && $f->contains($n5_plus_1)
      && $f->contains($n5_plus_2);
    my $by_LLL = $f3->contains($n5);
    $by_three == $by_LLL or $bad++;
  }
  is($bad, 0, $f3->{'name'});
}

{
  my $f = TurnLLLpred_flat();
  my $max_k = 12;
  is(join(',', MyFLAT::FLAT_count_contains($f,$max_k)),
     join(',', map {1/2 * TurnRuns3($_)} 0 .. $max_k),
     'TurnRuns3 by state machine');
}
{
  my $f = TurnRRRpred_flat();
  my $max_k = 12;
  is(join(',', MyFLAT::FLAT_count_contains($f,$max_k)),
     join(',', map {1/2 * TurnRuns3($_)} 0 .. $max_k),
     'TurnRuns3 by state machine');
}

{
  my $f = TurnLLLpred_flat()->MyFLAT::digits_increment (add => -1);
  ok($f->intersect(TurnLpred_flat()) ->is_empty,
     'no runs LLLL');
}
{
  my $f = TurnRRRpred_flat()->MyFLAT::digits_increment (add => -1);
  ok($f->intersect(TurnRpred_flat()) ->is_empty,
     'no runs LLLL');
}

ok(  TurnLpred_flat()->contains('1'));
ok(  TurnLpred_flat()->contains('2'));
ok(! TurnLpred_flat()->contains('3'));
ok(TurnLLLpred_flat()->contains('10'));
ok(TurnRRRpred_flat()->contains(Math::BaseCnv::cnv(13,10,5)));

{
  #      L      L      R      R
  # LLRR   LLRR   LLRR   LLRR
  # ^               ^
  # 1    5     10  13
  #     10     20  23
  # RRR precedes R of previous level
  my $f = TurnRpred_flat()->MyFLAT::concat(FLAT::Regex->new('0')->as_dfa)
    ->MyFLAT::digits_increment (add => -2);
  ok(TurnRRRpred_flat()->equals($f));
}
{
  # LLL is at eeach L of previous level
  my $f = TurnLpred_flat()->MyFLAT::concat(FLAT::Regex->new('0')->as_dfa)
     ->MyFLAT::digits_increment (add => 0);
  ok(TurnLLLpred_flat()->equals($f));

  # MyFLAT::FLAT_show_breadth($f_last_by_concat, 3);
  # MyFLAT::FLAT_show_breadth($f_last_by_add, 3);
  # MyFLAT::FLAT_show_breadth(TurnLLLpred_flat(), 3);
}


#------------------------------------------------------------------------------
# hanging squares

{
  my $max_k = 3;
  my @by_HQRpred = MyFLAT::FLAT_count_contains(HQRpred_flat(),$max_k);
  my @by_HQR     = map {HQR($_)} 0 .. $max_k;
  is_deeply(\@by_HQRpred,\@by_HQR);
}


#------------------------------------------------------------------------------
# join area tree diameter

require MyGraphs;
foreach my $k (0 .. 3) {
  my $graph = r5join_area_tree_Graph($k);
  my $num_vertices = $graph->vertices;
  is (scalar($graph->edges), $num_vertices ? $num_vertices-1 : 0,
      "k=$k join num edges");

  is ($graph->diameter||0, JAdiameter($k), "k=$k JAdiameter");
  # is (MyGraphs::Graph_indnum($graph), JAindnum($k), "k=$k JAindnum");
  is (MyGraphs::Graph_diameter_count($graph), JAdiameterCount($k),
      "k=$k JAdiameterCount");
  is (MyGraphs::Graph_tree_domnum($graph), JAdomnum($k), "k=$k JAdomnum");
}

#------------------------------------------------------------------------------
# quad area tree diameter

require MyGraphs;
foreach my $k (0 .. 3) {
  my $graph = r5quad_area_tree_Graph($k);
  is (scalar($graph->edges), scalar($graph->vertices) - 1);

  is ($graph->diameter||0, QAdiameter($k), "k=$k QAdiameter");
  is (MyGraphs::Graph_diameter_count($graph), QAdiameterCount($k),
      "k=$k QAdiameterCount");
  is (MyGraphs::Graph_tree_domnum($graph), QAdomnum($k), "k=$k QAdomnum");
}

#------------------------------------------------------------------------------
# quad turn graph

require MyGraphs;
foreach my $k (0 .. 3) {
  my $graph = r5quad_turn_Graph($k);
  # MyGraphs::Graph_view($graph);
  is ($graph->diameter||0, QTdiameter($k), "k=$k QTdiameter");
}

#------------------------------------------------------------------------------
# area tree is repeated 5-star replacement

{
  my $k_limit = 4;
  require Graph;
  require MyGraphs;
  my $graph = Graph->new (undirected => 1);
  $graph->add_vertex(1);
  foreach my $k (0 .. $k_limit) {
    my $area_tree  = r5quad_area_tree_Graph($k);
    ok (MyGraphs::Graph_is_isomorphic($graph, $area_tree), "k=$k");

    unless ($k == $k_limit) {
      $graph = MyGraphs::Graph_star_replacement($graph,5);
    }
  }
}

#------------------------------------------------------------------------------
# Epred and DpredSecond

is_deeply([quad_expand(0,0, 1,0)],
          [[0,0], [1,0], [1,1], [0,1]]);
           

#------------------------------------------------------------------------------
# Epred and DpredSecond

ok(DpredSecond_flat()->equals(Epred_flat()->MyFLAT::digits_increment));
#
# MyFLAT::FLAT_show_breadth(DpredSecond_flat(), 2);
# MyFLAT::FLAT_show_breadth(Epred_flat(), 2);

{
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $bad = 0;
  my $n_hi = 256;
 N: foreach my $n (0 .. $n_hi) {
    my $str = Math::BaseCnv::cnv($n,10,5);
    foreach my $zeros (0 .. 3) {
      my $str = ('0' x $zeros) . $str;
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      {
        my $f = SpredInf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 1 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredInf_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
      {
        my $f = DpredSecond_flat();
        my $f_pred = $f->contains($str) ? 1 : 0;
        my $p_pred = @n_list == 2 && $n==$n_list[1] ? 1 : 0;
        if ($p_pred != $f_pred) {
          print "$f->{name} at N=$n $str got flat=$f_pred want path=$p_pred\n";
          last N if $bad++ > 10;
        }
      }
    }
  }
  is($bad, 0, "Xpred etc vs path");
}

#------------------------------------------------------------------------------
# Opred_flat

{
  # Opred_flat vs path
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $bad = 0;
  my $k = 5;
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my %seen;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($x,$y) = $path->n_to_xy($n);
    my @n_list = $path->xy_to_n_list($x,$y);
    @n_list = grep {$_ < $n_hi} @n_list;
    next unless @n_list == 2;
    my $delta = abs($n_list[0]-$n_list[1]);
    $seen{$delta} = 1;
  }
  foreach my $delta ($n_lo .. int($n_hi/5)) {
    my $delta5 = Math::BaseCnv::cnv($delta,10,5);
    {
      my $d_by_path = $seen{$delta} ? 1:0;
      my $d_by_flat = Opred_flat()->contains($delta5) ?1:0;
      if ($d_by_path != $d_by_flat) {
        diag "delta=$delta [$delta5] bad Opred_flat=$d_by_flat",
          " path=$d_by_path\n";
        last K if $bad++ > 10;
      }
    }
  }
  is($bad,0, 'Opred_flat() vs path');
}

#------------------------------------------------------------------------------
# Dpred by Rpred

{
  require MyFLAT;
  my $Rpred_YZnon = MyFLAT::aref_to_FLAT_DFA
    (\@Rpred_state_table,
     accepting_list => [2,3,4]); # Y,Z,non
  my $Rpred_XYnon = MyFLAT::aref_to_FLAT_DFA
    (\@Rpred_state_table,
     accepting_list => [1,2,4]); # X,Y,non

  my $DpredLeft_1 = $Rpred_YZnon->MyFLAT::concat(FLAT::Regex->new('1 0*'));
  my $DpredLeft_2 = $Rpred_XYnon->MyFLAT::concat(FLAT::Regex->new('2 0*'));
  my $f = $DpredLeft_1->union($DpredLeft_2)->as_dfa;

  ok ($f->equals(DpredLeft_k_flat()), 'DpredLeft_k by Rpred');
}

{
  require MyFLAT;
  my $Lpred_YZnon = MyFLAT::aref_to_FLAT_DFA
    (\@Lpred_state_table,
     accepting_list => [2,3,4]); # Y,Z,non
  my $Lpred_XYnon = MyFLAT::aref_to_FLAT_DFA
    (\@Lpred_state_table,
     accepting_list => [1,2,4]); # X,Y,non

  my $DpredRight_3 = $Lpred_XYnon->MyFLAT::concat(FLAT::Regex->new('3 0*'));
  my $DpredRight_4 = $Lpred_YZnon->MyFLAT::concat(FLAT::Regex->new('4 0*'));
  my $f = $DpredRight_3->union($DpredRight_4)->as_dfa;

  ok ($f->equals(DpredRight_k_flat()), 'DpredRight_k by Lpred');
}

#------------------------------------------------------------------------------
# other_arm_and_n()

{
  my $path = Math::PlanePath::R5DragonCurve->new(arms=>4);
  my $bad = 0;
  for (my $n4 = 4; $n4 < 4000; $n4+=4) {
    my ($x,$y) = $path->n_to_xy($n4);
    my @n4_list = $path->xy_to_n_list($x,$y);
    @n4_list = grep {$_ != $n4} @n4_list;
    @n4_list==1 or die;
    my $want_other_n = int(($n4_list[0]) / 4);
    my $want_other_arm = $n4_list[0] % 4;
    if ($want_other_arm==3) { $want_other_arm=-1; }

    my $n = $n4/4;
    my ($got_other_arm,$got_other_n) = other_arm_and_n($n);
    if ($got_other_arm != $want_other_arm
        || $got_other_n != $want_other_n) {
      diag " other_arm_and_n() n=$n want $want_other_arm, $want_other_n got $got_other_arm, $got_other_n";
      last if $bad++ > 10;
    }
  }
  is($bad,0, 'other_arm_and_n() vs path');
}


#------------------------------------------------------------------------------
# Dpred

ok (Dpred_k_flat()->equals(Spred_k_flat()->complement));
ok (Spred_k_flat()->equals(Dpred_k_flat()->complement));
ok (DpredInf_flat()->equals(SpredInf_flat()->complement));
ok (SpredInf_flat()->equals(DpredInf_flat()->complement));

ok (Dpred_k_LtoH_flat()->equals(Spred_k_LtoH_flat()->complement));
ok (Spred_k_LtoH_flat()->equals(Dpred_k_LtoH_flat()->complement));
ok (DpredInf_LtoH_flat()->equals(SpredInf_LtoH_flat()->complement));
ok (SpredInf_LtoH_flat()->equals(DpredInf_LtoH_flat()->complement));

ok (DpredLeft_k_flat()->is_subset_of(TurnLpred_flat()));
ok (DpredRight_k_flat()->is_subset_of(TurnRpred_flat()));

# singles are all on the boundary
# single-visited right turns are on the left boundary
# single-visited left turns are on the right boundary
ok (SpredRight_k_flat()->is_subset_of(Lpred_k_flat()));
ok (SpredLeft_k_flat()->is_subset_of(Rpred_flat()));
ok (Spred_k_flat()->is_subset_of(Bpred_k_flat()));

{
  # Dpred vs path
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $bad = 0;
 K: foreach my $k (0 .. 5) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      my $n5 = Math::BaseCnv::cnv($n,10,5);
      {
        my $d_by_path = @n_list==2 ? 1:0;
        my $d_by_flat = DpredInf_flat()->contains($n5) ?1:0;
        if ($d_by_path != $d_by_flat) {
          diag "n=$n [$n5] bad DpredInf_flat=$d_by_flat",
            " path=$d_by_path (",join(',',@n_list),")\n";
          last K if $bad++ > 10;
        }
      }
      @n_list = grep {$_ < $n_hi} @n_list;
      $n5 = sprintf '%0*s', $k, $n5;
      {
        my $d_by_flat = Dpred_k_flat()->contains($n5) ?1:0;
        my $d_by_path = @n_list==2 ? 1:0;
        if ($d_by_path != $d_by_flat) {
          diag "k=$k n=$n [$n5] bad Dpred_k_flat=$d_by_flat",
            " path=$d_by_path (",join(',',@n_list),")\n";
          last K if $bad++ > 10;
        }
      }
    }
  }
  is($bad,0, 'Dpred() vs path');
}

#------------------------------------------------------------------------------
# Rpred and Lpred both boundary

my @all_pairs = map { my $high=$_; map {"$high$_"} 0..4 } 0..4;
{
  my %disallowed;
  @disallowed{@Rpred_disallowed_pairs} = ();
  @disallowed{@Lpred_disallowed_pairs} = ();
  my @disallowed = sort keys %disallowed;

  my %allowed;
  @allowed{@all_pairs} = ();
  delete @allowed{@disallowed};
  my @allowed = sort keys %allowed;
  is_deeply([sort keys %allowed],
            ['00','01','02','03','13','31','41','42','43','44']);
  
  require FLAT::Regex;
  my $by_allowed = FLAT::Regex->new
    ('[]
      | 0 | 1 | 2 | 3 | 4
      | 0* 00 | 0* 01 | 0* 02 | 0* 03
      | 0* 13
      | 0* 31
      | 4* 13
      | 4* 31
      | 4* 41 | 4* 42 | 4* 43 | 4* 44')
    ->as_dfa;

  my $both_flat = Rpred_flat()->intersect(Lpred_k_flat())->as_dfa;
  ok ($both_flat->equals($by_allowed));

  my $numbers = FLAT::Regex->new
    ('[] | 0 | 1 | 2 | 3 | 4 |'
     . join('|',
            map { '0*'. sprintf '%02s',Math::BaseCnv::cnv($_,10,5) }
            0,1,2,3,8,16)
    )->as_dfa;
  my $numbers_flip = MyFLAT::FLAT_rename($numbers,
                                         symbols_map => {0=>4,4=>0, 1=>3,3=>1});
  my $by_numbers = $numbers->union($numbers_flip)->as_dfa;

  ok ($both_flat->equals($by_numbers));
  # MyFLAT::FLAT_check_is_equal($both_flat,$by_numbers);
}


#------------------------------------------------------------------------------
# Rpred

ok (Rpred_flat->equals(NonRpred_flat()->complement), 'Rpred complement');
ok (NonRpred_flat->equals(Rpred_flat()->complement));

{
  # Rpred vs path
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $bad = 0;
  foreach my $n (0 .. 1000) {
    my $by_path = $path->_UNDOCUMENTED__n_segment_is_right_boundary($n) ?1:0;
    my $by_pred = Rpred($n) ?1:0;
    if ($by_path != $by_pred) {
      $bad++;
    }
  }
  is($bad,0, 'Rpred() vs _UNDOCUMENTED__n_segment_is_right_boundary');
}

{
  # Rpred_flat() vs digit patterns
  require MyFLAT;
  require FLAT::Regex;

  # any of @Rpred_disallowed_pairs, with anything before or after
  my $re = '(0|1|2|3|4)* ('
    . join(' | ',
                map {substr($_,0,1).' 1* '.substr($_,1)}
                @Rpred_disallowed_pairs)
    .  ') (0|1|2|3|4)*';
  ### $re
  my $f = FLAT::Regex->new($re)->as_dfa;
  # MyFLAT::view($f);
  # MyFLAT::view(NonRpred_flat());
  
  ok($f->equals(NonRpred_flat()), 'Rpred by regex');
}

#------------------------------------------------------------------------------
# Lpred

ok (Lpred_k_flat->equals(NonLpred_k_flat()->complement));
ok (NonLpred_k_flat->equals(Lpred_k_flat()->complement));

{
  my @swapped = map { my $str = $_;
                      $str =~ tr/01234/43210/;
                      $str
                    } @Rpred_disallowed_pairs;
  @swapped = sort @swapped;
  is_deeply(\@Lpred_disallowed_pairs, \@swapped);
}

{
  # Lpred k vs path
  my $path = Math::PlanePath::R5DragonCurve->new;
  my $bad = 0;
  foreach my $k (0 .. 4) {
    my ($n_lo, $n_hi) = $path->level_to_n_range($k);
    foreach my $n ($n_lo .. $n_hi-1) {
      my $by_path = $path->_UNDOCUMENTED__n_segment_is_left_boundary($n,$k)?1:0;
      my $by_pred = Lpred_k($k,$n) ?1:0;
      if ($by_path != $by_pred) {
        $bad++;
      }
    }
  }
  is($bad,0, 'Lpred_k() vs _UNDOCUMENTED__n_segment_is_left_boundary');
}
{
  # Lpred_flat() vs digit patterns
  require MyFLAT;
  require FLAT::Regex;

  # any of @Lpred_disallowed_pairs, with anything before or after
  my $re = '(0|1|2|3|4)* ('
    . join(' | ',
                map {substr($_,0,1).' 3* '.substr($_,1)}
                @Lpred_disallowed_pairs)
    .  ') (0|1|2|3|4)*';
  ### $re
  my $f = FLAT::Regex->new($re)->as_dfa;
  # MyFLAT::view($f);
  # MyFLAT::view(NonLpred_k_flat());

  ok($f->equals(NonLpred_k_flat()), 'Lpred k by regex');
}

#------------------------------------------------------------------------------
# Bpred

# MyFLAT::FLAT_check_is_equal(Bpred_k_flat(),
#                             NonBpred_k_flat()->complement);
# MyFLAT::FLAT_check_is_equal(NonBpred_k_flat(),
#                             Bpred_k_flat()->complement);

ok (Bpred_k_flat->equals(NonBpred_k_flat()->complement));
ok (NonBpred_k_flat->equals(Bpred_k_flat()->complement));


#------------------------------------------------------------------------------
# $B_squares_segs characterizes eventually enclosed confs

{
  &fBpred_hullsaround_init(verbose=>2);
  my $bad = 0;
  foreach my $state ($fNonBpred_hullsaround_flat->get_states) {
    my $is_accepting = $fNonBpred_hullsaround_flat->is_accepting($state) ?1:0;
    my $conf = $fBpred_hullsaround_state_to_conf[$state];
    my $conf_is_enc
      = $B_squares_segs->difference($conf)->num_segs == 0 ?1:0;
    if ($is_accepting != $conf_is_enc) {
      $bad++;
    }
  }
  is ($bad, 0, "B_squares_segs = enclosed confs");
}

#------------------------------------------------------------------------------
# unfold()

{
  my ($x,$y) = unfold(0,0, 1,0, 0);
  is($x,0); is($y,0);
}
{
  my ($x,$y) = unfold(0,0, 1,0, 1);
  is($x,1); is($y,0);
}
{
  my ($x,$y) = unfold(0,0, 1,0, 2);
  is($x,1); is($y,1);
}
{
  my ($x,$y) = unfold(0,0, 1,0, 3);
  is($x,0); is($y,1);
}
{
  my ($x,$y) = unfold(0,0, 1,0, 4);
  is($x,0); is($y,2);
}

{
  my ($x,$y) = unfold(1,0, 1,0, 0);
  is($x,1); is($y,0);
}
{
  my ($x,$y) = unfold(1,0, 1,0, 1);
  is($x,1); is($y,1);
}
{
  my ($x,$y) = unfold(1,0, 1,0, 2);
  is($x,0); is($y,1);
}
{
  my ($x,$y) = unfold(1,0, 1,0, 3);
  is($x,0); is($y,2);
}
{
  my ($x,$y) = unfold(1,0, 1,0, 4);
  is($x,1); is($y,2);
}

#------------------------------------------------------------------------------
# MyR5DragonConf

{
  my $s = MyR5DragonConf->initial;
  is($s->num_segs, 1);
  ok($s->equals($s));

  foreach my $digit (0 .. 4) {
    my $e = $s->expand($digit);
    is($e->num_segs, 5);
  }
  ok(! $s->expand(0)->equals($s));
  ok(! $s->expand(1)->equals($s));

  # ok(  $s->is_boundary(0, $R_square));
  # ok(! $s->is_boundary(1, $R_square));
  # ok(! $s->is_boundary(0, $L_square));
  # ok(  $s->is_boundary(1, $L_square));

  # foreach my $c ($R_square, $L_square) {
  #     $c->print;
  #   foreach my $digit (0 .. 4) {
  #     $c->expand($digit)->print;
  #     ok( $c->is_subset( $c->expand($digit) ),
  #       "$c->{name} digit=$digit");
  #   }
  # }

  {
    my $B = $R_square_segs->union($L_square_segs)->set_name('B squares');
    ok($B->equals($B_squares_segs));
  }
  foreach my $digit (0 .. 4) {
    # $B_squares_segs->expand($digit)->print;
    ok( $B_squares_segs->is_subset( $B_squares_segs->expand($digit) ),
        "$B_squares_segs->{name} expand digit=$digit");
  }

  {
    my $c = $R_square_segs->clone(within => $B_squares_segs);
    ok (! $c->is_boundary(0));
    ok (  $c->is_boundary(1));
  }
  {
    my $c = $L_square_segs->clone(within => $B_squares_segs);
    ok (  $c->is_boundary(0));
    ok (! $c->is_boundary(1));
  }
}

# MyR5DragonConf->initial->print;
# MyR5DragonConf->initial->expand(1)->print;

# {
#   my $c = $hullsaround_segs_all->difference
#     (MyR5DragonConf->new(segs=>['3,0,1']));
#   $c = $c->clone(within => $hullsaround_segs_all);
#   # $c->print;
#   ok (! $c->is_boundary(0));
#   ok (  $c->is_boundary(1));
# }

#------------------------------------------------------------------------------
# hullsaround

{
  my $int = $hullsaround_segs_inner->intersection($hullsaround_segs_outer);
  is($int->num_segs, 0,
     "$hullsaround_segs_inner->{name} disjoint $hullsaround_segs_outer->{name}");
}

foreach my $digit (0 .. 4) {
  # $hullsaround_segs_all->expand($digit)->print;
  ok( $hullsaround_segs_all->is_subset( $hullsaround_segs_all->expand($digit) ),
      "$hullsaround_segs_all->{name} expand digit=$digit");
}

&fBpred_hullsaround_init(verbose=>2);
is ($fRpred_hullsaround_flat->num_states, 152);  # not minimized
is ($fLpred_hullsaround_flat->num_states, 152);
is ($fBpred_hullsaround_flat->num_states, 152);

is (scalar(@{$fBpred_hullsaround_state_enclosed_lists[0]}), 43);  # no eventually accepting
is (scalar(@{$fBpred_hullsaround_state_enclosed_lists[1]}), 43);
is (scalar(@{$fBpred_hullsaround_state_enclosed_lists[2]}), 1); 

{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullsaround_state_table,
     name => 'fNonRpred_hullsaround',
     accepting_list => $fBpred_hullsaround_state_enclosed_lists[0],
    );
  is ($f->MyFLAT::num_accepting, 43);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 3);
  is (scalar(@{$info{'states'}}), 96-43);
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullsaround_state_table,
     name => 'fNonRpred_hullsaround',
     accepting_list => $fBpred_hullsaround_state_enclosed_lists[1],
    );
  is ($f->MyFLAT::num_accepting, 43);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 3);
  is (scalar(@{$info{'states'}}), 96-43);
}
{
  my $f = MyFLAT::aref_to_FLAT_DFA
    (\@fBpred_hullsaround_state_table,
     name => 'fNonRpred_hullsaround',
     accepting_list => $fBpred_hullsaround_state_enclosed_lists[2],
    );
  is ($f->MyFLAT::num_accepting, 1);  # without eventually accepting
  my %info = $f->MyFLAT::get_eventually_accepting_info;
  is ($info{'depth'}, 3);
  is (scalar(@{$info{'states'}}), 58-1);
}

is ($fNonRpred_hullsaround_flat->MyFLAT::num_accepting, 96);
is ($fNonLpred_hullsaround_flat->MyFLAT::num_accepting, 96);
is ($fNonBpred_hullsaround_flat->MyFLAT::num_accepting, 58);

is ($fRpred_hullsaround_flat->num_states, 152);  # not minimized


ok ($fRpred_hullsaround_flat->equals(Rpred_flat()),
    'fRpred hullsaround = Rpred');
ok ($fLpred_hullsaround_flat->equals(Lpred_k_flat()));
ok ($fBpred_hullsaround_flat->equals(Bpred_k_flat()));

ok ($fNonRpred_hullsaround_flat->equals(NonRpred_flat()));
ok ($fNonLpred_hullsaround_flat->equals(NonLpred_k_flat()));
ok ($fNonBpred_hullsaround_flat->equals(NonBpred_k_flat()));


#------------------------------------------------------------------------------
# LowestNon0()

{
  my @LowestNon0_samples = (1,2,3,4,1,1,2,3,4,2,1,2,3,4,3,1,2,3,4,4);
  foreach my $i (0 .. $#LowestNon0_samples) {
    my $n = $i+1;
    my $got = LowestNon0($n);
    my $want = $LowestNon0_samples[$i];
    is($got, $want, "LowestNon0 n=$n");
  }
}

#------------------------------------------------------------------------------
# turn()

{
  my @turn_samples = (1,1,-1,-1,1,1,1,-1,-1,1,1,1,-1,-1,-1,1,1,-1,-1,-1,1,1,-1,-1);
  foreach my $i (0 .. $#turn_samples) {
    my $n = $i+1;
    my $got = turn($n);
    my $want = $turn_samples[$i];
    is ($got, $want, "turn n=$n");
  }
}

#------------------------------------------------------------------------------
# sequences

{
  my @A_samples = (0, 0, 4, 36, 232, 1320, 7084, 36876);
  foreach my $k (0 .. $#A_samples) {
    my $got = A($k);
    my $want = $A_samples[$k];
    is($got, $want, "A k=$k");
  }
}
{
  my @J_samples = (0, 1, 4, 13, 40, 121, 364, 1093, 3280);
  foreach my $k (0 .. $#J_samples) {
    my $got = J($k);
    my $want = $J_samples[$k];
    is($got, $want, "J k=$k");
  }
}

#------------------------------------------------------------------------------
# mul()

{
  my ($x,$y) = mul(1,0);
  is($x, 1);
  is($y, 2);
  ($x,$y) = mul($x,$y);
  is($x, -3);
  is($y, 4);
  ($x,$y) = mul($x,$y);
  is($x, -11);
  is($y, -2);
}

#------------------------------------------------------------------------------
# turn graph is cycle replacement of previous area tree,
# but only with degree-4 replacements taking the existing edges according
# to the geometry to preserve arms order

require MyGraphs;
foreach my $k (1 .. 4) {
  my $turn_graph = r5quad_turn_Graph($k);
  ok ($turn_graph->is_undirected);
  my $area_tree  = r5quad_area_tree_Graph($k-1);
  ok ($area_tree->is_undirected);

  foreach my $xy (1) {
    my $replacement = MyGraphs::Graph_cycle_replacement($area_tree, 4,
                                                        edges_between => 2);
    ok ($replacement->is_undirected);

    ok (MyGraphs::Graph_is_isomorphic($replacement, $turn_graph),
        "k=$k xy=$xy");

    # # without the xy vertex naming gives possible different ordering around
    # # the cycle
    # $area_tree->set_graph_attribute('vertex_name_type',undef);
  }

  # my $s1 = join(',',MyGraphs::Graph_degree_sequence($turn_graph));
  # my $s2 = join(',',MyGraphs::Graph_degree_sequence($replacement));
  # print "$s1\n";
  # print "$s2\n";
  # is ($s1,$s2);
  # print MyGraphs::graph6_str_to_canonical(MyGraphs::Graph_to_graph6_str($turn_graph));
  # print MyGraphs::graph6_str_to_canonical(MyGraphs::Graph_to_graph6_str($replacement));
  # MyGraphs::Graph_view($turn_graph);
  # MyGraphs::Graph_view($area_tree);
  # MyGraphs::Graph_view($replacement);
}

#------------------------------------------------------------------------------
# turn graph is not line graph of area tree,
# since line graph makes complete-4 instead of 4-cycle

require MyGraphs;
foreach my $k (1 .. 3) {
  my $turn_graph = r5quad_turn_Graph($k);
  my $area_tree  = r5quad_area_tree_Graph($k);
  my $line_graph = MyGraphs::Graph_line_graph($area_tree);
  ok (! MyGraphs::Graph_is_isomorphic($line_graph, $turn_graph));
  # MyGraphs::Graph_view($turn_graph);
  # MyGraphs::Graph_view($area_tree);
  # MyGraphs::Graph_view($line_graph);
}

#------------------------------------------------------------------------------
# Hwp, Hep vs convex hull

ok (Hwp_factor - 1.418776 < 1e-6, "Hwp_factor ".Hwp_factor);

{
  # Hwp, Hep vs convex hull
  my $verbose = 0;
  my $max_k = 30;

  my $points = [ [to_bigint(0),to_bigint(0)],
                 [to_bigint(1),to_bigint(0)] ];
  my $ex = to_bigint(1);
  my $ey = to_bigint(0);

  foreach my $k (0 .. $max_k) {
    my $count = scalar(@$points);
    if ($verbose) {
      print "k=$k count $count  ending $ex,$ey\n";
    }

    my $hwp_norm = -1;
    my @hwp_i = 0;
    my ($hwp_x,$hwp_y);
    my $hep_norm = -1;
    my @hep_i = 0;
    my ($hep_x,$hep_y, $hep_wx,$hep_wy);
    foreach my $i (0 .. $#$points) {
      my $p = $points->[$i];
      my ($x,$y) = @$p;

      # z * conj(b^k) directs end towards X axis, Hwp=-Im Hep=-Re
      # (x+I*y)*conj(ex+ey*I) == (x*ex+y*ey) + (x*-ey+y*ex)*I
      my $wx = $x*$ex + $y*$ey;
      my $wy = $x*-$ey + $y*$ex;
      {
        my $m = -$wy;
        if ($m > $hwp_norm) {
          $hwp_norm = $m; @hwp_i = ();
          $hwp_x = $x;
          $hwp_y = $y;
        }
        if ($m == $hwp_norm) { push @hwp_i, $i; }
      }
      {
        my $m = -$wx;
        if ($m > $hep_norm) {
          $hep_norm = $m; @hep_i = ();
          $hep_x = $x;
          $hep_y = $y;
          $hep_wx = $x;
          $hep_wy = $y;
        }
        if ($m == $hep_norm) { push @hep_i, $i; }
      }
    }
    if ($verbose) {
      my $decimals = 35;
      $hwp_norm *= to_bigint(10)**$decimals;
      $hwp_norm /= to_bigint(5)**$k;
      printf "  hwp at i=".join(',',@hwp_i)." = %0*s\n", $decimals, $hwp_norm;
      print "  coords   $hwp_x,$hwp_y\n";

      $hep_norm *= to_bigint(10)**$decimals;
      $hep_norm /= to_bigint(5)**$k;
      printf "  hep at i=".join(',',@hep_i)." = %0*s\n", $decimals, $hep_norm;
      print "  coords   $hep_x,$hep_y  cf wxy $hep_wx,$hep_wy\n";
    }

    {
      my ($got_x,$got_y) = Hwp(to_bigint($k));
      is ($got_x, $hwp_x, "Hwp $k");
      is ($got_y, $hwp_y);
    }
    {
      my ($got_x,$got_y) = Hep(to_bigint($k));
      is ($got_x, $hep_x, "Hep $k");
      is ($got_y, $hep_y);
    }

    ($points, $ex,$ey) = hull_expand($points, $ex,$ey);
  }
}


#------------------------------------------------------------------------------
exit 0;
